/*
Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
This file contains trade secrets of Honeywell International, Inc. No part
may be reproduced or transmitted in any form by any means or for any
purpose without the express written permission of Honeywell.
*/

package com.honeywell.honplantcontroller.hmi.alarm;

import java.text.MessageFormat;
import java.util.Iterator;
import java.util.Optional;
import java.util.logging.Logger;

import javax.baja.alarm.BAlarmRecord;
import javax.baja.alarm.BAlarmService;
import javax.baja.alarm.ext.BAlarmState;
import javax.baja.alarm.ext.offnormal.BBooleanChangeOfStateAlgorithm;
import javax.baja.alarm.ext.offnormal.BBooleanCommandFailureAlgorithm;
import javax.baja.alarm.ext.offnormal.BEnumChangeOfStateAlgorithm;
import javax.baja.alarm.ext.offnormal.BEnumCommandFailureAlgorithm;
import javax.baja.bacnet.BacnetAlarmConst;
import javax.baja.bacnet.datatypes.BBacnetObjectIdentifier;
import javax.baja.bacnet.datatypes.BBacnetTimeStamp;
import javax.baja.bacnet.datatypes.BBacnetUnsigned;
import javax.baja.bacnet.enums.BBacnetBinaryPv;
import javax.baja.bacnet.enums.BBacnetEngineeringUnits;
import javax.baja.bacnet.enums.BBacnetEventState;
import javax.baja.control.BControlPoint;
import javax.baja.naming.BOrd;
import javax.baja.naming.BOrdList;
import javax.baja.nre.util.ByteArrayUtil;
import javax.baja.sys.BBoolean;
import javax.baja.sys.BComponent;
import javax.baja.sys.BDouble;
import javax.baja.sys.BEnumRange;
import javax.baja.sys.BFacets;
import javax.baja.sys.BIBoolean;
import javax.baja.sys.BIEnum;
import javax.baja.sys.BINumeric;
import javax.baja.sys.BObject;
import javax.baja.sys.BString;
import javax.baja.sys.BValue;
import javax.baja.sys.Sys;
import javax.baja.units.BUnit;

import com.tridium.bacnet.asn.AsnUtil;
import com.tridium.bacnet.stack.server.BBacnetExportTable;

public class BacnetAlarmSummary {
  
  protected BBacnetObjectIdentifier objectIdentifier;
  protected BBacnetEventState eventState;
  protected BBacnetTimeStamp eventTimeStamp;
  protected BBacnetUnsigned priority;
  protected BValue alarmPresentValue;
  protected BBacnetEngineeringUnits unit;

  public BacnetAlarmSummary() {
    //
  }

  /**
   * Factory method to create an HMI AlarmSummary from given AlarmRecord
   * @param alarmRecord
   * @param exportTable
   * @return
   */
  public static BacnetAlarmSummary make(BAlarmRecord alarmRecord, BBacnetExportTable exportTable) {

    if(alarmRecord == null || exportTable == null) {
      return null;
    }
    BacnetAlarmSummary bacnetAlarmSummary = new BacnetAlarmSummary();

    try {
      if(bacnetAlarmSummary.initialize(alarmRecord, exportTable)) {
        return bacnetAlarmSummary;
      }
      
    } catch (Exception e) {
      LOG.warning(()->
        MessageFormat.format("could not make BacnetAlarmSummary for given alarmRecord {0} because {1}", 
        alarmRecord, e));
    }
    return null;

  }

  protected boolean initialize(BAlarmRecord alarmRecord, BBacnetExportTable exportTable) {
    BControlPoint controlPoint = getAlarmSourceControlPointOrd(alarmRecord);

    if(controlPoint != null) {
      BBacnetObjectIdentifier bacnetObjectId = exportTable.byObjectOrd(controlPoint.getHandleOrd());

      if(bacnetObjectId == null) {
        return false;
      }

      BBacnetUnsigned bacnetPriority = new BBacnetUnsigned(alarmRecord.getPriority());
      
      BFacets alarmData = alarmRecord.getAlarmData();
      
      Optional<BValue> alarmValueOptional = getAlarmValue(alarmData, controlPoint);
      if(!alarmValueOptional.isPresent()) {
        return false;
      }
      
      BUnit niagaraUnit = (BUnit)controlPoint.getFacets().get(BFacets.UNITS);
  
      BString toState = (BString) alarmData.get(BAlarmRecord.TO_STATE);
      
      BAlarmState alarmState = BAlarmState.make(toState.getString());
  
      BBacnetEventState bacnetEventState = BBacnetEventState.make(alarmState);
      
      BBacnetEngineeringUnits bacnetUnits = BBacnetEngineeringUnits.make(niagaraUnit);

      this.objectIdentifier = bacnetObjectId;
      this.eventState = bacnetEventState;
      this.eventTimeStamp = new BBacnetTimeStamp(alarmRecord.getTimestamp());
      this.priority = bacnetPriority;
      this.alarmPresentValue = alarmValueOptional.get();
      this.unit = bacnetUnits;
      return true;
    }

    return false;
  }

  /**
   * Helper method to retrieve the real control point for a given alarm record
   * @param alarmRecord
   * @return
   */
  public BControlPoint getAlarmSourceControlPointOrd(BAlarmRecord alarmRecord) {

    if(alarmRecord == null) {
      return null;
    }

    BOrdList sources = alarmRecord.getSource();
    Iterator<BOrd> sourceIterator = sources.iterator();

    while (sourceIterator.hasNext()) {
      BOrd sourceOrd = sourceIterator.next();

      BObject source = sourceOrd.resolve(Sys.getStation()).get();

      if(source instanceof BControlPoint) {
        return (BControlPoint) source;
      } else {
        if(source instanceof BComponent) {
          BComponent parent = source.asComplex().getParent().asComponent();
          if(parent instanceof BControlPoint) {
            return (BControlPoint) parent;
          }
        }
      }
      
    }

    return null;
  }

  public static BComponent getEventObject(BAlarmRecord alarmRecord) {
    BComponent eventObject = null;
    BAlarmService base = (BAlarmService) Sys.getService(BAlarmService.TYPE);
    try {
      BObject alarmSource = alarmRecord.getSource().get(0).get(base);
      if (alarmSource instanceof BComponent) {
        eventObject = alarmSource.asComplex().getParent().asComponent();
      }
    } catch (Exception e) {
      LOG.warning(()->
      MessageFormat.format("could not retrieve source point for given alarmRecord {0} because {1}", 
      alarmRecord, e));
    }
    return eventObject;
  }

  public Optional<BValue> getAlarmValue(BFacets alarmData, BControlPoint controlPoint) {
      BValue alarmValue = (BValue)alarmData.get(BAlarmRecord.ALARM_VALUE);

      if(null != alarmValue) {
        return Optional.of(alarmValue);
      }

      return getAlarmValueForBacnetAlarm(alarmData, controlPoint);
  }

  public Optional<BValue> getAlarmValueForBacnetAlarm(BFacets alarmData, BControlPoint controlPoint) {

    if(controlPoint instanceof BINumeric) {
      BValue alarmValue = (BValue)alarmData.get(BacnetAlarmConst.BAC_EXCEEDING_VALUE);

      return Optional.of(BDouble.make(((BString)alarmValue).getString()));
    }

    BString newState  = (BString) alarmData.get(BacnetAlarmConst.BAC_NEW_STATE);

    if(null != newState) {
      return extractBacnetAlarmValueFromInputObject(newState, controlPoint);
    }

    BString commandValue  = (BString) alarmData.get(BacnetAlarmConst.BAC_COMMAND_VALUE);
    if(null!= commandValue){
      return extractBacnetAlarmValueFromOutputObject(commandValue, controlPoint);
    }

    return Optional.empty();

  }

  /**
   * Parse alarm value, that was generated by a {@link BBooleanCommandFailureAlgorithm} or {@link BEnumCommandFailureAlgorithm} 
   * @param value
   * @param controlPoint
   * @return
   */
  private Optional<BValue> extractBacnetAlarmValueFromOutputObject(BString value, BControlPoint controlPoint) {
    byte[] valueAsByte = ByteArrayUtil.hexStringToBytes(value.toString());

    LOG.fine(()-> MessageFormat.format("raw bytes {0} to parse", 
      ByteArrayUtil.toHexString(valueAsByte)));

    try {
      if(controlPoint instanceof BIBoolean) {
        BBoolean booleanValue = BBoolean.make(BBacnetBinaryPv.make(AsnUtil.fromAsnEnumerated(valueAsByte)).isActive());
        return Optional.of(booleanValue);
      }else if(controlPoint instanceof BIEnum) {
        int index = AsnUtil.fromAsnUnsignedInt(valueAsByte);
        return Optional.of(getValueForEnum(controlPoint, index));
      }
    } catch (Exception e) {
      LOG.warning(()-> MessageFormat.format("could not extract value from {0} because {1}", value ,e));
    }
    return Optional.empty();
  }

  /**
   * Parse alarm value, that was generated by {@link BBooleanChangeOfStateAlgorithm} or {@link BEnumChangeOfStateAlgorithm}
   * @param newState
   * @param controlPoint
   * @return
   */
  private Optional<BValue> extractBacnetAlarmValueFromInputObject(BString newState, BControlPoint controlPoint) {

    String newStateString = newState.getString();
    if(!newStateString.contains("NBacnetPropertyStates:")) {
      return Optional.empty();
    }

    if(controlPoint instanceof BIBoolean) {
      String[] x = newStateString.split("binaryValue:");

      return Optional.of(BBoolean.make(BBacnetBinaryPv.make(x[1].trim().toLowerCase()).isActive()));

    }else if(controlPoint instanceof BIEnum) {
      String[] x = newStateString.split("unsignedValue:");

      return Optional.of(getValueForEnum(controlPoint, Integer.parseInt(x[1].trim())));
    }

    return Optional.empty();

  }

  /**
   * retrieve the alarm value, based on the enum range of the point 
   * @param controlPoint
   * @param enumIndex
   * @return
   */
  private BValue getValueForEnum(BControlPoint controlPoint, int enumIndex) {
    BIEnum enumPoint = (BIEnum)controlPoint;
    BEnumRange enumRange = (BEnumRange) enumPoint.getEnumFacets().get("range");
    return enumRange.get(enumIndex);
  }

  public String toDebugString() {
    StringBuilder sb = new StringBuilder();
    sb.append("objectIdentifier: " + objectIdentifier);
    sb.append("\neventState: " + eventState);
    sb.append("\neventTimeStamp: " + eventTimeStamp);
    sb.append("\npriority: " + priority);
    sb.append("\nalarmPresentValue: " + alarmPresentValue);
    sb.append("\nunit: " + unit);

    return sb.toString();
  }

  public BBacnetObjectIdentifier getObjectIdentifier() {
      return objectIdentifier;
  }

  public BBacnetEventState getEventState() {
      return eventState;
  }
  public BBacnetTimeStamp getEventTimeStamp() {
      return eventTimeStamp;
  }

  public BBacnetUnsigned getPriority() {
      return priority;
  }
  public BValue getAlarmPresentValue() {
      return alarmPresentValue;
  }
  public BBacnetEngineeringUnits getUnit() {
      return unit;
  }

  private static final Logger LOG = Logger.getLogger(BacnetAlarmSummary.class.getName());
}
