/*
Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
This file contains trade secrets of Honeywell International, Inc. No part
may be reproduced or transmitted in any form by any means or for any
purpose without the express written permission of Honeywell.
*/

package com.honeywell.honplantcontroller.hmi.comm;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.logging.Level;
import java.util.stream.Collectors;

import javax.baja.alarm.BAlarmRecord;
import javax.baja.alarm.BIAlarmSource;
import javax.baja.bacnet.BacnetException;
import javax.baja.bacnet.export.BIBacnetExportObject;
import javax.baja.control.BControlPoint;
import javax.baja.nre.annotations.NiagaraType;
import javax.baja.sys.BAbsTime;
import javax.baja.sys.BComponent;
import javax.baja.sys.Sys;
import javax.baja.sys.Type;

import com.honeywell.honplantcontroller.hmi.alarm.BacnetAlarmSummary;
import com.honeywell.honplantcontroller.hmi.alarm.IAlarmResolver;
import com.honeywell.honplantcontroller.hmi.enums.BHMIAlarmLevelEnum;
import com.honeywell.honplantcontroller.hmi.messages.QueryAlarmPointsResultMessage;
import com.honeywell.honplantcontroller.hmi.messages.QueryUnackAlarmPointsResultMessage;
import com.honeywell.honplantcontroller.hmi.messages.RequestAlarmAckMessage;
import com.honeywell.honplantcontroller.hmi.service.BHonPlantControllerHMI;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerBacnetErrorFactory;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerHMILogger;
import com.tridium.bacnet.stack.server.BBacnetExportTable;

@NiagaraType
public class BHMIAlarmHandler extends BComponent {
/*+ ------------ BEGIN BAJA AUTO GENERATED CODE ------------ +*/
/*@ $com.honeywell.honplantcontroller.hmi.comm.BHMIAlarmHandler(2979906276)1.0$ @*/
/* Generated Wed Aug 17 10:52:21 CEST 2022 by Slot-o-Matic (c) Tridium, Inc. 2012 */

////////////////////////////////////////////////////////////////
// Type
////////////////////////////////////////////////////////////////
  
  @Override
  public Type getType() { return TYPE; }
  public static final Type TYPE = Sys.loadType(BHMIAlarmHandler.class);

/*+ ------------ END BAJA AUTO GENERATED CODE -------------- +*/
  

  /**
   * Query all alarms
   * @param alarmLevel
   * @param startIndex
   * @param stopIndex
   * @param hmiAlarmResolver
   * @param exportTable
   * @return
   */
  public QueryAlarmPointsResultMessage queryAlarms(BHMIAlarmLevelEnum alarmLevel, int startIndex, int stopIndex, IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable) throws BacnetException {
		setTimeOfQueryAlarm(MessageFormat.format("handleQueryUnackCalled at {0}", BAbsTime.now().getMillis()));
		checkIndicesValid(startIndex, stopIndex);
		setTimeAfterCheckingIndicesValid(MessageFormat.format("time after checkIndicesValid is {0}", BAbsTime.now().getMillis()));
		hmiAlarmResolver.collectActiveAlarms((BHonPlantControllerHMI) getParent());
		setTimeAfterCollectingALarms(MessageFormat.format("time after collecting alarms is {0}", BAbsTime.now().getMillis()));
		if (checkOnlyReadAlarmCounts(startIndex, stopIndex)) {
			return alarmCountsResult(hmiAlarmResolver);
		}

    return getAlarms(alarmLevel, startIndex, stopIndex, hmiAlarmResolver, exportTable);
  }

  /**
   * Query un-acknowledged alarms
   * @param alarmLevel
   * @param startIndex
   * @param stopIndex
   * @param hmiAlarmResolver
   * @param exportTable
   * @return
   */
  public QueryUnackAlarmPointsResultMessage queryUnAckAlarms(BHMIAlarmLevelEnum alarmLevel, int startIndex, int stopIndex, IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable) throws BacnetException {
	  setTimeOfQueryAlarm(MessageFormat.format("handleQueryUnackCalled at {0}", BAbsTime.now().getMillis()));
    checkIndicesValid(startIndex, stopIndex);
    setTimeAfterCheckingIndicesValid(MessageFormat.format("time after checkIndicesValid is {0}", BAbsTime.now().getMillis()));
    hmiAlarmResolver.collectUnackedAlarms((BHonPlantControllerHMI) getParent());
    setTimeAfterCollectingALarms(MessageFormat.format("time after collecting alarms is {0}", BAbsTime.now().getMillis()));
    if(checkOnlyReadAlarmCounts(startIndex, stopIndex)) {
      return unAckAlarmCountsResult(hmiAlarmResolver);
    }

    return getUnAckAlarms(alarmLevel, startIndex, stopIndex, hmiAlarmResolver, exportTable);
  }
  
  protected void checkIndicesValid(int startIndex, int stopIndex) throws BacnetException {
    //indices are not negative
    //stopIndex is not less then startIndex, but not zero

    if(startIndex < 0 || stopIndex < 0 || !(stopIndex >= 0 && stopIndex >= startIndex)) {
      throw HonPlantControllerBacnetErrorFactory.makeInvalidIndicesError();

    }
  }

  /**
   * Check if HMI only wants to read the alarm counts
   * @param startIndex
   * @param stopIndex
   * @return
   */
  protected boolean checkOnlyReadAlarmCounts(int startIndex, int stopIndex) {
    return startIndex == 0 && stopIndex == 0;
  }

  protected QueryAlarmPointsResultMessage alarmCountsResult(IAlarmResolver hmiAlarmResolver) throws BacnetException  {
    
    long alarmPointsCount = hmiAlarmResolver.getAlarmCount().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);

    long urgentAlarmCount = hmiAlarmResolver.getUrgentAlarmCount().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    long highAlarmCount = hmiAlarmResolver.getHighAlarmCount().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    long lowAlarmCount = hmiAlarmResolver.getLowAlarmCount().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);

    return QueryAlarmPointsResultMessage.make(alarmPointsCount, urgentAlarmCount, highAlarmCount, lowAlarmCount);
  }

  protected QueryUnackAlarmPointsResultMessage unAckAlarmCountsResult(IAlarmResolver hmiAlarmResolver) throws BacnetException {
    
    long alarmPointsCount = hmiAlarmResolver.getUnAckAlarmCount().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    long urgentAlarmCount = hmiAlarmResolver.getUrgentUnAckAlarmCount().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    long highAlarmCount = hmiAlarmResolver.getHighUnAckAlarmCount().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    long lowAlarmCount = hmiAlarmResolver.getLowUnAckAlarmCount().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);


    return QueryUnackAlarmPointsResultMessage.make(alarmPointsCount, urgentAlarmCount, highAlarmCount, lowAlarmCount);
  }

  protected QueryAlarmPointsResultMessage getAlarms(BHMIAlarmLevelEnum alarmLevel, int startIndex, int stopIndex, IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable) throws BacnetException {
    final String LOG_MSG_PREFIX = BHMIAlarmHandler.TYPE.getTypeName() + "." + getClass().getSimpleName() + ".getAlarms";

    List<BacnetAlarmSummary> alarmSummaries = new ArrayList<>();
    switch (alarmLevel.getOrdinal()) {
      case BHMIAlarmLevelEnum.ALL_ALARM:
        alarmSummaries = getAlarms(hmiAlarmResolver, exportTable, startIndex, stopIndex);
        break;

      case BHMIAlarmLevelEnum.URGENT_ALARM:
        alarmSummaries = getUrgentAlarms(hmiAlarmResolver, exportTable, startIndex, stopIndex);
        break;

      case BHMIAlarmLevelEnum.HIGH_ALARM:
        alarmSummaries = getHighAlarms(hmiAlarmResolver, exportTable, startIndex, stopIndex);
        break;

      case BHMIAlarmLevelEnum.LOW_ALARM:
        alarmSummaries = getLowAlarms(hmiAlarmResolver, exportTable, startIndex, stopIndex);
        break;
    
      default:
        break;
    }

    QueryAlarmPointsResultMessage hmiAlarmResult = alarmCountsResult(hmiAlarmResolver);
    hmiAlarmResult =  QueryAlarmPointsResultMessage.make(hmiAlarmResult, alarmSummaries);
    HonPlantControllerHMILogger.checkAndLog(Level.FINEST, MessageFormat.format("{0} return alarm result {1}", LOG_MSG_PREFIX, hmiAlarmResult));
    return hmiAlarmResult;
  }

  protected QueryUnackAlarmPointsResultMessage getUnAckAlarms(BHMIAlarmLevelEnum alarmLevel, int startIndex, int stopIndex, IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable) throws BacnetException {

    List<BacnetAlarmSummary> alarmSummaries = new ArrayList<>();
    switch (alarmLevel.getOrdinal()) {
      case BHMIAlarmLevelEnum.ALL_ALARM:
        alarmSummaries = getUnAckAlarms(hmiAlarmResolver, exportTable, startIndex, stopIndex);

        break;

      case BHMIAlarmLevelEnum.URGENT_ALARM:

        alarmSummaries = getUrgentUnAckAlarms(hmiAlarmResolver, exportTable, startIndex, stopIndex);
        
        break;

      case BHMIAlarmLevelEnum.HIGH_ALARM:
        alarmSummaries = getHighUnAckAlarms(hmiAlarmResolver, exportTable, startIndex, stopIndex);
        break;

      case BHMIAlarmLevelEnum.LOW_ALARM:
        alarmSummaries = getLowUnAckAlarms(hmiAlarmResolver, exportTable, startIndex, stopIndex);
        break;
    
      default:
        break;
    }

    QueryUnackAlarmPointsResultMessage hmiAlarmResult = unAckAlarmCountsResult(hmiAlarmResolver);
    return QueryUnackAlarmPointsResultMessage.make(hmiAlarmResult, alarmSummaries);
  }

  protected List<BacnetAlarmSummary> getAlarms(IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable, int startIndex, int stopIndex) throws BacnetException {
    List<BAlarmRecord> urgentAlarms = hmiAlarmResolver.getAlarms().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    return mapAlarmRecordsToAlarmSummary(urgentAlarms, startIndex, stopIndex, exportTable);
  }

  protected List<BacnetAlarmSummary> getUrgentAlarms(IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable, int startIndex, int stopIndex) throws BacnetException {
    
    List<BAlarmRecord> urgentAlarms = hmiAlarmResolver.getUrgentAlarms().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    return mapAlarmRecordsToAlarmSummary(urgentAlarms, startIndex, stopIndex, exportTable);
  }

  protected List<BacnetAlarmSummary> getHighAlarms(IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable, int startIndex, int stopIndex) throws BacnetException {

    List<BAlarmRecord> highAlarms = hmiAlarmResolver.getHighAlarms().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    return mapAlarmRecordsToAlarmSummary(highAlarms, startIndex, stopIndex, exportTable);
  }

  protected List<BacnetAlarmSummary> getLowAlarms(IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable, int startIndex, int stopIndex) throws BacnetException {

    List<BAlarmRecord> lowAlarms = hmiAlarmResolver.getLowAlarms().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    return mapAlarmRecordsToAlarmSummary(lowAlarms, startIndex, stopIndex, exportTable);
  }

  protected List<BacnetAlarmSummary> getUnAckAlarms(IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable, int startIndex, int stopIndex) throws BacnetException {
    List<BAlarmRecord> unackAlarms = hmiAlarmResolver.getUnAckAllAlarms().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    return mapAlarmRecordsToAlarmSummary(unackAlarms, startIndex, stopIndex, exportTable);
  }

  protected List<BacnetAlarmSummary> getUrgentUnAckAlarms(IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable, int startIndex, int stopIndex) throws BacnetException {
    List<BAlarmRecord> urgentUnAckAlarms = hmiAlarmResolver.getUrgentUnAckAlarms().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    return mapAlarmRecordsToAlarmSummary(urgentUnAckAlarms, startIndex, stopIndex, exportTable);
  }

  protected List<BacnetAlarmSummary> getHighUnAckAlarms(IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable, int startIndex, int stopIndex) throws BacnetException {
    List<BAlarmRecord> highUnAckAlarms = hmiAlarmResolver.getHighUnAckAlarms().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    return mapAlarmRecordsToAlarmSummary(highUnAckAlarms, startIndex, stopIndex, exportTable);

  }

  protected List<BacnetAlarmSummary> getLowUnAckAlarms(IAlarmResolver hmiAlarmResolver, BBacnetExportTable exportTable, int startIndex, int stopIndex) throws BacnetException {
    List<BAlarmRecord> lowUnAckAlarms = hmiAlarmResolver.getLowUnAckAlarms().orElseThrow(HonPlantControllerBacnetErrorFactory::makeAlarmDbError);
    return mapAlarmRecordsToAlarmSummary(lowUnAckAlarms, startIndex, stopIndex, exportTable);

  }

  protected List<BacnetAlarmSummary> mapAlarmRecordsToAlarmSummary(List<BAlarmRecord> alarmRecords, int startIndex, int stopIndex, BBacnetExportTable exportTable) {
    if(alarmRecords.isEmpty()) {
      return new ArrayList<>();
    }
    
    int fromIndex = 0;
    int toIndex = alarmRecords.size()-1;

    if(validateIndicesForList(alarmRecords.size(), startIndex, stopIndex)) {
      fromIndex = startIndex;
      toIndex = stopIndex;
    }

    return alarmRecords.subList(fromIndex, toIndex)
                        .stream()
                        .filter(Objects::nonNull)
                        .map(alarmRecord -> BacnetAlarmSummary.make(alarmRecord, exportTable))
                        .filter(Objects::nonNull)
                        .collect(Collectors.toList());

  }

  protected boolean validateIndicesForList(int alarmRecordsSize, int startIndex, int stopIndex) {
    final String LOG_MSG_PREFIX = BHMIAlarmHandler.TYPE.getTypeName() + "." + getClass().getSimpleName() + ".validateIndicesForList";

    if(startIndex > alarmRecordsSize) {
      HonPlantControllerHMILogger.checkAndLog(Level.FINEST, MessageFormat.format("{0} startIndex {1} > alarmRecordsSize {2} ", LOG_MSG_PREFIX, startIndex, alarmRecordsSize));
      return INVALID_INDICES_IN_DB;
    }
    
    if(stopIndex > alarmRecordsSize) {
      HonPlantControllerHMILogger.checkAndLog(Level.FINEST, MessageFormat.format("{0} stopIndex {1} > alarmRecordsSize {2} ", LOG_MSG_PREFIX, stopIndex, alarmRecordsSize));
      return INVALID_INDICES_IN_DB;
    }

    return VALID_INDICES_IN_DB;
  }

  boolean proccessAlarmAckRequest(RequestAlarmAckMessage alarmAck, BBacnetExportTable exportTable, IAlarmResolver alarmResolver) {
    final String LOG_MSG_PREFIX = BHMIAlarmHandler.TYPE.getTypeName() + "." + getClass().getSimpleName() + ".proccessAlarmAckRequest";
    BIBacnetExportObject export = exportTable.byObjectId(alarmAck.getEventObjectId());

    if (null == export) {
      HonPlantControllerHMILogger.checkAndLog(Level.WARNING, 
        MessageFormat.format("{0} can not process alarm ack, no object available for bacnet object id {1} ", 
        LOG_MSG_PREFIX, alarmAck.getEventObjectId()));
      return false;
    }

    Object exportedObject = export.getObject();
    if (null == exportedObject) {
      HonPlantControllerHMILogger.checkAndLog(Level.WARNING, 
        MessageFormat.format("{0} can not process alarm ack, no exported object/point available ", 
        LOG_MSG_PREFIX));
      return false;
    }
    BControlPoint point = (BControlPoint) exportedObject;
	List<BIAlarmSource> alarmsSourceExtList = Arrays.stream(point.getChildren(BIAlarmSource.class)).filter(Objects::nonNull)
			.sequential().collect(Collectors.toList());
	boolean value = false;
	if (alarmsSourceExtList.size() > 0) {
		for (int i = 0; i < alarmsSourceExtList.size(); i++) {
			Optional<BIAlarmSource> alarmSourceOptionalss = Optional.of(alarmsSourceExtList.get(i));
			if (alarmSourceOptionalss.isPresent()) {
				value = alarmResolver.acknowledgeAlarm(alarmSourceOptionalss.get(), alarmAck);
			} else {
				value = alarmResolver.acknowledgeAlarm((BIAlarmSource) point, alarmAck);
			}
			if (value) {
				return true;
			}
		}
	} else {
		value = alarmResolver.acknowledgeAlarm((BIAlarmSource) point, alarmAck);
	}
	return value;
  }

  private static final boolean VALID_INDICES_IN_DB = true;
  private static final boolean INVALID_INDICES_IN_DB = false;

	public String getTimeOfQueryAlarm() {
		return timeOfQueryAlarm;
	}

	public void setTimeOfQueryAlarm(String timeOfQueryAlarm) {
		this.timeOfQueryAlarm = timeOfQueryAlarm;
	}

	public String getTimeAfterCheckingIndicesValid() {
		return timeAfterCheckingIndicesValid;
	}

	public void setTimeAfterCheckingIndicesValid(String timeAfterCheckingIndicesValid) {
		this.timeAfterCheckingIndicesValid = timeAfterCheckingIndicesValid;
	}

	public String getTimeAfterCollectingALarms() {
		return timeAfterCollectingALarms;
	}

	public void setTimeAfterCollectingALarms(String timeAfterCollectingALarms) {
		this.timeAfterCollectingALarms = timeAfterCollectingALarms;
	}

	private String timeOfQueryAlarm = "";
	private String timeAfterCheckingIndicesValid = "";
	private String timeAfterCollectingALarms = "";
  

}
