/*
 * Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
 * This file contains trade secrets of Honeywell International, Inc. No part
 * may be reproduced or transmitted in any form by any means or for any
 * purpose without the express written permission of Honeywell.
 */
package com.honeywell.honplantcontroller.hmi.comm;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.security.SecureRandom;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.interfaces.ECPublicKey;
import java.text.MessageFormat;
import java.util.List;
import java.util.Optional;
import java.util.Timer;
import java.util.logging.Level;

import javax.baja.bacnet.BBacnetNetwork;
import javax.baja.bacnet.BacnetException;
import javax.baja.bacnet.datatypes.BBacnetAddress;
import javax.baja.bacnet.datatypes.BBacnetObjectIdentifier;
import javax.baja.bacnet.datatypes.BBacnetUnsigned;
import javax.baja.bacnet.datatypes.BErrorType;
import javax.baja.bacnet.enums.BBacnetErrorClass;
import javax.baja.bacnet.enums.BBacnetErrorCode;
import javax.baja.bacnet.enums.BBacnetObjectType;
import javax.baja.bacnet.enums.BCharacterSetEncoding;
import javax.baja.bacnet.io.BBacnetComm;
import javax.baja.bacnet.io.ErrorException;
import javax.baja.bacnet.io.RejectException;
import javax.baja.naming.SlotPath;
import javax.baja.nre.annotations.NiagaraType;
import javax.baja.nre.util.ByteBuffer;
import javax.baja.sys.BAbsTime;
import javax.baja.sys.BComponent;
import javax.baja.sys.BDynamicEnum;
import javax.baja.sys.BEnumRange;
import javax.baja.sys.BFacets;
import javax.baja.sys.BString;
import javax.baja.sys.BValue;
import javax.baja.sys.Sys;
import javax.baja.sys.Type;
import javax.baja.user.BUser;

import com.honeywell.honbacnethelper.listener.HonPrivateTransferListener;
import com.honeywell.honplantcontroller.hmi.alarm.IAlarmResolver;
import com.honeywell.honplantcontroller.hmi.messages.QueryAlarmPointsMessage;
import com.honeywell.honplantcontroller.hmi.messages.QueryAlarmPointsResultMessage;
import com.honeywell.honplantcontroller.hmi.messages.QueryUnackAlarmPointsResultMessage;
import com.honeywell.honplantcontroller.hmi.messages.QueryUserLoginMessage;
import com.honeywell.honplantcontroller.hmi.messages.QueryUserLogoutMessage;
import com.honeywell.honplantcontroller.hmi.messages.QueryUserMessage;
import com.honeywell.honplantcontroller.hmi.messages.RequestAlarmAckMessage;
import com.honeywell.honplantcontroller.hmi.messages.RequestCertAckMessage;
import com.honeywell.honplantcontroller.hmi.messages.RequestCertMessage;
import com.honeywell.honplantcontroller.hmi.service.BHonPlantControllerHMI;
import com.honeywell.honplantcontroller.hmi.user.authenticator.BHonPlantControllerHMIAuthenticator;
import com.honeywell.honplantcontroller.hmi.utils.BHonPlantControllerHMIInitializationHandler;
import com.honeywell.honplantcontroller.hmi.utils.HMIConfirmedPrivateTransferRequest;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerBacnetUtils;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerHMILogger;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerUserUtils;
import com.tridium.bacnet.asn.AsnInputStream;
import com.tridium.bacnet.asn.AsnOutputStream;
import com.tridium.bacnet.asn.AsnUtil;
import com.tridium.bacnet.asn.NErrorType;
import com.tridium.bacnet.services.BacnetComplexAck;
import com.tridium.bacnet.services.BacnetServicePrimitive;
import com.tridium.bacnet.services.BacnetSimpleAck;
import com.tridium.bacnet.services.confirmed.ConfirmedPrivateTransferAck;
import com.tridium.bacnet.services.error.ConfirmedPrivateTransferError;
import com.tridium.bacnet.stack.BBacnetStack;
import com.tridium.bacnet.stack.BacnetStackException;
import com.tridium.bacnet.stack.client.BBacnetClientLayer;
import com.tridium.bacnet.stack.network.BNetworkPort;
import com.tridium.bacnet.stack.network.BNetworkPriority;
import com.tridium.bacnet.stack.server.BBacnetExportTable;


@NiagaraType
public class BHMICommHandler extends BComponent implements IUnconfirmedPrivateTransferSender, IConfirmedPrivateTransferSender {
/*+ ------------ BEGIN BAJA AUTO GENERATED CODE ------------ +*/
/*@ $com.honeywell.honplantcontroller.hmi.comm.BHMICommHandler(2979906276)1.0$ @*/
/* Generated Tue Mar 15 15:46:46 CET 2022 by Slot-o-Matic (c) Tridium, Inc. 2012 */

////////////////////////////////////////////////////////////////
// Type
////////////////////////////////////////////////////////////////

  @Override
  public Type getType() { return TYPE; }
  public static final Type TYPE = Sys.loadType(BHMICommHandler.class);

/*+ ------------ END BAJA AUTO GENERATED CODE -------------- +*/

    public BHMICommHandler(BBacnetNetwork bacnetNetwork, BHonPlantControllerHMI hmiService) {
        this.setBacnetNetwork(bacnetNetwork);
        this.hmiService = hmiService;
    }

    public QueryAlarmPointsResultMessage handleQueryAlarmPointsMessage(QueryAlarmPointsMessage msg) throws BacnetException {
        BBacnetExportTable exportTable = (BBacnetExportTable) getBacnetNetwork().getLocalDevice().getExportTable();

        setTimeBeforeQueryAlarms(MessageFormat.format("Time before calling queryAlarms in handleQueryAlarmPointsMessage is {0}", BAbsTime.now().getMillis()));
        BHMIAlarmHandler hmiAlarmHandler = hmiService.getHmiAlarmHandler();
        QueryAlarmPointsResultMessage result = hmiAlarmHandler.queryAlarms(msg.getHmiAlarmLevel(), msg.getStartIndex(), msg.getStopIndex(), getAlarmResolver(), exportTable);
        setTimeAfterQueryAlarms(MessageFormat.format("Time After calling queryAlarms in handleQueryAlarmPointsMessage is {0}", BAbsTime.now().getMillis()));
        return result;
    }

    public QueryUnackAlarmPointsResultMessage handleQueryUnackAlarmPointsMessage(QueryAlarmPointsMessage msg) throws BacnetException {
        BBacnetExportTable exportTable = (BBacnetExportTable) getBacnetNetwork().getLocalDevice().getExportTable();
        setTimeBeforeQueryAlarms(MessageFormat.format("Time before calling queryUnackAlarms in handleQueryUnackAlarmPointsMessage is {0}", BAbsTime.now().getMillis()));
        BHMIAlarmHandler hmiAlarmHandler = hmiService.getHmiAlarmHandler();
        QueryUnackAlarmPointsResultMessage result =  hmiAlarmHandler.queryUnAckAlarms(msg.getHmiAlarmLevel(), msg.getStartIndex(), msg.getStopIndex(), getAlarmResolver(), exportTable);
        setTimeAfterQueryAlarms(MessageFormat.format("Time After calling queryUnackAlarms in handleQueryUnackAlarmPointsMessage is {0}", BAbsTime.now().getMillis()));
        return result;
    }


	public void handleServiceDeviceInfo(byte[] bytes) throws BacnetException {
		AsnInputStream inputStream = AsnInputStream.make(bytes);
		BBacnetObjectIdentifier hmiInstance = inputStream.readObjectIdentifier();
		hmiService.setSerialNumber(new String(inputStream.readOctetString()));
		HonPlantControllerHMILogger.finest(MessageFormat.format("{0}.HMI Instance from HMI is {1}; Serial Number of HMI is {2} ", LOG_CLS_PREFIX,
				hmiInstance, hmiService.getSerialNumber()));
		BHonPlantControllerBacnetProperties hmiBacnetProperties = hmiService.getBacnetProperties();
		if (hmiBacnetProperties.getHMIInstanceId().getInstanceNumber() != hmiInstance.getInstanceNumber()) {
			hmiService.setInstanceIdFromHMi(hmiInstance);
			hmiBacnetProperties.validateAndThrowExceptionIfinstanceIdIsDuplicate(hmiBacnetProperties.getHMIInstanceId().getInstanceNumber());
			sendInstanceidToHMI(hmiBacnetProperties.getHMIInstanceId(), hmiService, true);
		} else {
			hmiService.addHMIEntryInBacnetNetwork();
			if (null != getHmiIAmListener()) {
				getHmiIAmListener().threadVerifyFirmware();
			}
		}
	}

	public byte[] handleUserLogin(QueryUserLoginMessage msg) throws BacnetException {
        HonPlantControllerHMILogger.finest(MessageFormat.format("{0}.handleUserLogin() userName: {1}; escaped userName {2} ", LOG_CLS_PREFIX, msg.getUserName(), SlotPath.escape(msg.getUserName())));
        Optional<BUser> foundUser = HonPlantControllerUserUtils.getNiagaraUserbyName(SlotPath.escape(msg.getUserName()));
        // get user out of optional or throws Exception.
        BUser user = foundUser.orElseThrow(() ->
            new ErrorException(new BErrorType(BBacnetErrorClass.SERVICES, BBacnetErrorCode.AUTHENTICATION_FAILED)));

        byte[] decryptedDk = new byte[32];
        byte[] decryptedRand = new byte[32];
        try {
            HonPlantControllerHMILogger.finest(MessageFormat.format("{0}.handleUserLogin(): user: {1} ", LOG_CLS_PREFIX, user.getName()));
            hmiService.getHmiInitializationListener().UserLoginDecryptDkRand(msg.getUserDk(), decryptedDk, decryptedRand);

            ///CLOVER:OFF
            BValue authProp = user.getChildren(BHonPlantControllerHMIAuthenticator.class)[0];
            if (authProp instanceof BHonPlantControllerHMIAuthenticator) {
                BHonPlantControllerHMIAuthenticator hmiAuthenticator = (BHonPlantControllerHMIAuthenticator) authProp;

                if (!hmiAuthenticator.validateDk(decryptedDk)) {
                    HonPlantControllerHMILogger.info(MessageFormat.format("{0}.handleUserLogin(): Pin is invalid", LOG_CLS_PREFIX));
                    throw new ErrorException(new BErrorType(BBacnetErrorClass.SERVICES, BBacnetErrorCode.AUTHENTICATION_FAILED));
                }

                HonPlantControllerHMILogger.info(MessageFormat.format("{0}.handleUserLogin(): Pin is valid", LOG_CLS_PREFIX));
                return new byte[0];
            }
        } catch (Exception e) {
            HonPlantControllerHMILogger.info(MessageFormat.format("{0}.handleUserLogin(): Pin is invalid", LOG_CLS_PREFIX));
        }

        //throw an exception if we come here!
        throw new ErrorException(new BErrorType(BBacnetErrorClass.SERVICES, BBacnetErrorCode.AUTHENTICATION_FAILED));
    }

    public byte[] handleChangePassword(QueryUserMessage msg) throws BacnetException {
        HonPlantControllerHMILogger.finest(MessageFormat.format("{0}.handleChangePassword()", LOG_CLS_PREFIX));
        final int ERR_DUPLICATE_PWD = 4;
        Optional<BUser> foundUser = HonPlantControllerUserUtils.getNiagaraUserbyName(SlotPath.escape(msg.getUserName()));

        // get user out of optional or throws Exception.
        BUser user = foundUser.orElseThrow(() ->
            new ErrorException(
                new BErrorType(BBacnetErrorClass.SERVICES,
                BBacnetErrorCode.INCONSISTENT_PARAMETERS)
            )
        );

        byte[] decryptedDk = new byte[32];
        byte[] decryptedRand = new byte[32];
        byte[] decryptedSalt = new byte[32];
        try {
            HonPlantControllerHMILogger.finest(MessageFormat.format("{0}.handleChangePassword(): user: {1} ", LOG_CLS_PREFIX, user.getName()));
            hmiService.getHmiInitializationListener().UserLoginDecryptSaltDkRand(msg.getData(), decryptedSalt, decryptedDk, decryptedRand);

            // This code will be rewritten because of certificate limitations
            ///CLOVER:OFF

            BValue authProp = user.getChildren(BHonPlantControllerHMIAuthenticator.class)[0];
            if (authProp instanceof BHonPlantControllerHMIAuthenticator) {
                BHonPlantControllerHMIAuthenticator hmiAuthenticator = (BHonPlantControllerHMIAuthenticator) authProp;

                int iterationCount = 2; //Fixed
                if (hmiAuthenticator.updateDk(decryptedSalt, iterationCount, decryptedDk))
                    return new byte[0];

                //duplicate password
                throw new RejectException(ERR_DUPLICATE_PWD);
            }
        } catch (Exception e) {
            HonPlantControllerHMILogger.info(
                    MessageFormat.format("{0}.handleChangePassword(): Pin is invalid", LOG_CLS_PREFIX)
                );
        }

        //throw an exception if we come here!
        throw new ErrorException(new BErrorType(BBacnetErrorClass.SERVICES, BBacnetErrorCode.INCONSISTENT_PARAMETERS));
        ///CLOVER:OFF
    }

    public byte[] handleUserLogout(QueryUserLogoutMessage msg)  throws BacnetException  {

        HonPlantControllerHMILogger.finest(MessageFormat.format("{0}.handleUserLogout():", LOG_CLS_PREFIX));

        if (null == msg) {
            throw new ErrorException(
                new BErrorType(BBacnetErrorClass.SERVICES, BBacnetErrorCode.CONFIGURATION_IN_PROGRESS)
            );
        }

//        Optional<BUser> foundUser = HonPlantControllerUserUtils.getNiagaraUserbyName(SlotPath.escape(msg.getUserName()));
//
//        if (!foundUser.isPresent()) {
//            throw new ErrorException(
//                new BErrorType(BBacnetErrorClass.SERVICES, BBacnetErrorCode.CONFIGURATION_IN_PROGRESS)
//            );
//        }


        return new byte[0];
    }

    public byte[] handleReadUserRequest(int index) {
        
        HonPlantControllerHMILogger.finest(
            MessageFormat.format("{0}.handleReadUserRequest(): Handle read users", LOG_CLS_PREFIX)
        );
        
        // make sure that index cannot be less than 0 because that would break the calculation.
        if (index < 0) {
            HonPlantControllerHMILogger.finest(
                MessageFormat.format("{0}.handleReadUserRequest(): could not select users because received index is < 0!", LOG_CLS_PREFIX)
            );
            return new byte[0]; // TODO find better return in case of an error.
        }

        List<String> users = HonPlantControllerUserUtils.getUnescapedNiagaraUserList();
        
        // calculate start and end index:
        //   startIndex = index * 5
        //   endIndex   = startIndex + 4
        // | Req. Idx. | start | end |
        // |-----------+-------+-----|
        // | 0         |  0    |  4  |
        // | 1         |  5    |  9  |
        // | 2         | 10    | 14  |
        // | 3         | 15    | 19  |
        // +-----------+-------+-----+
        
        final int stepSize = 5;
        
        int startIndex = index * stepSize;
        // need to use 5 because the toIndex of List<T>::sublist(int fromIndex, int toIndex) is exclusive
        int endIndex = startIndex + stepSize;
        
        if (endIndex > users.size()) {
            endIndex = users.size();
        }
        
        if (startIndex > endIndex) {
            return new byte[0];
        }

        if (startIndex == endIndex) {
            return new byte[0];
        }

        if (startIndex >= users.size()) {
            return new byte[0];
        }
        
        List<String> selectedUsers = users.subList(startIndex, endIndex);
        
        AsnOutputStream userList = new AsnOutputStream();
        for (String user : selectedUsers) {
            userList.writeCharacterString(BString.make(user), BCharacterSetEncoding.iso10646_UTF8);
        }

        return userList.toByteArray();
    }

    // handleRequestRand will be rewritten due to certificate limitations
    ///CLOVER:OFF
    public byte[] handleRequestRand() throws Exception {
        //byte[] rand = mmiService.getMmiInitializationListener().generateRequestRand();
        byte[] rand = new byte[32];
        new SecureRandom().nextBytes(rand);
        AsnOutputStream asnOut = new AsnOutputStream();
        asnOut.writeOctetString(rand);
        return asnOut.toByteArray();
    }
    ///CLOVER:ON

    // handleRequestSalt will be rewritten due to certificate limitations
    ///CLOVER:OFF
    public byte[] handleRequestSalt(QueryUserLoginMessage msg) throws Exception {
        byte[] encryptedSalt = new byte[64];
        Optional<BUser> foundUser = HonPlantControllerUserUtils.getNiagaraUserbyName(SlotPath.escape(msg.getUserName()));

        // get user out of optional or throws Exception.
        BUser user = foundUser.orElseThrow(() ->
            new ErrorException(
                new BErrorType(BBacnetErrorClass.SERVICES,
                    BBacnetErrorCode.INCONSISTENT_PARAMETERS)
            )
        );

        try {
        	BValue authProp = user.getChildren(BHonPlantControllerHMIAuthenticator.class)[0];
        	if (authProp instanceof BHonPlantControllerHMIAuthenticator) {
                BHonPlantControllerHMIAuthenticator hmiAuthenticator = (BHonPlantControllerHMIAuthenticator) authProp;

                //byte[] dk = hmiAuthenticator.getKey();
                //byte[] salt = hmiAuthenticator.getSalt();
                //int ic = hmiAuthenticator.getIterationCount();

                String dk = hmiAuthenticator.getDk();
                String[] tokens = dk.split(":");
                String salt = tokens[0];
                int ic = Integer.parseInt(tokens[1]);
                String key = tokens[2];

                //include 32 random bytes
                byte[] rand = new byte[32];
                new SecureRandom().nextBytes(rand);
                //encrypt everything using sharedKey
                encryptedSalt = hmiService.getHmiInitializationListener().GenerateEncryptedSalt(hmiAuthenticator.hexStringToBytes(salt), rand);

            	}
        } catch (Exception e) {
        	HonPlantControllerHMILogger.finest(e.getMessage(), e);
        }

        AsnOutputStream asnOut = new AsnOutputStream();
        asnOut.writeOctetString(encryptedSalt);
        return asnOut.toByteArray();
    }

    ///CLOVER:ON
    public void handleCertResponse (byte[] response) throws Exception {
        final String LOG_MSG_PREFIX = BHonPlantControllerHMIInitializationHandler.TYPE.getTypeName() + "." + getClass().getSimpleName() + ".handleCertResponse";
        CertificateFactory certFactory = CertificateFactory.getInstance("X.509", "BC");
        InputStream in = new ByteArrayInputStream(response);
        X509Certificate hmiCert = (X509Certificate)certFactory.generateCertificate(in);
        BDynamicEnum certificateValue = hmiService.getDeviceCertificate();
		BFacets enumFacet = certificateValue.getEnumFacets();
		BEnumRange enumRange = (BEnumRange) enumFacet.get(BFacets.RANGE);
		String certAlias = SlotPath.unescape(enumRange.getTag(HONEYWELL_CA_CERT_ORIDINAL));
        X509Certificate caHoneywell = hmiService.getCertFromSystemTrustStore(certAlias);
        try {
            hmiCert.verify(caHoneywell.getPublicKey());
            hmiService.getHmiInitializationListener().setCertVerified(true);
            hmiService.getHmiInitializationListener().setPubKeyHMI((ECPublicKey) hmiCert.getPublicKey());
            HonPlantControllerHMILogger.finest(MessageFormat.format("{0} Certificate received and issued by {1} is verified", LOG_MSG_PREFIX, hmiCert.getIssuerDN().getName()));
        } catch (Exception e) {
            hmiService.getHmiInitializationListener().setCertVerified(false);
            HonPlantControllerHMILogger.finest(MessageFormat.format("{0} Certificate received and issued {1} cannot be verified", LOG_MSG_PREFIX, hmiCert.getIssuerDN().getName()));
        }
    }

    ///CLOVER:OFF
    public void sendDynamicDataSyncRequest(Integer category) throws BacnetException {

        AsnOutputStream out = new AsnOutputStream();
        
        // As per new 'dynamic data synchronization' design it is not necessary to distinguish between Users, Datapoints
        // and Schedules. It is enough to send 0x01 to the HMI to restart. The other options are optional:
        // | Value | Meaning                                   |
        // +-------+-------------------------------------------+
        // | 0x01  | There is data changed from the controller |
        // | 0x02  | Datapoint is changed [optional]           |
        // | 0x04  | Time schedule is changed [optional]       |
        // | 0x08  | User is changed [optional]                |
        // +-------+-------------------------------------------+
        // * original text from the spec.
        out.writeUnsigned(BBacnetUnsigned.make(1));
        sendConfirmedPrivateTransferMessage(HMIPrivateTransferListener.SERVICE_ID_DYNAMIC_DATA_SYNC, out.toByteArray());
    }
    ///CLOVER:ON


    public byte[] handleRequestLanguageFile(byte[] request)
      throws BacnetException
    {
        final String LOG_MSG_PREFIX = TYPE.getTypeName() + ".handleRequestLanguageFile";

        String reqLang;
        AsnInputStream asnInputStream = new AsnInputStream(request);
        try
        {
            byte[] lang = asnInputStream.readOctetString();
            reqLang = new String(lang, StandardCharsets.UTF_8);
            HonPlantControllerHMILogger.info(MessageFormat.format("{0} REQUEST_LANGUAGE_FILE lang[{1}]",
                                                                  LOG_MSG_PREFIX, reqLang));
        }
        catch (Exception exc)
        {
            HonPlantControllerHMILogger.warning(MessageFormat.format("{0} REQUEST_LANGUAGE_FILE lang exc:",
                                                                     LOG_MSG_PREFIX), exc);
            reqLang = null;
        }

        if (hmiService.sendLanguageFile(reqLang))
        {
            return new byte[0];
        }

        // language is not supported
        throw new ErrorException(new BErrorType(BBacnetErrorClass.SERVICES,
                                                BBacnetErrorCode.INCONSISTENT_PARAMETERS));
    }

    public byte[] handleCertRequestFromHMI(RequestCertMessage msg) throws CertificateEncodingException {
        int index = msg.getIndex();

        byte [] certFull = hmiService.getCertBytes();
        int packageAmount = certFull.length / MAX_PACKAGE_SIZE;
        if (certFull.length % MAX_PACKAGE_SIZE != 0) {
            packageAmount++;
        }

        if (packageAmount > 1 && (index < (packageAmount - 1))) {
            lastPackagageSent = false;
            byte[] chunk = new byte[MAX_PACKAGE_SIZE];
            System.arraycopy(certFull, index * MAX_PACKAGE_SIZE, chunk, 0, MAX_PACKAGE_SIZE);
            RequestCertAckMessage requestCertAckMessage = new RequestCertAckMessage(index, lastPackagageSent, chunk);


            return requestCertAckMessage.getBytes();
        } else {
            lastPackagageSent = true;
            int remainingSize = certFull.length - (index * MAX_PACKAGE_SIZE);
            byte [] chunk = new byte[remainingSize];
            System.arraycopy(certFull, (index*MAX_PACKAGE_SIZE), chunk, 0, remainingSize);
            RequestCertAckMessage requestCertAckMessage = new RequestCertAckMessage(index, lastPackagageSent, chunk);


            return requestCertAckMessage.getBytes();
        }
    }
    
    public byte[] handleAlarmAck(RequestAlarmAckMessage alarmAck) {

        BBacnetExportTable exportTable = (BBacnetExportTable) getBacnetNetwork().getLocalDevice().getExportTable();

        BHMIAlarmHandler hmiAlarmHandler = hmiService.getHmiAlarmHandler();

        boolean success = hmiAlarmHandler.proccessAlarmAckRequest(alarmAck, exportTable, getAlarmResolver());

        if(success) {
            return new BacnetSimpleAck(0).toEncodedBytes();
        }

        return new ConfirmedPrivateTransferError(
            new NErrorType(BBacnetErrorClass.OBJECT,BBacnetErrorCode.UNKNOWN_OBJECT),
            HMIPrivateTransferListener.VENDOR_ID_HON, 
            HMIPrivateTransferListener.SERVICE_ID_ALARM_ACK).toEncodedBytes();
    }

    public void registerListeners()
    {
        HonPlantControllerHMILogger.finest(MessageFormat.format("{0}.registerListeners(): try register Listeners", LOG_CLS_PREFIX));
        if (listenersAreRegistered) {
            return;
        }
        if (getBacnetNetwork() == null) {
            HonPlantControllerHMILogger.finest(MessageFormat.format("{0}.registerListeners(): BacnetNetwork is not available!", LOG_CLS_PREFIX));
            return;
        }
        if (privateTransferListener == null) {
            HonPlantControllerHMILogger.finest(MessageFormat.format("{0}.registerListeners(): register Private Transfer Listener!", LOG_CLS_PREFIX));
            privateTransferListener = new HMIPrivateTransferListener(this);
        }
        if (iAmListener == null) {
            HonPlantControllerHMILogger.finest(MessageFormat.format("{0}.registerListeners(): register I Am Listener!", LOG_CLS_PREFIX));
            iAmListener = new HMIIAmListener(hmiService);
        }

        BBacnetComm bacnetComm = getBacnetNetwork().getBacnetComm();
        if (null == bacnetComm) {
            return;
        }
        HonPrivateTransferListener.getInstance().registerListener(bacnetNetwork, privateTransferListener, CONFIRMED_PRIVATE_TRANSFER_ID);
        HonPrivateTransferListener.getInstance().registerListener(bacnetNetwork, privateTransferListener, UNCONFIRMED_PRIVATE_TRANSFER_ID);
        bacnetComm.registerBacnetListener(iAmListener, SERVICE_ID_IAM);
        listenersAreRegistered = true;
    }

    public void unregisterListeners()
    {
    	try {
	        BBacnetComm bacnetComm = getBacnetNetwork().getBacnetComm();
	        if (null == bacnetComm) {
	            return;
	        }
	
	        if (privateTransferListener != null && listenersAreRegistered) {
	        	// Unregister from Common Honeywell Listener.
	        	HonPrivateTransferListener.getInstance().unregisterListener(privateTransferListener, CONFIRMED_PRIVATE_TRANSFER_ID);
	        	HonPrivateTransferListener.getInstance().unregisterListener(privateTransferListener, UNCONFIRMED_PRIVATE_TRANSFER_ID);
	            privateTransferListener = null;
	        }
	        if (iAmListener != null && listenersAreRegistered) {
	            bacnetComm.unregisterBacnetListener(iAmListener, SERVICE_ID_IAM);
	            iAmListener = null;
	        }
    	}catch (Exception e) {
    		HonPlantControllerHMILogger.finest(e.getMessage());
		}
        privateTransferListener = null;
        iAmListener = null;
        listenersAreRegistered = false;
    }

    @Override
    public void sendUnconfirmedPrivateTransferMessage(int serviceID, byte[] rawMsg) {
        try {
            if (rawMsg.length > 0) {
                getBacnetNetwork().getBacnetComm().unconfirmedPrivateTransfer(getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
                HonPlantControllerHMILogger.checkAndLog(Level.FINE, "[BHonPlantControllerHMIPointModeHandler.notifyHMI] hmi notified");
            }
        } catch (Exception e) {
            HonPlantControllerHMILogger.checkAndLog(Level.FINEST,"[BHonPlantControllerHMIPointModeHandler.pointModeChanged] could not notify HMI of point mode change ", e);
        }
    }

    @Override
    public void sendConfirmedPrivateTransferMessage(int serviceID, byte[] rawMsg) throws BacnetException{
        if (serviceID == HMIPrivateTransferListener.SERVICE_ID_DYNAMIC_DATA_SYNC) {
            try {
                if (rawMsg.length > 0) {
                    getBacnetNetwork().getBacnetComm().confirmedPrivateTransfer(getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
                    HonPlantControllerHMILogger.checkAndLog(Level.FINE, "[BHonPlantControllerHMIPointModeHandler.sendConfirmedPrivateTransferMessage] Confirmed Private Transfer with Service ID: "+ serviceID);
                }
            } catch (Exception e) {
                HonPlantControllerHMILogger.checkAndLog(Level.FINEST,"[BHonPlantControllerHMIPointModeHandler.sendConfirmedPrivateTransferMessage] Error in Confirmed Private Transfer with Service ID: "+ serviceID+", reason: "+e.getMessage());
            }
        }else if (serviceID == BHonPlantControllerHMIInitializationHandler.SERVICE_ID_REQUEST_CERT_EXCHANGE) {
            try {
                //initialize variables
                verifyDeviceCertSend = false;
                lastPackagageSent = false;
                HonPlantControllerHMILogger.finest("[BHonPlantControllerHMIPointModeHandler.sendConfirmedPrivateTransferMessage] Confirmed Private Transfer with Service ID: "+ serviceID);
                ByteBuffer dataBuffer = new ByteBuffer();
                int expectedIndex = 0;

                byte [] response = confirmedPrivateTransfer((BBacnetStack)getBacnetNetwork().getBacnetComm(), getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
                RequestCertAckMessage requestCertAckMessage = getRequestCertAckMessageWithRetry(serviceID, expectedIndex, new RequestCertAckMessage(response));

                if (requestCertAckMessage != null) {
                    dataBuffer.write(requestCertAckMessage.getCertBytes());

                    while (!requestCertAckMessage.getIsLastFrame()) {
                        expectedIndex++;
                        rawMsg[0] = (byte)(expectedIndex | FLAG_NEW_INITIALIZATION);
                        response = confirmedPrivateTransfer((BBacnetStack)getBacnetNetwork().getBacnetComm(), getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
                        requestCertAckMessage = getRequestCertAckMessageWithRetry(serviceID, expectedIndex, new RequestCertAckMessage(response));
                        dataBuffer.write(requestCertAckMessage.getCertBytes());
                    }
                    handleCertResponse(dataBuffer.toByteArray());
                    HonPlantControllerHMILogger.finest("[BHonPlantControllerHMIPointModeHandler.sendConfirmedPrivateTransferMessage] Confirmed Private Transfer: ");
                } else {
                    throw new Exception();
                }
            } catch (Exception e) {
                HonPlantControllerHMILogger.finest("[BHonPlantControllerHMIPointModeHandler.sendConfirmedPrivateTransferMessage] Error in Confirmed Private Transfer with Service ID: "+ serviceID+", reason: "+e.getMessage());
            }
        } else if (serviceID == BHonPlantControllerHMIInitializationHandler.SERVICE_ID_VERIFY_DEVICE_CERT) {
            try {
                HonPlantControllerHMILogger.finest("[BHonPlantControllerHMIPointModeHandler.sendConfirmedPrivateTransferMessage] Confirmed Private Transfer with Service ID: "+ serviceID);
                getBacnetNetwork().getBacnetComm().confirmedPrivateTransfer(getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
                hmiService.getHmiInitializationListener().startECDH();
            } catch (Exception e) {
                HonPlantControllerHMILogger.finest("[BHonPlantControllerHMIPointModeHandler.sendConfirmedPrivateTransferMessage] Error in Confirmed Private Transfer with Service ID: "+ serviceID+", reason: "+e.getLocalizedMessage());

            }

        } else if (serviceID == BHonPlantControllerHMIInitializationHandler.SERVICE_ID_START_ECDH) {
            try {
                HonPlantControllerHMILogger.finest("[BHonPlantControllerHMIPointModeHandler.sendConfirmedPrivateTransferMessage] Confirmed Private Transfer with Service ID: "+ serviceID);
                getBacnetNetwork().getBacnetComm().confirmedPrivateTransfer(getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
                hmiService.getHmiInitializationListener().sendVerifyMstSecret();
            } catch (Exception e) {
        	// confirmedPrivateTransfer expects complex ack but for notification controller receives simple ack but to which exception will be thrown from Niagara  BACnet stack
            HonPlantControllerHMILogger.checkAndLog(Level.FINE,"[BHonPlantControllerHMIPointModeHandler.pointModeChanged] could not notify HMI of point mode change ", e);
            }
        } else if (serviceID == BHonPlantControllerHMIInitializationHandler.SERVICE_ID_VERIFY_MST_SECRET) {
            byte [] response = getBacnetNetwork().getBacnetComm().confirmedPrivateTransfer(getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
            try {
                if (hmiService.getHmiInitializationListener().handleVerifyMstSecretAck(rawMsg, response)) {
                    hmiService.getHmiInitializationListener().sendChangeCipherMstSecret();
                } else {
                    //TODO failure mechanism
                    //mmiService.getMmiInitializationListener().getKeyExchangeState().fail();
                    //mmiService.doStartKeyExchange();
                	HonPlantControllerHMILogger.finest("[BHonPlantControllerHMIPointModeHandler.sendConfirmedPrivateTransferMessage] Could not verify MstSecret");
                }
            } catch (Exception e) {
            	HonPlantControllerHMILogger.finest(e.getMessage(), e);
            }
        } else if (serviceID == BHonPlantControllerHMIInitializationHandler.SERVICE_ID_CHANGE_CIPHER_MST_SECRET) {
            //we are not interested in response information
            getBacnetNetwork().getBacnetComm().confirmedPrivateTransfer(getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
            try {
                hmiService.getHmiInitializationListener().sendDeriveSharedKey();
            } catch (Exception e) {
            	HonPlantControllerHMILogger.finest(e.getMessage(), e);
            }

        } else if (serviceID == BHonPlantControllerHMIInitializationHandler.SERVICE_ID_DERIVE_SHARED_KEY) {
            byte [] response = getBacnetNetwork().getBacnetComm().confirmedPrivateTransfer(getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
            try {
                hmiService.getHmiInitializationListener().sendVerifySharedKey();
            } catch (Exception e) {
            	HonPlantControllerHMILogger.finest(e.getMessage(), e);
            }
        } else if (serviceID == BHonPlantControllerHMIInitializationHandler.SERVICE_ID_VERIFY_SHARED_KEY) {
            byte[] response = getBacnetNetwork().getBacnetComm().confirmedPrivateTransfer(getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
            try {
                if (hmiService.getHmiInitializationListener().handleVerifySharedKeyAck(response)) {
                    HonPlantControllerHMILogger.finest("[BHonPlantControllerHMIPointModeHandler.sendConfirmedPrivateTransferMessage] Verify Shared Key successful");
                    hmiService.getHmiInitializationListener().sendChangeCipherSharedKey();
                }
                else {
                    //mmiService.getMmiInitializationListener().getKeyExchangeState().fail();
                    //mmiService.doStartKeyExchange();
                	HonPlantControllerHMILogger.finest("[BHonPlantControllerHMIPointModeHandler.sendConfirmedPrivateTransferMessage] Could not verify Shared Key");
                }
            } catch (Exception e) {
            	HonPlantControllerHMILogger.finest(e.getMessage(), e);
            }
        } else if (serviceID == BHonPlantControllerHMIInitializationHandler.SERVICE_ID_CHANGE_CIPHER_SHARED_KEY) {
            //we are not interested in response information
            getBacnetNetwork().getBacnetComm().confirmedPrivateTransfer(getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
        }else if(serviceID == BHonPlantControllerHMIInitializationHandler.SERVICE_ID_SET_DEVICE_INSTANCE) {
        	getBacnetNetwork().getBacnetComm().confirmedPrivateTransfer(getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, rawMsg);
        }

    }

    private RequestCertAckMessage getRequestCertAckMessageWithRetry(int serviceID, int expectedIndex, RequestCertAckMessage requestCertAckMessage) throws Exception {
        byte[] response;

        for (int retryCount = 0; retryCount < MAX_TRIES; retryCount++){

            if (!(requestCertAckMessage.getIndex() == expectedIndex)) {
                HonPlantControllerHMILogger.finest("BHMICommHadler.getRequestCertAckMessageWithRetry expected index " + expectedIndex+ "does not match received index "+ requestCertAckMessage.getIndex());
                response = confirmedPrivateTransfer((BBacnetStack)getBacnetNetwork().getBacnetComm(), getHMIbacnetAddress(), HMIPrivateTransferListener.VENDOR_ID_HON, serviceID, new byte[]{(byte)expectedIndex});
                requestCertAckMessage = new RequestCertAckMessage(response);
                continue;
            }
            HonPlantControllerHMILogger.finest("BHMICommHadler.getRequestCertAckMessageWithRetry has valid index");

            return requestCertAckMessage;
        }
        return null;

    }


    // reimplementation of BBacnetClientLayer.confirmedPrivateTransfer
    public byte[] confirmedPrivateTransfer(BBacnetStack bacnetStack, BBacnetAddress deviceAddress, int vendorId, int serviceNumber, byte[] serviceParameters) throws BacnetException {
        if (!bacnetStack.isCommInitiationEnabled()) {
            throw new BacnetStackException("Stack Disabled");
        } else {

            HonPlantControllerHMILogger.fine("BHMICommHadler.confirmedPrivateTransfer(" + vendorId + ";" + serviceNumber + ")");

            HMIConfirmedPrivateTransferRequest hmiRequest = new HMIConfirmedPrivateTransferRequest((long)vendorId, (long)serviceNumber, serviceParameters);

            BacnetServicePrimitive response = (BacnetComplexAck)bacnetStack.getTransport().sendConfirmedRequestComplex(hmiRequest, deviceAddress, BNetworkPriority.normal);

            HonPlantControllerHMILogger.fine("BHMICommHadler.confirmedPrivateTransfer: Response = ");

            return ((ConfirmedPrivateTransferAck)response).getResultBlock();
        }
    }

    public boolean isCertFullyVerified() {
        HonPlantControllerHMILogger.finest("[BHMICommHandler.isCertFullyVerified] hmiService.getHmiInitializationListener().isCertVerified() = " + hmiService.getHmiInitializationListener().isCertVerified());
        HonPlantControllerHMILogger.finest("[BHMICommHandler.isCertFullyVerified] lastPackagageSent = " + lastPackagageSent);

        return (hmiService.getHmiInitializationListener().isCertVerified() && lastPackagageSent);
    }

    public void sendVerifyDeviceCert() {
        HonPlantControllerHMILogger.finest("[BHMICommHandler.sendVerifyDeviceCert]");

        if(!verifyDeviceCertSend) {
            HonPlantControllerHMILogger.finest("[BHMICommHandler.sendVerifyDeviceCert] waitTime = " + waitTime);
            HonPlantControllerHMILogger.finest("[BHMICommHandler.sendVerifyDeviceCert] Before scheduling (second)= " + BAbsTime.now().getSecond());

            Timer t = new java.util.Timer();
            t.scheduleAtFixedRate(
                    new java.util.TimerTask() {
                        int cnt = 0;
                        final int MAX_CNT = 10;

                        @Override
                        public void run() {
                            HonPlantControllerHMILogger.finest("[BHMICommHandler.sendVerifyDeviceCert] cnt= " + cnt);

                            try {
                                if (isCertFullyVerified()) {
                                    HonPlantControllerHMILogger.finest("[BHMICommHandler.sendVerifyDeviceCert] timer start (second)= " + BAbsTime.now().getSecond());

                                    sendConfirmedPrivateTransferMessage(BHonPlantControllerHMIInitializationHandler.SERVICE_ID_VERIFY_DEVICE_CERT, new byte[]{0});
                                    verifyDeviceCertSend = true;
                                    t.cancel();
                                }
                            } catch (BacnetException e) {
                                HonPlantControllerHMILogger.finest(e.getMessage(), e);
                            }
                            if (++cnt == MAX_CNT) {
                                t.cancel();
                            }
                        }
                    },
                waitTime, //delay of firt trial
                waitTime //dely between consecutive trials
            );
        }

    }
    
	public void sendInstanceidToHMI(BBacnetObjectIdentifier proposedHMIInst, BHonPlantControllerHMI hmiDevice, boolean isFwToBeVerified) {
		new Thread(() -> {
			try {
				hmiDevice.getHmiInitializationListener().getConfirmedPrivateTransferSender().sendConfirmedPrivateTransferMessage(
						BHonPlantControllerHMIInitializationHandler.SERVICE_ID_SET_DEVICE_INSTANCE, getServiceParamBytes(proposedHMIInst, hmiDevice));
				hmiDevice.setInstanceIdFromHMi(proposedHMIInst);
				hmiService.getBacnetProperties().setHmiInstanceChanged(false);
				if (isFwToBeVerified && null != getHmiIAmListener()) {
						getHmiIAmListener().threadVerifyFirmware();
				}
			} catch (BacnetException e) {
				HonPlantControllerHMILogger.checkAndLog(Level.SEVERE, e.getMessage());
			}
		}).start();
	}
	
	public static byte[] getServiceParamBytes(BBacnetObjectIdentifier hmiInstanceId, BHonPlantControllerHMI hmiDevice) {
		int instanceNo = ((BBacnetObjectType.DEVICE << LEFT_SHIFT_BACNET_DEVICE_ID_INT) | hmiInstanceId.getInstanceNumber());
		String serialNumber = hmiDevice.getSerialNumber();
		AsnOutputStream out = new AsnOutputStream();
		out.writeObjectIdentifier(hmiInstanceId.getObjectType(), instanceNo);
		if (null != serialNumber) {
			out.writeOctetString(serialNumber.getBytes());
		}
		return out.toByteArray();
	}

    public BBacnetAddress getHMIbacnetAddress() {

        byte[] hmiMacAddress = iAmListener.getHmiMacAddress();
        BNetworkPort mstpPort =  HonPlantControllerBacnetUtils.getHMIPort(getBacnetNetwork(), hmiService.getBacnetProperties());

        if(mstpPort != null) {

            if(null == hmiMacAddress || hmiMacAddress.length <= 0) {
            	return new BBacnetAddress(mstpPort.getNetworkNumber(), new byte[] {0x01});	
            }

            return new BBacnetAddress(mstpPort.getNetworkNumber(), hmiMacAddress);
        }
        
        return null;
    }

	public static BValue readProperty(BBacnetClientLayer clientLayer, BBacnetObjectIdentifier objId,
			BBacnetAddress hmiAddress, int propertyId, int asnType, BValue obj) throws BacnetException {
		byte[] encodedValue = clientLayer.readProperty(hmiAddress, objId, propertyId);
		return AsnUtil.fromAsn(asnType, encodedValue, obj);
	}

	public static void writeProperty(BBacnetClientLayer clientLayer, BBacnetObjectIdentifier objId,
			BBacnetAddress hmiAddress, int propertyId, int asnType, BValue value) throws BacnetException {
		byte[] encodedValue = AsnUtil.toAsn(asnType, value);
		clientLayer.writeProperty(hmiAddress, objId, propertyId, encodedValue);
	}

    public HMIPrivateTransferListener getPrivateTransferListener(){
    	return privateTransferListener;
    }
    
    public HMIIAmListener getHmiIAmListener(){
    	return iAmListener;
    }

    public BBacnetNetwork getBacnetNetwork() {
		return bacnetNetwork;
	}
	public void setBacnetNetwork(BBacnetNetwork bacnetNetwork) {
		this.bacnetNetwork = bacnetNetwork;
	}

    protected IAlarmResolver getAlarmResolver() {
        if(null == alarmResolver) {
            alarmResolver =  HonPlantControllerBacnetUtils.getHMIAlarmConsoleRecipient(Sys.getStation());
        }

        return alarmResolver;
    }
    
	
	public BHonPlantControllerHMI getHmiService() {
		return hmiService;	
	}

	private boolean lastPackagageSent = false;
    private boolean verifyDeviceCertSend = false;

    private static final int MAX_PACKAGE_SIZE                   = 400;
    private static final int MAX_TRIES = 3;
    private static final int FLAG_NEW_INITIALIZATION = 0x80;

	private BBacnetNetwork bacnetNetwork;
    private BHonPlantControllerHMI hmiService;
    private HMIPrivateTransferListener privateTransferListener;
    private HMIIAmListener iAmListener;
    private boolean listenersAreRegistered = false;
    private IAlarmResolver alarmResolver;

    // Bacnet service id for unconfirmed private transfer
    private static final int UNCONFIRMED_PRIVATE_TRANSFER_ID = 30;
    private static final int CONFIRMED_PRIVATE_TRANSFER_ID = 18;
    // Bacnet service id for I-Am
    private static final int SERVICE_ID_IAM = 26;
    private long waitTime = 1000;
    private static final String LOG_CLS_PREFIX = TYPE.getTypeName();
	private  static final int WAIT_TIME_FOR_INSTANCE_ID_TO_UPDATE_IN_HMI = 2000;
	private static final int LEFT_SHIFT_BACNET_DEVICE_ID_INT = 22;
	 private static final int HONEYWELL_CA_CERT_ORIDINAL = 1;
	 
	public String getTimeBeforeQueryAlarms() {
		return timeBeforeQueryAlarms;
	}

	public void setTimeBeforeQueryAlarms(String timeBeforeQueryAlarms) {
		this.timeBeforeQueryAlarms = timeBeforeQueryAlarms;
	}

	public String getTimeAfterQueryAlarms() {
		return timeAfterQueryAlarms;
	}

	public void setTimeAfterQueryAlarms(String timeAfterQueryAlarms) {
		this.timeAfterQueryAlarms = timeAfterQueryAlarms;
	}

	private String timeBeforeQueryAlarms = "";
	private String timeAfterQueryAlarms = "";

}
