/*
 * Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
 * This file contains trade secrets of Honeywell International, Inc. No part
 * may be reproduced or transmitted in any form by any means or for any
 * purpose without the express written permission of Honeywell.
 */
package com.honeywell.honplantcontroller.hmi.comm;

import java.text.MessageFormat;
import java.util.Arrays;
import java.util.Timer;

import javax.baja.bacnet.BBacnetNetwork;
import javax.baja.bacnet.datatypes.BBacnetAddress;
import javax.baja.job.BJobService;
import javax.baja.sys.Sys;

import com.honeywell.honplantcontroller.hmi.service.BHonPlantControllerHMI;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerBacnetUtils;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerHMILogger;
import com.tridium.bacnet.job.BBacnetDiscoverDevicesJob;
import com.tridium.bacnet.services.unconfirmed.IAmRequest;
import com.tridium.bacnet.stack.IAmListener;
import com.tridium.bacnet.stack.network.BNetworkPort;

public class HMIIAmListener implements IAmListener {
    public HMIIAmListener(BHonPlantControllerHMI hmiService) {
        this.hmiService = hmiService;
    }

    @Override
    public void receiveIAm(IAmRequest iAmRequest, BBacnetAddress bacnetAddress) {
        final String LOG_MSG_PREFIX = getClass().getSimpleName() + ".receiveIAm";
        if (iAmRequest.getVendorId() != HonPlantControllerBacnetUtils.HONEYWELL_BACNET_VENDOR_ID) {
            return;
        }
         if (hmiService == null) {
            HonPlantControllerHMILogger.finest(MessageFormat.format("{0} no HMI service", LOG_MSG_PREFIX));
            return;
        }
        
        if (hmiService.getIsHmiJobActive()) {
            HonPlantControllerHMILogger.finest(MessageFormat.format("{0} A HMI Job is already Active. ", LOG_MSG_PREFIX));
            return;
        }
       
		if (!hmiService.isHMIDevice(iAmRequest.getObjectId(), bacnetAddress)) {
			return;
		}
		
		hmiService.setInstanceIdFromHMi(iAmRequest.getObjectId());
		hmiService.setHMIInstSet(true);
		
        // check HMI port
        
        BBacnetNetwork bacnetNetwork = hmiService.getBacnetNetwork();
        BHonPlantControllerBacnetProperties hmiBacnetProperties = hmiService.getBacnetProperties();
        BNetworkPort hmiNetworkPort = HonPlantControllerBacnetUtils.getHMIPort(bacnetNetwork, hmiBacnetProperties);
        if (hmiNetworkPort == null) {
            HonPlantControllerHMILogger.finest(MessageFormat.format("{0} no HMI port", LOG_MSG_PREFIX));
            return;
        }

        if (bacnetAddress.getNetworkNumber() != hmiNetworkPort.getAddress().getNetworkNumber()) {
            HonPlantControllerHMILogger.finest(MessageFormat.format("{0} wrong port: bacnetAddress: networkNumber[{1}] " +
                    "hmiPort.networkNumber[{2}]", LOG_MSG_PREFIX,
                bacnetAddress.getNetworkNumber(),
                hmiNetworkPort.getAddress().getNetworkNumber()));
            return;
        }

        if (null != findRunningBacnetDiscoverJob()) {
            HonPlantControllerHMILogger.fine(MessageFormat.format("{0} IAM-99 from DiscoveryJob",
                    LOG_MSG_PREFIX));
            return;
        }
        
        // it is better to keep this messages at level INFO, so we can see, when the HMI connects to the station.
        HonPlantControllerHMILogger.info(MessageFormat.format("{0} IAM-99 from HMI port, address[{1}]",
                LOG_MSG_PREFIX, bacnetAddress));

        hmiPresent = true;
        byte[] macAddr = bacnetAddress.getMacAddress().getAddr();
        int addrLen = macAddr.length;
        hmiMacAddress = new byte[addrLen];
        System.arraycopy(macAddr, 0, hmiMacAddress, 0, addrLen);
       
           
        //Since we don't know when Controller sends out own IAM, wait for 1 second before starting Key Exchange
		startKeyExchangeWithDelay();
    }

	public void threadVerifyFirmware() {
		if((threadVerifyFirmware == null) || !threadVerifyFirmware.isAlive()) {
			threadVerifyFirmware = new Thread(hmiService::verifyFirmware);
			threadVerifyFirmware.start();
        }
	}

    public BBacnetDiscoverDevicesJob findRunningBacnetDiscoverJob() {

        BJobService jobService = Sys.getStation().getServices().getChildren(BJobService.class)[0];


        return Arrays.stream(jobService.getChildComponents())
                .filter(BBacnetDiscoverDevicesJob.class::isInstance)
                .map(BBacnetDiscoverDevicesJob.class::cast)
                .filter(job -> job.getJobState().isRunning())
                .findFirst().orElse(null);
    }

    protected void startKeyExchangeWithDelay() {
        Timer t = new java.util.Timer();
        t.schedule(
                new java.util.TimerTask() {
                    @Override
                    public void run() {
                        try {
                            hmiService.doStartKeyExchange();
                        } catch (Exception exception) {
                        	HonPlantControllerHMILogger.finest(exception.getMessage(), exception);
                        }
                        t.cancel();
                    }
                },
                waitTime
        );
    }

    public boolean hmiIsPresent() {
        return hmiPresent;
    }

    public void resetHmiIsPresent() {
        hmiPresent = false;
        hmiMacAddress = null;
    }

    public byte[] getHmiMacAddress() {
        return hmiMacAddress;
    }

    public void stopThreadVerifyFirmware() {
        if (threadVerifyFirmware != null) {
            threadVerifyFirmware.interrupt();
        }
    }

    private boolean hmiPresent = false;
    private byte[] hmiMacAddress = null;
    private final BHonPlantControllerHMI hmiService;
    private Thread threadVerifyFirmware = null;
    private final Object waitObject = new Object();
    private long waitTime = 1000;
}
