/*
 * Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
 * This file contains trade secrets of Honeywell International, Inc. No part
 * may be reproduced or transmitted in any form by any means or for any
 * purpose without the express written permission of Honeywell.
 */
package com.honeywell.honplantcontroller.hmi.comm;

import java.security.cert.CertificateEncodingException;
import java.text.MessageFormat;
import java.util.logging.Level;

import javax.baja.bacnet.BacnetException;
import javax.baja.bacnet.datatypes.BBacnetAddress;
import javax.baja.bacnet.datatypes.BBacnetOctetString;
import javax.baja.bacnet.enums.BBacnetErrorClass;
import javax.baja.bacnet.enums.BBacnetErrorCode;
import javax.baja.nre.util.ByteArrayUtil;
import javax.baja.nre.util.ByteBuffer;
import javax.baja.nre.util.TextUtil;
import javax.baja.sys.BAbsTime;

import com.honeywell.honbacnethelper.listener.IPrivateTransferClient;
import com.honeywell.honplantcontroller.hmi.messages.QueryAlarmPointsMessage;
import com.honeywell.honplantcontroller.hmi.messages.QueryAlarmPointsResultMessage;
import com.honeywell.honplantcontroller.hmi.messages.QueryUnackAlarmPointsResultMessage;
import com.honeywell.honplantcontroller.hmi.messages.QueryUserLoginMessage;
import com.honeywell.honplantcontroller.hmi.messages.QueryUserLogoutMessage;
import com.honeywell.honplantcontroller.hmi.messages.QueryUserMessage;
import com.honeywell.honplantcontroller.hmi.messages.RequestAlarmAckMessage;
import com.honeywell.honplantcontroller.hmi.messages.RequestCertMessage;
import com.honeywell.honplantcontroller.hmi.utils.BHonPlantControllerHMIInitializationHandler;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerHMILogger;
import com.tridium.bacnet.asn.AsnInputStream;
import com.tridium.bacnet.asn.NErrorType;
import com.tridium.bacnet.services.error.ConfirmedPrivateTransferError;

public class HMIPrivateTransferListener implements IPrivateTransferClient
{
    public HMIPrivateTransferListener(BHMICommHandler commHandler) {
        this.commHandler = commHandler;
    }

    //This method receives ConfirmedPrivateTransfer from HMI to controller
	@Override
	public byte[] handleConfirmedPrivateTransfer(long vendorId, long serviceId, byte[] bytes, BBacnetAddress bacnetAddress) throws BacnetException {
			final String LOG_MSG_PREFIX = BHMICommHandler.TYPE.getTypeName() + "." + getClass().getSimpleName() + ".receiveConfirmedPrivateTransfer";
		if (vendorId == VENDOR_ID_HON) {
			ByteBuffer dataBuffer = new ByteBuffer(1);
			java.nio.ByteBuffer resultBuffer = java.nio.ByteBuffer.allocate(0);
			switch ((int) serviceId) {
			case SERVICE_ID_QUERY_ALARM_POINTS: {
				setTimeOfRequestEntry(MessageFormat.format("Time of request entry is {0}", BAbsTime.now().getMillis()));
				QueryAlarmPointsResultMessage result = commHandler.handleQueryAlarmPointsMessage(new QueryAlarmPointsMessage(bytes));
				if (result != null) {
					byte[] rawBytes = result.getBytes();
					dataBuffer.write(rawBytes);
				}
				setTimeOfRequestComplete(MessageFormat.format("Time of request complete is {0}", BAbsTime.now().getMillis()));
				
				if (null != commHandler.getHmiService()) {
					HonPlantControllerHMILogger.timeLog(Level.FINEST, getTimeOfRequestEntry());
					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getTimeBeforeQueryAlarms());

					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getHmiService().getHmiAlarmHandler().getTimeOfQueryAlarm());
					HonPlantControllerHMILogger.timeLog(Level.FINEST,
							commHandler.getHmiService().getHmiAlarmHandler().getTimeAfterCheckingIndicesValid());

					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getAlarmResolver().getTimeBeforeGettingAlarmSources());
					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getAlarmResolver().getTimeAfterGettingAlarmSources());

					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getAlarmResolver().getTimeBeforeStartingRecurssionOfAlarmRecords());
					HonPlantControllerHMILogger.timeLog(Level.FINEST,
							commHandler.getAlarmResolver().getTimeAfterCompletingRecurssionOfAlarmRecords());

					HonPlantControllerHMILogger.timeLog(Level.FINEST,
							commHandler.getHmiService().getHmiAlarmHandler().getTimeAfterCollectingALarms());

					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getTimeAfterQueryAlarms());
					HonPlantControllerHMILogger.timeLog(Level.FINEST, getTimeOfRequestComplete());
				}
			}
				break;
			case SERVICE_ID_QUERY_UNACK_ALARM_POINTS: {
				setTimeOfRequestEntry(MessageFormat.format("Time of request entry is {0}", BAbsTime.now().getMillis()));
				QueryUnackAlarmPointsResultMessage result = commHandler.handleQueryUnackAlarmPointsMessage(new QueryAlarmPointsMessage(bytes));
				// for testing, remove later
				if (result != null) {
					byte[] rawBytes = result.getBytes();
					dataBuffer.write(rawBytes);
				}
				setTimeOfRequestComplete(MessageFormat.format("Time of request complete is {0}", BAbsTime.now().getMillis()));
				if (null != commHandler.getHmiService()) {
					HonPlantControllerHMILogger.timeLog(Level.FINEST, getTimeOfRequestEntry());
					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getTimeBeforeQueryAlarms());

					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getHmiService().getHmiAlarmHandler().getTimeOfQueryAlarm());
					HonPlantControllerHMILogger.timeLog(Level.FINEST,
							commHandler.getHmiService().getHmiAlarmHandler().getTimeAfterCheckingIndicesValid());

					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getAlarmResolver().getTimeBeforeGettingAlarmSources());
					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getAlarmResolver().getTimeAfterGettingAlarmSources());

					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getAlarmResolver().getTimeBeforeStartingRecurssionOfAlarmRecords());
					HonPlantControllerHMILogger.timeLog(Level.FINEST,
							commHandler.getAlarmResolver().getTimeAfterCompletingRecurssionOfAlarmRecords());

					HonPlantControllerHMILogger.timeLog(Level.FINEST,
							commHandler.getHmiService().getHmiAlarmHandler().getTimeAfterCollectingALarms());

					HonPlantControllerHMILogger.timeLog(Level.FINEST, commHandler.getTimeAfterQueryAlarms());
					HonPlantControllerHMILogger.timeLog(Level.FINEST, getTimeOfRequestComplete());
				}
			}
				break;
			case SERVICE_ID_REQUEST_LANGUAGE_FILE:
				HonPlantControllerHMILogger.finest(MessageFormat.format("{0} REQUEST_LANGUAGE_FILE from MHMI port, " + "address[{1}] bytes[{2}:{3}]",
						LOG_MSG_PREFIX, bacnetAddress, bytes.length, TextUtil.bytesToHexString(bytes)));
				dataBuffer.write(commHandler.handleRequestLanguageFile(bytes));
				break;
			case SERVICE_ID_VERIFY_INTER_CERT:
				HonPlantControllerHMILogger.finest(
						MessageFormat.format("{0} {1} from HMI port, address[{2}]", LOG_MSG_PREFIX, SERVICE_ID_VERIFY_INTER_CERT, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_REQUEST_CERT_EXCHANGE:
				HonPlantControllerHMILogger.finest(
						MessageFormat.format("{0} {1} from hMI port, address[{2}]", LOG_MSG_PREFIX, SERVICE_ID_REQUEST_CERT_EXCHANGE, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_VERIFY_DEVICE_CERT:
				HonPlantControllerHMILogger.finest(
						MessageFormat.format("{0} {1} from HMI port, address[{2}]", LOG_MSG_PREFIX, SERVICE_ID_VERIFY_DEVICE_CERT, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_START_ECDH:
				HonPlantControllerHMILogger
						.finest(MessageFormat.format("{0} {1} from HMI port, address[{2}]", LOG_MSG_PREFIX, SERVICE_ID_START_ECDH, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_VERIFY_MST_SECRET:
				HonPlantControllerHMILogger.finest(
						MessageFormat.format("{0} {1} from HMI port, address[{2}]", LOG_MSG_PREFIX, SERVICE_ID_VERIFY_MST_SECRET, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_CHANGE_CIPHER_MST_SECRET:
				HonPlantControllerHMILogger.finest(MessageFormat.format("{0} {1} from HMI port, address[{2}]", LOG_MSG_PREFIX,
						SERVICE_ID_CHANGE_CIPHER_MST_SECRET, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_DERIVE_SHARED_KEY:
				HonPlantControllerHMILogger.finest(
						MessageFormat.format("{0} {1} from HMI port, address[{2}]", LOG_MSG_PREFIX, SERVICE_ID_DERIVE_SHARED_KEY, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_VERIFY_SHARED_KEY:
				HonPlantControllerHMILogger.finest(
						MessageFormat.format("{0} {1} from HMI port, address[{2}]", LOG_MSG_PREFIX, SERVICE_ID_VERIFY_SHARED_KEY, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_CHANGE_CIPHER_SHARED_KEY:
				HonPlantControllerHMILogger.finest(MessageFormat.format("{0} {1} from HMI port, address[{2}]", LOG_MSG_PREFIX,
						SERVICE_ID_CHANGE_CIPHER_SHARED_KEY, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_CERT_REQUEST_FROM_HMI:
				HonPlantControllerHMILogger.finest(
						MessageFormat.format("{0} {1} from HMI port, address[{2}]", LOG_MSG_PREFIX, SERVICE_ID_CERT_REQUEST_FROM_HMI, bacnetAddress));
				try {
					dataBuffer.write(commHandler.handleCertRequestFromHMI(new RequestCertMessage(bytes)));
				} catch (CertificateEncodingException e) {
					HonPlantControllerHMILogger.finest(e.getMessage(), e);
				}
				commHandler.sendVerifyDeviceCert();
				break;
			case SERVICE_ID_USER_LOGIN:
				HonPlantControllerHMILogger.finest(MessageFormat.format("{0} USER_LOGIN from HMI port, address[{1}]", LOG_MSG_PREFIX, bacnetAddress));

				dataBuffer.write(commHandler.handleUserLogin(QueryUserLoginMessage.make(bytes)));
				break;
			case SERVICE_ID_USER_LOGOUT:
				HonPlantControllerHMILogger
						.finest(MessageFormat.format("{0} USER_LOGOUT from HMI port, address[{1}]", LOG_MSG_PREFIX, bacnetAddress));
				dataBuffer.write(commHandler.handleUserLogout(QueryUserLogoutMessage.make(bytes)));
				break;
			case SERVICE_ID_MODIFY_PASSWORD:
				HonPlantControllerHMILogger
						.finest(MessageFormat.format("{0} MODIFY_PASSWORD from HMI port, address[{1}]", LOG_MSG_PREFIX, bacnetAddress));
				dataBuffer.write(commHandler.handleChangePassword(QueryUserMessage.make(bytes)));
				break;
			case SERVICE_ID_CREATE_USER:
				HonPlantControllerHMILogger
						.finest(MessageFormat.format("{0} CREATE_USER from HMI port, address[{1}]", LOG_MSG_PREFIX, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_DELETE_USER:
				HonPlantControllerHMILogger
						.finest(MessageFormat.format("{0} DELETE_USER from HMI port, address[{1}]", LOG_MSG_PREFIX, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_READ_USERS:

				AsnInputStream serviceParameter = AsnInputStream.make(bytes);

				BBacnetOctetString requestIndex = serviceParameter.readBacnetOctetString();

				HonPlantControllerHMILogger.finest(MessageFormat.format("{0} READ_USERS at index {1}", LOG_MSG_PREFIX, requestIndex));

				// encode the index from it's byte representation to integer.
				byte[] idxBytes = requestIndex.getBytes();
				int userRequestIndex = 0;
				for (int index = 0; index < idxBytes.length; index++) {
					userRequestIndex |= (idxBytes[index] << (index * 4));
				}
				HonPlantControllerHMILogger.finest(MessageFormat.format("{0} requested Index: {1}", LOG_MSG_PREFIX, userRequestIndex));

				dataBuffer.write(commHandler.handleReadUserRequest(userRequestIndex));
				break;

			case SERVICE_ID_REQ_SALT:
				HonPlantControllerHMILogger
						.finest(MessageFormat.format("{0} {1} from HMI port, address[{2}]", LOG_MSG_PREFIX, SERVICE_ID_REQ_SALT, bacnetAddress));

				// for testing, remove later
				try {
					dataBuffer.write(commHandler.handleRequestSalt(QueryUserLoginMessage.make(bytes)));
				} catch (Exception e) {
					HonPlantControllerHMILogger.finest(e.getMessage(), e);
					return new byte[0];
				}
				break;
			case SERVICE_ID_REQ_RAND:
				HonPlantControllerHMILogger
						.finest(MessageFormat.format("{0} {1} from HMI port, address[{2}]", LOG_MSG_PREFIX, SERVICE_ID_REQ_RAND, bacnetAddress));
				// for testing, remove later
				try {
					dataBuffer.write(commHandler.handleRequestRand());
				} catch (Exception e) {
					HonPlantControllerHMILogger.finest(e.getMessage(), e);
					return new byte[0];
				}
				break;
			case SERVICE_ID_USE_INIT:
				HonPlantControllerHMILogger.finest(MessageFormat.format("{0} USE_INIT from HMI port, address[{1}]", LOG_MSG_PREFIX, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_DATA_SYNC:
				HonPlantControllerHMILogger.finest(MessageFormat.format("{0} DATA_SYNC from HMI port, address[{1}]", LOG_MSG_PREFIX, bacnetAddress));
				// for testing, remove later
				dataBuffer.writeByte((byte) serviceId);
				break;
			case SERVICE_ID_ALARM_ACK: {
				byte[] result = handleAlarmAck(bytes);
				if (result != null) {
					dataBuffer.write(result);
				}
			}
				break;
			case BHonPlantControllerHMIInitializationHandler.SERVICE_ID_DEVICE_INFO_NOTIFICATION: 
					commHandler.handleServiceDeviceInfo(bytes);

				break;	
			default:
				
				
				HonPlantControllerHMILogger
						.finest(MessageFormat.format("HMI instance ID is not same as controller configuration", LOG_MSG_PREFIX, bacnetAddress));
				break;
			}
			return dataBuffer.toByteArray();
		} else {
			return new byte[0];
		}
	}

    private byte[] handleAlarmAck(byte[] bytes) {
        byte[] result;
        try {
            result = commHandler.handleAlarmAck(RequestAlarmAckMessage.make(bytes));
        } catch (Exception e) {
            HonPlantControllerHMILogger.warning(MessageFormat.format("can not decode alarm ack message {0} because of {1}", ByteArrayUtil.toHexString(bytes), e));
            result = new ConfirmedPrivateTransferError(
                    new NErrorType(BBacnetErrorClass.COMMUNICATION, BBacnetErrorCode.MALFORMED_MESSAGE),
                    HMIPrivateTransferListener.VENDOR_ID_HON,
                    HMIPrivateTransferListener.SERVICE_ID_ALARM_ACK).toEncodedBytes();
        }
        return result;
    }

    @Override
    public void handleUnconfirmedPrivateTransfer(long vendorId, long serviceId, byte[] bytes, BBacnetAddress bacnetAddress) throws BacnetException {
        final String LOG_MSG_PREFIX = BHMICommHandler.TYPE.getTypeName() + "." + getClass().getSimpleName() + ".receiveUnconfirmedPrivateTransfer";
        HonPlantControllerHMILogger.finest(MessageFormat.format("{0} VendorId: {1}, ServiceId: {2}", LOG_MSG_PREFIX, vendorId, serviceId));
        if (vendorId == VENDOR_ID_HON) {
            if (serviceId == SERVICE_ID_WRITE_POINT_MODE) {
                HonPlantControllerHMILogger.finest(MessageFormat.format("{0} SERVICE_ID_WRITE_POINT_MODE from HMI port, address[{1}]", LOG_MSG_PREFIX, bacnetAddress));
            }
        }
    }
    
	public String getTimeOfRequestEntry() {
		return timeOfRequestEntry;
	}

	public void setTimeOfRequestEntry(String timeOfRequestEntry) {
		this.timeOfRequestEntry = timeOfRequestEntry;
	}
	
	public String getTimeOfRequestComplete() {
		return timeOfRequestComplete;
	}

	public void setTimeOfRequestComplete(String timeOfRequestComplete) {
		this.timeOfRequestComplete = timeOfRequestComplete;
	}
	
	private String timeOfRequestEntry = "";
	private String timeOfRequestComplete = "";

    private static final int SELF_VENDOR_ID                     = 36;
    private BHMICommHandler commHandler                         = null;
     public static final int VENDOR_ID_HON                       = 17;
    //User Login
    private static final int SERVICE_ID_USER_LOGIN               = 60;
    private static final int SERVICE_ID_USER_LOGOUT              = 61;
    private static final int SERVICE_ID_MODIFY_PASSWORD          = 62;
    private static final int SERVICE_ID_CREATE_USER              = 63;
    private static final int SERVICE_ID_DELETE_USER              = 64;
    public  static final int SERVICE_ID_READ_USERS               = 65;

    private static final int SERVICE_ID_REQ_SALT                 = 66;
    private static final int SERVICE_ID_REQ_RAND                 = 67;
    private static final int SERVICE_ID_USE_INIT                 = 68;

    //Alarm
    private static final int SERVICE_ID_QUERY_ALARM_POINTS       = 70;
    private static final int SERVICE_ID_QUERY_UNACK_ALARM_POINTS = 71;
    //Point Mode
    private static final int SERVICE_ID_WRITE_POINT_MODE         = 76;
    //Language File
    private static final int SERVICE_ID_REQUEST_LANGUAGE_FILE    = 77;
    //Data Synchronization
    private static final int SERVICE_ID_DATA_SYNC                = 78;

    //Initialize Shared Key
    private static final int SERVICE_ID_VERIFY_INTER_CERT        = 79;
    private static final int SERVICE_ID_REQUEST_CERT_EXCHANGE    = 80;
    private static final int SERVICE_ID_VERIFY_DEVICE_CERT       = 81;
    private static final int SERVICE_ID_START_ECDH               = 82;
    private static final int SERVICE_ID_VERIFY_MST_SECRET        = 83;
    private static final int SERVICE_ID_CHANGE_CIPHER_MST_SECRET = 84;
    private static final int SERVICE_ID_DERIVE_SHARED_KEY        = 85;
    private static final int SERVICE_ID_VERIFY_SHARED_KEY        = 86;
    private static final int SERVICE_ID_CHANGE_CIPHER_SHARED_KEY = 87;
    private static final int SERVICE_ID_CERT_REQUEST_FROM_HMI    = 88;

    public static final int SERVICE_ID_DYNAMIC_DATA_SYNC         = 90;
    public static final int SERVICE_ID_ALARM_ACK                 = 91;
    
}
