/*
Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
This file contains trade secrets of Honeywell International, Inc. No part
may be reproduced or transmitted in any form by any means or for any
purpose without the express written permission of Honeywell.
*/

package com.honeywell.honplantcontroller.hmi.messages;

import java.util.ArrayList;
import java.util.List;

import javax.baja.bacnet.datatypes.BBacnetUnsigned;
import javax.baja.bacnet.enums.BBacnetEngineeringUnits;
import javax.baja.sys.BBoolean;
import javax.baja.sys.BDouble;
import javax.baja.sys.BEnum;
import javax.baja.sys.BFloat;
import javax.baja.sys.BInteger;
import javax.baja.sys.BLong;
import javax.baja.sys.BNumber;
import javax.baja.sys.BString;
import javax.baja.sys.BValue;

import com.honeywell.honplantcontroller.hmi.alarm.BacnetAlarmSummary;
import com.tridium.bacnet.asn.AsnOutputStream;

public class QueryAlarmPointsResultMessage {

    private BBacnetUnsigned alarmPointsCount;
    private BBacnetUnsigned urgentAlarmCount;
    private BBacnetUnsigned highAlarmCount;
    private BBacnetUnsigned lowAlarmCount;
    private List<BacnetAlarmSummary> alarmSummaries = new ArrayList<>();

    private QueryAlarmPointsResultMessage() {

    }

    public static QueryAlarmPointsResultMessage make(long alarmPointsCount, long urgentAlarmCount, long highAlarmCount, long lowAlarmCount) {

        QueryAlarmPointsResultMessage hmiAlarmResult = new QueryAlarmPointsResultMessage();

        hmiAlarmResult.alarmPointsCount = new BBacnetUnsigned(alarmPointsCount);
        hmiAlarmResult.urgentAlarmCount = new BBacnetUnsigned(urgentAlarmCount);
        hmiAlarmResult.highAlarmCount = new BBacnetUnsigned(highAlarmCount);
        hmiAlarmResult.lowAlarmCount = new BBacnetUnsigned(lowAlarmCount);


        return hmiAlarmResult;
    }

    public static QueryAlarmPointsResultMessage make(long alarmPointsCount, long urgentAlarmCount, long highAlarmCount, long lowAlarmCount, List<BacnetAlarmSummary> alarmSummaries) {
        
        QueryAlarmPointsResultMessage hmiAlarmResult = QueryAlarmPointsResultMessage.make(alarmPointsCount, urgentAlarmCount, highAlarmCount, lowAlarmCount);
        hmiAlarmResult.alarmSummaries = alarmSummaries;

        return hmiAlarmResult;
    }
    
    public static QueryAlarmPointsResultMessage make(QueryAlarmPointsResultMessage hmiAlarmResult, List<BacnetAlarmSummary> alarmSummaries) {

        hmiAlarmResult.alarmSummaries = alarmSummaries;
        return hmiAlarmResult;
    }

    public byte[] getBytes() {

        AsnOutputStream asnOutputStream = new AsnOutputStream();

        asnOutputStream.writeUnsignedInteger(0, alarmPointsCount.getInt());
        asnOutputStream.writeUnsignedInteger(1, urgentAlarmCount.getInt());
        asnOutputStream.writeUnsignedInteger(2, highAlarmCount.getInt());
        asnOutputStream.writeUnsignedInteger(3, lowAlarmCount.getInt());

        if (!alarmSummaries.isEmpty()) {
            asnOutputStream.writeOpeningTag(4);
            alarmSummaries.stream().forEach(alarmSummary -> alarmSummaryBytes(alarmSummary, asnOutputStream));
            asnOutputStream.writeClosingTag(4);
        }

        return asnOutputStream.toByteArray();
    }

    public static void alarmSummaryBytes(BacnetAlarmSummary alarmSummary, AsnOutputStream asnOutputStream) {

        asnOutputStream.writeObjectIdentifier(alarmSummary.getObjectIdentifier());

        asnOutputStream.writeEnumerated(alarmSummary.getEventState());

        asnOutputStream.writeOpeningTag(2);
        asnOutputStream.writeDate(alarmSummary.getEventTimeStamp().getDateTime().getDate());
        asnOutputStream.writeTime(alarmSummary.getEventTimeStamp().getDateTime().getTime());
        asnOutputStream.writeClosingTag(2);

        asnOutputStream.writeUnsigned(alarmSummary.getPriority());

        asnOutputStream.writeOpeningTag(4);
        QueryAlarmPointsResultMessage.writeAlarmValue(alarmSummary.getAlarmPresentValue(), asnOutputStream);
        asnOutputStream.writeClosingTag(4);

        BBacnetEngineeringUnits unit = alarmSummary.getUnit();
        asnOutputStream.writeEnumerated(unit);

    }

    public static void writeAlarmValue(BValue alarmValue, AsnOutputStream asnOutputStream) {

        if (alarmValue instanceof BDouble || alarmValue instanceof BFloat) {
            asnOutputStream.writeReal((BNumber)alarmValue);
        } else if (alarmValue instanceof BInteger) {
            asnOutputStream.writeSignedInteger((BInteger)alarmValue);
        } else if (alarmValue instanceof BLong) {
            asnOutputStream.writeUnsignedInteger(((BLong)alarmValue).getLong());
        } else if (alarmValue instanceof BBoolean) {
            asnOutputStream.writeEnumerated((BBoolean)alarmValue);
        }
        else if(alarmValue instanceof BEnum) {
            asnOutputStream.writeUnsignedInteger(((BEnum)alarmValue).getOrdinal());
        }
        else if (alarmValue instanceof BString) {
            asnOutputStream.writeCharacterString((BString)alarmValue);
        }
    }

    public BBacnetUnsigned getAlarmPointsCount() {
        return alarmPointsCount;
    }
    public BBacnetUnsigned getUrgentAlarmCount() {
        return urgentAlarmCount;
    }
    public BBacnetUnsigned getHighAlarmCount() {
        return highAlarmCount;
    }
    public BBacnetUnsigned getLowAlarmCount() {
        return lowAlarmCount;
    }

    
    @Override
    public String toString() {
        //{----------------------only for testing, remove for production-------------------------

        StringBuilder stringBuilder = new StringBuilder();
        stringBuilder.append("Alarm Points Count: "+alarmPointsCount);
        stringBuilder.append("\n");
        stringBuilder.append("Urgent Alarm Count: "+urgentAlarmCount);
        stringBuilder.append("\n");
        stringBuilder.append("High Alarm Count: "+highAlarmCount);
        stringBuilder.append("\n");
        stringBuilder.append("Low Alarm Count: "+lowAlarmCount);
        stringBuilder.append("\n");
        
        stringBuilder.append(" List of alarm summaries {");
        stringBuilder.append("\n");

        for (BacnetAlarmSummary alarmSummary : alarmSummaries) {
            stringBuilder.append("  Object identifier: "+alarmSummary.getObjectIdentifier());
            stringBuilder.append("\n");
            stringBuilder.append("  Event State: "+alarmSummary.getEventState());
            stringBuilder.append("\n");
            stringBuilder.append("  Event Time Stamp: "+alarmSummary.getEventTimeStamp());
            stringBuilder.append("\n");
            stringBuilder.append("  Priority: "+alarmSummary.getPriority());
            stringBuilder.append("\n");
            stringBuilder.append("  Present Value: "+alarmSummary.getAlarmPresentValue());
            stringBuilder.append("\n");
            stringBuilder.append("  Unit: "+alarmSummary.getUnit());
            stringBuilder.append("\n");
            stringBuilder.append("----");
        }
        stringBuilder.append("\n");
        stringBuilder.append("}");


        return stringBuilder.toString();
        //}----------------------only for testing, remove for production-------------------------
    }
}
