/*
Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
This file contains trade secrets of Honeywell International, Inc. No part
may be reproduced or transmitted in any form by any means or for any
purpose without the express written permission of Honeywell.
*/

package com.honeywell.honplantcontroller.hmi.messages;

import java.nio.charset.Charset;
import java.text.MessageFormat;
import java.util.logging.Logger;

import javax.baja.alarm.BSourceState;
import javax.baja.bacnet.datatypes.BBacnetObjectIdentifier;
import javax.baja.bacnet.datatypes.BBacnetTimeStamp;
import javax.baja.bacnet.enums.BBacnetEventState;
import javax.baja.bacnet.enums.BCharacterSetEncoding;
import javax.baja.bacnet.io.AsnException;
import javax.baja.nre.util.ByteArrayUtil;
import javax.baja.sys.BAbsTime;
import javax.baja.sys.BEnum;

import com.tridium.bacnet.asn.AsnInputStream;
import com.tridium.bacnet.services.confirmed.AcknowledgeAlarmRequest;

/**
 * This class represents a private transfer message, that HMI sends to acknowledge an alarm
 * Its code is derived from Tridium's {@link AcknowledgeAlarmRequest} 
 */

//TODO: The UUID of alarm is missing :-(
//TODO: except eventObjectId, the mapping of such AckMessage to the Alarms in AlarmDB is not possible
public class RequestAlarmAckMessage {
  
  private BBacnetObjectIdentifier eventObjectId;
  private BEnum eventStateAcknowledged;
  private BBacnetTimeStamp timeStamp;
  private String acknowledgementSource;
  private BBacnetTimeStamp timeOfAcknowledgement;
  private BCharacterSetEncoding encoding;

  private static final Logger LOG = Logger.getLogger("HonPlantControllerHMI");

  public static RequestAlarmAckMessage make(byte[] bytes) throws AsnException  {
    if (null == bytes) {
      return null;
    }

    AsnInputStream serviceParameter = AsnInputStream.make(bytes);

    byte[] data = null;

    try {
      data = serviceParameter.readOctetString();
    }
    catch (AsnException asne) {
      LOG.finest(() -> "RequestAlarmAckMessage.make() CharacterString: Could not read data from ServiceParameter");
      return null;
    }

    return new RequestAlarmAckMessage(data);
  }

  private RequestAlarmAckMessage(byte[] rawMsg) throws AsnException  {
    readEncoded(rawMsg);
  }

  protected void readEncoded(byte[] rawMsg) throws AsnException {
      AsnInputStream inputStream = new AsnInputStream(rawMsg);

      inputStream.readUnsignedInteger(0);
      this.eventObjectId = inputStream.readObjectIdentifier(1);
      this.eventStateAcknowledged = BBacnetEventState.make(inputStream.readEnumerated(2));
      inputStream.skipTag();
      this.timeStamp = new BBacnetTimeStamp();
      this.timeStamp.readAsn(inputStream);
      inputStream.skipTag();
      this.encoding = inputStream.peekEncoding(4);
      this.acknowledgementSource = inputStream.readCharacterString(4);
      inputStream.skipTag();
      this.timeOfAcknowledgement = new BBacnetTimeStamp();
      this.timeOfAcknowledgement.readAsn(inputStream);
      inputStream.skipTag();
  }


  public BBacnetObjectIdentifier getEventObjectId() {
    return eventObjectId;
  }

  public BSourceState getEventStateAcknowledged() {
    return BBacnetEventState.make(eventStateAcknowledged.getOrdinal()).toSourceState();
  }

  public BAbsTime getTimeStamp() {
    return timeStamp.toBAbsTime();
  }

  public BAbsTime getTimeOfAcknowledgement() {
    return timeOfAcknowledgement.toBAbsTime();
  }

  public String getAcknowledgementSource() {
    return acknowledgementSource;
  }

  public void setEventObjectId(BBacnetObjectIdentifier eventObjectId) {
    this.eventObjectId = eventObjectId;
  }

  public void setEventStateAcknowledged(BEnum eventStateAcknowledged) {
    this.eventStateAcknowledged = eventStateAcknowledged;
  }

  public void setTimeStamp(BBacnetTimeStamp timeStamp) {
    this.timeStamp = timeStamp;
  }

  public void setAcknowledgementSource(String acknowledgementSource) {
    this.acknowledgementSource = acknowledgementSource;
  }

  public void setTimeOfAcknowledgement(BBacnetTimeStamp timeOfAcknowledgement) {
    this.timeOfAcknowledgement = timeOfAcknowledgement;
  }

  public void setEncoding(BCharacterSetEncoding encoding) {
    this.encoding = encoding;
  }

  public String toDebugString() {
    StringBuilder sb = new StringBuilder();
    sb.append("eventObjectId: " + eventObjectId);
    sb.append("\neventStateAcknowledged: " + eventStateAcknowledged);
    sb.append("\ntimeStamp: " + timeStamp);
    sb.append("\nacknowledgementSource: " + acknowledgementSource);
    sb.append("\ntimeOfAcknowledgement: " + timeOfAcknowledgement);

    return sb.toString();
  }

}
