/*
Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
This file contains trade secrets of Honeywell International, Inc. No part
may be reproduced or transmitted in any form by any means or for any
purpose without the express written permission of Honeywell.
*/
package com.honeywell.honplantcontroller.hmi.service;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.security.AccessController;
import java.security.KeyStore;
import java.security.PrivilegedAction;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.security.interfaces.ECPrivateKey;
import java.security.interfaces.ECPublicKey;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Base64;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Optional;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.baja.bacnet.BBacnetNetwork;
import javax.baja.bacnet.BacnetConst;
import javax.baja.bacnet.BacnetException;
import javax.baja.bacnet.datatypes.BBacnetAddress;
import javax.baja.bacnet.datatypes.BBacnetBitString;
import javax.baja.bacnet.datatypes.BBacnetObjectIdentifier;
import javax.baja.bacnet.datatypes.BBacnetUnsigned;
import javax.baja.bacnet.enums.BBacnetDeviceStatus;
import javax.baja.bacnet.enums.BBacnetObjectType;
import javax.baja.bacnet.enums.BBacnetPropertyIdentifier;
import javax.baja.bacnet.enums.BCharacterSetEncoding;
import javax.baja.bacnet.export.BBacnetCalendarDescriptor;
import javax.baja.bacnet.export.BBacnetPointDescriptor;
import javax.baja.bacnet.export.BBacnetScheduleDescriptor;
import javax.baja.bacnet.util.BacnetBitStringUtil;
import javax.baja.file.BDataFile;
import javax.baja.file.BDirectory;
import javax.baja.file.BIFile;
import javax.baja.job.BJobState;
import javax.baja.job.BSimpleJob;
import javax.baja.naming.BOrd;
import javax.baja.naming.SlotPath;
import javax.baja.nre.annotations.Facet;
import javax.baja.nre.annotations.NiagaraAction;
import javax.baja.nre.annotations.NiagaraProperty;
import javax.baja.nre.annotations.NiagaraType;
import javax.baja.security.crypto.CertManagerFactory;
import javax.baja.security.crypto.ICryptoManager;
import javax.baja.security.crypto.IKeyStore;
import javax.baja.security.crypto.ITrustStore;
import javax.baja.sys.Action;
import javax.baja.sys.BBoolean;
import javax.baja.sys.BComponent;
import javax.baja.sys.BDynamicEnum;
import javax.baja.sys.BEnumRange;
import javax.baja.sys.BFacets;
import javax.baja.sys.BIService;
import javax.baja.sys.BInteger;
import javax.baja.sys.BString;
import javax.baja.sys.BValue;
import javax.baja.sys.Context;
import javax.baja.sys.Flags;
import javax.baja.sys.LocalizableException;
import javax.baja.sys.LocalizableRuntimeException;
import javax.baja.sys.Property;
import javax.baja.sys.Subscriber;
import javax.baja.sys.Sys;
import javax.baja.sys.Type;
import javax.baja.units.BDimension;
import javax.baja.units.BUnit;
import javax.baja.user.BUser;
import javax.baja.util.BFormat;
import javax.baja.util.BServiceContainer;
import javax.baja.util.Lexicon;
import javax.baja.util.Version;

import com.honeywell.honplantcontroller.hmi.alarm.BAlarmPriorityConfig;
import com.honeywell.honplantcontroller.hmi.comm.BHMIAlarmHandler;
import com.honeywell.honplantcontroller.hmi.comm.BHMICommHandler;
import com.honeywell.honplantcontroller.hmi.comm.BHonPlantControllerBacnetProperties;
import com.honeywell.honplantcontroller.hmi.comm.HMIIAmListener;
import com.honeywell.honplantcontroller.hmi.enums.BHMIPointExportOptions;
import com.honeywell.honplantcontroller.hmi.enums.BHMISyncCategory;
import com.honeywell.honplantcontroller.hmi.enums.BHMISyncStatus;
import com.honeywell.honplantcontroller.hmi.enums.KeyExchangeState;
import com.honeywell.honplantcontroller.hmi.job.BBaseHMIJob;
import com.honeywell.honplantcontroller.hmi.job.BCommissionHMIJob;
import com.honeywell.honplantcontroller.hmi.job.BHMIFileDownloadJob;
import com.honeywell.honplantcontroller.hmi.utils.BHMIFirmwareDownloadConfig;
import com.honeywell.honplantcontroller.hmi.utils.BHonPlantControllerHMIInitializationHandler;
import com.honeywell.honplantcontroller.hmi.utils.BHonPlantControllerHMIPointModeHandler;
import com.honeywell.honplantcontroller.hmi.utils.HMIReinitializeDeviceDownloadRequest;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerBacnetUtils;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerHMILexicon;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerHMILogger;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerUserUtils;
import com.honeywell.honplantcontroller.hmi.view.BHonPlantControllerHome;
import com.tridium.bacnet.services.BacnetServicePrimitive;
import com.tridium.bacnet.stack.BBacnetStack;
import com.tridium.bacnet.stack.BacnetStackException;
import com.tridium.bacnet.stack.client.BBacnetClientLayer;
import com.tridium.bacnet.stack.network.BNetworkPort;
import com.tridium.bacnet.stack.network.BNetworkPriority;
import com.tridium.bacnet.stack.server.BBacnetExportTable;
import com.tridium.crypto.core.io.CoreTrustStore;
import com.tridium.platcrypto.core.BCertManagerService;
import com.tridium.platcrypto.core.NTrustStore;

@NiagaraProperty(
    name = "enabled",
    type = "BBoolean",
    defaultValue = "BBoolean.FALSE",
    flags=Flags.READONLY
)
@NiagaraProperty(
    name = "BacnetProperties",
    type = "BHonPlantControllerBacnetProperties",
    defaultValue = "new BHonPlantControllerBacnetProperties()",
    flags = Flags.DEFAULT_ON_CLONE
)
@NiagaraProperty(
    name = "home",
    type="BHonPlantControllerHome",
    defaultValue = "new BHonPlantControllerHome()"
)

@NiagaraProperty(
    name = "hmiAlarmHandler",
    type="BHMIAlarmHandler",
    defaultValue = "new BHMIAlarmHandler()",
    flags = Flags.HIDDEN | Flags.READONLY | Flags.TRANSIENT
)
@NiagaraProperty(
    name = "hmiPointModeListener",
    type = "BHonPlantControllerHMIPointModeHandler",
    defaultValue = "new BHonPlantControllerHMIPointModeHandler()",
    flags = Flags.HIDDEN | Flags.READONLY | Flags.TRANSIENT
)
@NiagaraProperty(
    name = "supportHmiFirmwareDowngrade",
    type = "boolean",
    defaultValue = "false",
    flags = Flags.HIDDEN
)
@NiagaraProperty(
        name = "hmiInitializationListener",
        type = "BHonPlantControllerHMIInitializationHandler",
        defaultValue = "new BHonPlantControllerHMIInitializationHandler()",
        flags = Flags.HIDDEN | Flags.READONLY | Flags.TRANSIENT
)
@NiagaraAction(
    name="commissionHMI" , flags = Flags.HIDDEN
)
@NiagaraAction(
    name="notifyChangesToHMI" ,flags = Flags.HIDDEN
)

@NiagaraAction(
    name="firmwareUpdateHMI",
    parameterType = "BString",
    defaultValue = "BString.DEFAULT",
    flags = Flags.ASYNC
)
@NiagaraAction(
    name="getLanguageFileHMI",
    flags = Flags.ASYNC | Flags.HIDDEN
)
@NiagaraAction(
    name="languageFileHMI",
    parameterType = "BString",
    defaultValue = "BString.DEFAULT",
    flags = Flags.ASYNC | Flags.HIDDEN
)
@NiagaraAction(
    name="isHMIPresent",
    returnType = "BBoolean",
    flags = Flags.HIDDEN
)
@NiagaraAction(
    name="initiateFileCopy",
    parameterType = "BString",
    defaultValue = "BString.DEFAULT",
    returnType = "BString",
    flags = Flags.HIDDEN
)
@NiagaraAction(
    name="copyBinaryString",
    parameterType = "BHMIFirmwareDownloadConfig",
    defaultValue = "BHMIFirmwareDownloadConfig.make()",
    returnType = "BString",
    flags = Flags.HIDDEN
)
@NiagaraProperty(
    name = "falSyncStatus",
    type = "BHMISyncStatus",
    defaultValue = "BHMISyncStatus.DEFAULT",
    flags = Flags.READONLY
)

@NiagaraProperty(
    name = "deviceCertificate",
    type = "baja:DynamicEnum",
    defaultValue = "BDynamicEnum.DEFAULT",
    flags = Flags.HIDDEN | Flags.READONLY
)

@NiagaraProperty(
    name="pointName",
    type="BFormat",
    defaultValue = "BFormat.make(\"%parent.displayName%-%displayName%\")"
)

@NiagaraProperty(
    name="scheduleName",
    type="BFormat",
    defaultValue = "BFormat.make(\"%parent.displayName%-%displayName%\")"
)

@NiagaraProperty(
    name="calendarName",
    type="BFormat",
    defaultValue = "BFormat.make(\"%parent.displayName%-%displayName%\")"
)

@NiagaraProperty(
		name = "hmiAlarmDbLimit",
	    type = "int",
	    defaultValue = "100",
	    facets = {
	    		@Facet(name = "\"units\"", value = "BUnit.make(\"records\", BDimension.DEFAULT)"),
	    		@Facet(name = "\"min\"", value = "1"), @Facet(name = "\"max\"", value = "1500")
	    }
)
@NiagaraProperty(
    name = "alarmPriorityConfiguration",
    type = "BAlarmPriorityConfig",
    defaultValue = "new BAlarmPriorityConfig()"
)

@NiagaraProperty(
    name ="pointExportOption",
    type = "BHMIPointExportOptions",
    defaultValue = "BHMIPointExportOptions.DEFAULT"
)
@NiagaraProperty(
	    name = "isHmiJobActive",
	    type = "boolean",
	    defaultValue = "false",
	    flags=Flags.READONLY | Flags.HIDDEN | Flags.TRANSIENT
	)

@SuppressWarnings({"squid:S1845"})
@NiagaraType
public class BHonPlantControllerHMI extends BComponent {
/*+ ------------ BEGIN BAJA AUTO GENERATED CODE ------------ +*/
/*@ $com.honeywell.honplantcontroller.hmi.service.BHonPlantControllerHMI(574437010)1.0$ @*/
/* Generated Fri May 26 17:57:37 IST 2023 by Slot-o-Matic (c) Tridium, Inc. 2012 */

////////////////////////////////////////////////////////////////
// Property "enabled"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code enabled} property.
   * @see #getEnabled
   * @see #setEnabled
   */
  public static final Property enabled = newProperty(Flags.READONLY, ((BBoolean)(BBoolean.FALSE)).getBoolean(), null);
  
  /**
   * Get the {@code enabled} property.
   * @see #enabled
   */
  public boolean getEnabled() { return getBoolean(enabled); }
  
  /**
   * Set the {@code enabled} property.
   * @see #enabled
   */
  public void setEnabled(boolean v) { setBoolean(enabled, v, null); }

////////////////////////////////////////////////////////////////
// Property "BacnetProperties"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code BacnetProperties} property.
   * @see #getBacnetProperties
   * @see #setBacnetProperties
   */
  public static final Property BacnetProperties = newProperty(Flags.DEFAULT_ON_CLONE, new BHonPlantControllerBacnetProperties(), null);
  
  /**
   * Get the {@code BacnetProperties} property.
   * @see #BacnetProperties
   */
  public BHonPlantControllerBacnetProperties getBacnetProperties() { return (BHonPlantControllerBacnetProperties)get(BacnetProperties); }
  
  /**
   * Set the {@code BacnetProperties} property.
   * @see #BacnetProperties
   */
  public void setBacnetProperties(BHonPlantControllerBacnetProperties v) { set(BacnetProperties, v, null); }

////////////////////////////////////////////////////////////////
// Property "home"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code home} property.
   * @see #getHome
   * @see #setHome
   */
  public static final Property home = newProperty(0, new BHonPlantControllerHome(), null);
  
  /**
   * Get the {@code home} property.
   * @see #home
   */
  public BHonPlantControllerHome getHome() { return (BHonPlantControllerHome)get(home); }
  
  /**
   * Set the {@code home} property.
   * @see #home
   */
  public void setHome(BHonPlantControllerHome v) { set(home, v, null); }

////////////////////////////////////////////////////////////////
// Property "hmiAlarmHandler"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code hmiAlarmHandler} property.
   * @see #getHmiAlarmHandler
   * @see #setHmiAlarmHandler
   */
  public static final Property hmiAlarmHandler = newProperty(Flags.HIDDEN | Flags.READONLY | Flags.TRANSIENT, new BHMIAlarmHandler(), null);
  
  /**
   * Get the {@code hmiAlarmHandler} property.
   * @see #hmiAlarmHandler
   */
  public BHMIAlarmHandler getHmiAlarmHandler() { return (BHMIAlarmHandler)get(hmiAlarmHandler); }
  
  /**
   * Set the {@code hmiAlarmHandler} property.
   * @see #hmiAlarmHandler
   */
  public void setHmiAlarmHandler(BHMIAlarmHandler v) { set(hmiAlarmHandler, v, null); }

////////////////////////////////////////////////////////////////
// Property "hmiPointModeListener"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code hmiPointModeListener} property.
   * @see #getHmiPointModeListener
   * @see #setHmiPointModeListener
   */
  public static final Property hmiPointModeListener = newProperty(Flags.HIDDEN | Flags.READONLY | Flags.TRANSIENT, new BHonPlantControllerHMIPointModeHandler(), null);
  
  /**
   * Get the {@code hmiPointModeListener} property.
   * @see #hmiPointModeListener
   */
  public BHonPlantControllerHMIPointModeHandler getHmiPointModeListener() { return (BHonPlantControllerHMIPointModeHandler)get(hmiPointModeListener); }
  
  /**
   * Set the {@code hmiPointModeListener} property.
   * @see #hmiPointModeListener
   */
  public void setHmiPointModeListener(BHonPlantControllerHMIPointModeHandler v) { set(hmiPointModeListener, v, null); }

////////////////////////////////////////////////////////////////
// Property "supportHmiFirmwareDowngrade"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code supportHmiFirmwareDowngrade} property.
   * @see #getSupportHmiFirmwareDowngrade
   * @see #setSupportHmiFirmwareDowngrade
   */
  public static final Property supportHmiFirmwareDowngrade = newProperty(Flags.HIDDEN, false, null);
  
  /**
   * Get the {@code supportHmiFirmwareDowngrade} property.
   * @see #supportHmiFirmwareDowngrade
   */
  public boolean getSupportHmiFirmwareDowngrade() { return getBoolean(supportHmiFirmwareDowngrade); }
  
  /**
   * Set the {@code supportHmiFirmwareDowngrade} property.
   * @see #supportHmiFirmwareDowngrade
   */
  public void setSupportHmiFirmwareDowngrade(boolean v) { setBoolean(supportHmiFirmwareDowngrade, v, null); }

////////////////////////////////////////////////////////////////
// Property "hmiInitializationListener"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code hmiInitializationListener} property.
   * @see #getHmiInitializationListener
   * @see #setHmiInitializationListener
   */
  public static final Property hmiInitializationListener = newProperty(Flags.HIDDEN | Flags.READONLY | Flags.TRANSIENT, new BHonPlantControllerHMIInitializationHandler(), null);
  
  /**
   * Get the {@code hmiInitializationListener} property.
   * @see #hmiInitializationListener
   */
  public BHonPlantControllerHMIInitializationHandler getHmiInitializationListener() { return (BHonPlantControllerHMIInitializationHandler)get(hmiInitializationListener); }
  
  /**
   * Set the {@code hmiInitializationListener} property.
   * @see #hmiInitializationListener
   */
  public void setHmiInitializationListener(BHonPlantControllerHMIInitializationHandler v) { set(hmiInitializationListener, v, null); }

////////////////////////////////////////////////////////////////
// Property "falSyncStatus"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code falSyncStatus} property.
   * @see #getFalSyncStatus
   * @see #setFalSyncStatus
   */
  public static final Property falSyncStatus = newProperty(Flags.READONLY, BHMISyncStatus.DEFAULT, null);
  
  /**
   * Get the {@code falSyncStatus} property.
   * @see #falSyncStatus
   */
  public BHMISyncStatus getFalSyncStatus() { return (BHMISyncStatus)get(falSyncStatus); }
  
  /**
   * Set the {@code falSyncStatus} property.
   * @see #falSyncStatus
   */
  public void setFalSyncStatus(BHMISyncStatus v) { set(falSyncStatus, v, null); }

////////////////////////////////////////////////////////////////
// Property "deviceCertificate"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code deviceCertificate} property.
   * @see #getDeviceCertificate
   * @see #setDeviceCertificate
   */
  public static final Property deviceCertificate = newProperty(Flags.HIDDEN | Flags.READONLY, BDynamicEnum.DEFAULT, null);
  
  /**
   * Get the {@code deviceCertificate} property.
   * @see #deviceCertificate
   */
  public BDynamicEnum getDeviceCertificate() { return (BDynamicEnum)get(deviceCertificate); }
  
  /**
   * Set the {@code deviceCertificate} property.
   * @see #deviceCertificate
   */
  public void setDeviceCertificate(BDynamicEnum v) { set(deviceCertificate, v, null); }

////////////////////////////////////////////////////////////////
// Property "pointName"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code pointName} property.
   * @see #getPointName
   * @see #setPointName
   */
  public static final Property pointName = newProperty(0, BFormat.make("%parent.displayName%-%displayName%"), null);
  
  /**
   * Get the {@code pointName} property.
   * @see #pointName
   */
  public BFormat getPointName() { return (BFormat)get(pointName); }
  
  /**
   * Set the {@code pointName} property.
   * @see #pointName
   */
  public void setPointName(BFormat v) { set(pointName, v, null); }

////////////////////////////////////////////////////////////////
// Property "scheduleName"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code scheduleName} property.
   * @see #getScheduleName
   * @see #setScheduleName
   */
  public static final Property scheduleName = newProperty(0, BFormat.make("%parent.displayName%-%displayName%"), null);
  
  /**
   * Get the {@code scheduleName} property.
   * @see #scheduleName
   */
  public BFormat getScheduleName() { return (BFormat)get(scheduleName); }
  
  /**
   * Set the {@code scheduleName} property.
   * @see #scheduleName
   */
  public void setScheduleName(BFormat v) { set(scheduleName, v, null); }

////////////////////////////////////////////////////////////////
// Property "calendarName"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code calendarName} property.
   * @see #getCalendarName
   * @see #setCalendarName
   */
  public static final Property calendarName = newProperty(0, BFormat.make("%parent.displayName%-%displayName%"), null);
  
  /**
   * Get the {@code calendarName} property.
   * @see #calendarName
   */
  public BFormat getCalendarName() { return (BFormat)get(calendarName); }
  
  /**
   * Set the {@code calendarName} property.
   * @see #calendarName
   */
  public void setCalendarName(BFormat v) { set(calendarName, v, null); }

////////////////////////////////////////////////////////////////
// Property "hmiAlarmDbLimit"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code hmiAlarmDbLimit} property.
   * @see #getHmiAlarmDbLimit
   * @see #setHmiAlarmDbLimit
   */
  public static final Property hmiAlarmDbLimit = newProperty(0, 100, BFacets.make(BFacets.make(BFacets.make("units", BUnit.make("records", BDimension.DEFAULT)), BFacets.make("min", 1)), BFacets.make("max", 1500)));
  
  /**
   * Get the {@code hmiAlarmDbLimit} property.
   * @see #hmiAlarmDbLimit
   */
  public int getHmiAlarmDbLimit() { return getInt(hmiAlarmDbLimit); }
  
  /**
   * Set the {@code hmiAlarmDbLimit} property.
   * @see #hmiAlarmDbLimit
   */
  public void setHmiAlarmDbLimit(int v) { setInt(hmiAlarmDbLimit, v, null); }

////////////////////////////////////////////////////////////////
// Property "alarmPriorityConfiguration"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code alarmPriorityConfiguration} property.
   * @see #getAlarmPriorityConfiguration
   * @see #setAlarmPriorityConfiguration
   */
  public static final Property alarmPriorityConfiguration = newProperty(0, new BAlarmPriorityConfig(), null);
  
  /**
   * Get the {@code alarmPriorityConfiguration} property.
   * @see #alarmPriorityConfiguration
   */
  public BAlarmPriorityConfig getAlarmPriorityConfiguration() { return (BAlarmPriorityConfig)get(alarmPriorityConfiguration); }
  
  /**
   * Set the {@code alarmPriorityConfiguration} property.
   * @see #alarmPriorityConfiguration
   */
  public void setAlarmPriorityConfiguration(BAlarmPriorityConfig v) { set(alarmPriorityConfiguration, v, null); }

////////////////////////////////////////////////////////////////
// Property "pointExportOption"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code pointExportOption} property.
   * @see #getPointExportOption
   * @see #setPointExportOption
   */
  public static final Property pointExportOption = newProperty(0, BHMIPointExportOptions.DEFAULT, null);
  
  /**
   * Get the {@code pointExportOption} property.
   * @see #pointExportOption
   */
  public BHMIPointExportOptions getPointExportOption() { return (BHMIPointExportOptions)get(pointExportOption); }
  
  /**
   * Set the {@code pointExportOption} property.
   * @see #pointExportOption
   */
  public void setPointExportOption(BHMIPointExportOptions v) { set(pointExportOption, v, null); }

////////////////////////////////////////////////////////////////
// Property "isHmiJobActive"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code isHmiJobActive} property.
   * @see #getIsHmiJobActive
   * @see #setIsHmiJobActive
   */
  public static final Property isHmiJobActive = newProperty(Flags.READONLY | Flags.HIDDEN | Flags.TRANSIENT, false, null);
  
  /**
   * Get the {@code isHmiJobActive} property.
   * @see #isHmiJobActive
   */
  public boolean getIsHmiJobActive() { return getBoolean(isHmiJobActive); }
  
  /**
   * Set the {@code isHmiJobActive} property.
   * @see #isHmiJobActive
   */
  public void setIsHmiJobActive(boolean v) { setBoolean(isHmiJobActive, v, null); }

////////////////////////////////////////////////////////////////
// Action "commissionHMI"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code commissionHMI} action.
   * @see #commissionHMI()
   */
  public static final Action commissionHMI = newAction(Flags.HIDDEN, null);
  
  /**
   * Invoke the {@code commissionHMI} action.
   * @see #commissionHMI
   */
  public void commissionHMI() { invoke(commissionHMI, null, null); }

////////////////////////////////////////////////////////////////
// Action "notifyChangesToHMI"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code notifyChangesToHMI} action.
   * @see #notifyChangesToHMI()
   */
  public static final Action notifyChangesToHMI = newAction(Flags.HIDDEN, null);
  
  /**
   * Invoke the {@code notifyChangesToHMI} action.
   * @see #notifyChangesToHMI
   */
  public void notifyChangesToHMI() { invoke(notifyChangesToHMI, null, null); }

////////////////////////////////////////////////////////////////
// Action "firmwareUpdateHMI"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code firmwareUpdateHMI} action.
   * @see #firmwareUpdateHMI(BString parameter)
   */
  public static final Action firmwareUpdateHMI = newAction(Flags.ASYNC, BString.DEFAULT, null);
  
  /**
   * Invoke the {@code firmwareUpdateHMI} action.
   * @see #firmwareUpdateHMI
   */
  public void firmwareUpdateHMI(BString parameter) { invoke(firmwareUpdateHMI, parameter, null); }

////////////////////////////////////////////////////////////////
// Action "getLanguageFileHMI"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code getLanguageFileHMI} action.
   * @see #getLanguageFileHMI()
   */
  public static final Action getLanguageFileHMI = newAction(Flags.ASYNC | Flags.HIDDEN, null);
  
  /**
   * Invoke the {@code getLanguageFileHMI} action.
   * @see #getLanguageFileHMI
   */
  public void getLanguageFileHMI() { invoke(getLanguageFileHMI, null, null); }

////////////////////////////////////////////////////////////////
// Action "languageFileHMI"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code languageFileHMI} action.
   * @see #languageFileHMI(BString parameter)
   */
  public static final Action languageFileHMI = newAction(Flags.ASYNC | Flags.HIDDEN, BString.DEFAULT, null);
  
  /**
   * Invoke the {@code languageFileHMI} action.
   * @see #languageFileHMI
   */
  public void languageFileHMI(BString parameter) { invoke(languageFileHMI, parameter, null); }

////////////////////////////////////////////////////////////////
// Action "isHMIPresent"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code isHMIPresent} action.
   * @see #isHMIPresent()
   */
  public static final Action isHMIPresent = newAction(Flags.HIDDEN, null);
  
  /**
   * Invoke the {@code isHMIPresent} action.
   * @see #isHMIPresent
   */
  public BBoolean isHMIPresent() { return (BBoolean)invoke(isHMIPresent, null, null); }

////////////////////////////////////////////////////////////////
// Action "initiateFileCopy"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code initiateFileCopy} action.
   * @see #initiateFileCopy(BString parameter)
   */
  public static final Action initiateFileCopy = newAction(Flags.HIDDEN, BString.DEFAULT, null);
  
  /**
   * Invoke the {@code initiateFileCopy} action.
   * @see #initiateFileCopy
   */
  public BString initiateFileCopy(BString parameter) { return (BString)invoke(initiateFileCopy, parameter, null); }

////////////////////////////////////////////////////////////////
// Action "copyBinaryString"
////////////////////////////////////////////////////////////////
  
  /**
   * Slot for the {@code copyBinaryString} action.
   * @see #copyBinaryString(BHMIFirmwareDownloadConfig parameter)
   */
  public static final Action copyBinaryString = newAction(Flags.HIDDEN, BHMIFirmwareDownloadConfig.make(), null);
  
  /**
   * Invoke the {@code copyBinaryString} action.
   * @see #copyBinaryString
   */
  public BString copyBinaryString(BHMIFirmwareDownloadConfig parameter) { return (BString)invoke(copyBinaryString, parameter, null); }

////////////////////////////////////////////////////////////////
// Type
////////////////////////////////////////////////////////////////
  
  @Override
  public Type getType() { return TYPE; }
  public static final Type TYPE = Sys.loadType(BHonPlantControllerHMI.class);

/*+ ------------ END BAJA AUTO GENERATED CODE -------------- +*/

  public BHonPlantControllerHMI() {
    //
  }

  @Override
  public boolean isParentLegal(BComponent parent) {
    // only when on station. Workbench does not care
    if (!Sys.isStation()) {
      return true;
    }

    if (!(parent.getParent() instanceof BServiceContainer) ||
        (parent.getChildren(BHonPlantControllerHMI.class).length > 0)) {
      return false;
    }

    return super.isParentLegal(parent);
  }

  ///CLOVER:OFF
  @Override
  public  void started() throws Exception {
	if(getEnabled() && !BHonPlantControllerHMI.isOnBeatsAdvController()) {
		this.setEnabled(false);
	}
    BDynamicEnum certAliases = getAvailableCertificates();
    setDeviceCertificate(certAliases);
  }

  @Override
	public void stopped() throws Exception {
    super.stopped();
    if (hmiCommHandler != null) {
      hmiCommHandler.unregisterListeners();
      HMIIAmListener hmiIAmListener = hmiCommHandler.getHmiIAmListener();
      if (hmiIAmListener != null) {
        hmiIAmListener.stopThreadVerifyFirmware();
      }
    }

    if (subscriber != null) {
      //unsubscribe
      subscriber.unsubscribe();
    }
  }
  ///CLOVER:ON

  @Override
	public void atSteadyState() throws Exception {
		super.atSteadyState();
		if (!HonPlantControllerBacnetUtils.containsHMIPort(getBacnetNetwork(), getBacnetProperties())) {
			HonPlantControllerHMILogger.log(Level.WARNING, "Not proceeding with registering the listeners as the station doesn't have the HMI port configured.");
			return;
		}
		registerListeners();
		notifyChangesToHMI();
		getHmiInitializationListener().setKeyExchangeState(KeyExchangeState.Idle_State);
	}
  
	public static boolean isOnBeatsAdvController() {
		final String LOG_MSG_PREFIX = "BHonPLantControllerService.isOnBeatsAdvController";
		String os = System.getProperty("os.name").toUpperCase().trim();
		boolean ret = false;

		// compare with known OS name. LINUX = BeatsAdv
		if (os.equals("LINUX")) {
			HonPlantControllerHMILogger.log(Level.FINEST, MessageFormat.format("{0} running on a beats adv controller", LOG_MSG_PREFIX));
			ret = true;
		} else {
			HonPlantControllerHMILogger.log(Level.FINEST, MessageFormat.format("{0} not running on a beats adv controller", LOG_MSG_PREFIX));
		}
		return ret;
	}

  private void registerListeners() {
    logger.finest("[BHonPlantControllerHMI.registerListeners]");

    if (HonPlantControllerBacnetUtils.containsHMIPort(getBacnetNetwork(), getBacnetProperties())) {
      if (hmiCommHandler == null) {
        hmiCommHandler = new BHMICommHandler(getBacnetNetwork(), this);
      }
      hmiCommHandler.registerListeners();
      getHmiInitializationListener().setConfirmedPrivateTransferSender(hmiCommHandler);
      getHmiPointModeListener().setUnconfirmedPrivateTransferSender(hmiCommHandler);
      getHmiPointModeListener().startCollect();
      getHmiInitializationListener().setKeyExchangeState(KeyExchangeState.Idle_State);
    }

    logger.finest("[BHonPlantControllerHMI.registerListeners] initializeSubscriber");
    initializeSubscriber();
  }

  public void initializeSubscriber() {
    if (null == subscriber) {
      subscriber = new BHonPlantControllerHMISubscriber(this);
    } else {
      subscriber.unsubscribe();
    }
    subscriber.initializeSubscriber();
    subscriber.subscribe();
  }

  @Override
  public void changed(Property property, Context context) {

    super.changed(property, context);

    if (!isRunning()) {
      return;
    }

    if (property == deviceCert) {
      establishDeviceCert();
    } else if (property.getName().equals(HMI_PROP_NAME_FW_SOURCE)) {
	} else if (property == enabled) {
		if (!getEnabled()) {
			notifyChangesToHMI();
			setFlags(commissionHMI, getFlags(commissionHMI) | Flags.HIDDEN);
			setFlags(notifyChangesToHMI, getFlags(notifyChangesToHMI) | Flags.HIDDEN);
			setFlags(getLanguageFileHMI, getFlags(getLanguageFileHMI) | Flags.HIDDEN);
			if (getBacnetNetwork() != null) {
				removeHMIDeviceEntryFromBacnet();
			}
		} else {
			setFlags(commissionHMI, getFlags(commissionHMI) & ~Flags.HIDDEN);
			setFlags(notifyChangesToHMI, getFlags(notifyChangesToHMI) & ~Flags.HIDDEN);
			setFlags(getLanguageFileHMI, getFlags(getLanguageFileHMI) & ~Flags.HIDDEN);
			validateAndSetHMIInst();
		}
	}
 	
  }

	public void validateAndSetHMIInst() {
		BBacnetNetwork network = getBacnetNetwork();
		if (null == network) {
			HonPlantControllerHMILogger.checkAndLog(Level.WARNING,
					"can not set BacnetSettings on a null BacnetNetwork");
			return;
		}
		getBacnetProperties().validateAndThrowExceptionIfinstanceIdIsDuplicate(getBacnetProperties().getHMIInstanceId().getInstanceNumber());
		if(null != getHMICommhandler()) {
			getHMICommhandler().sendInstanceidToHMI(getBacnetProperties().getHMIInstanceId(), this, false);
		}
		if (!getBacnetProperties().isHmiInstanceChanged()) {
			addHMIEntryInBacnetNetwork();
		}
	}

	public void addHMIEntryInBacnetNetwork() {
		BBacnetNetwork network = getBacnetNetwork();
		if (null == network) {
			return;
		}
		String hMIDevPresentInBacnet = getHMIDeviceNamePresentInBacnetNetwork();
		BHMIDevice hmiDeviceNw = new BHMIDevice();
		if (getEnabled() && null==hMIDevPresentInBacnet) {
			BHonPlantControllerBacnetProperties bacnetProperties = getBacnetProperties();
			bacnetProperties.validateAndThrowExceptionIfinstanceIdIsDuplicate(bacnetProperties.getHMIInstanceId().getInstanceNumber());
			hmiDeviceNw.getConfig().getDeviceObject().setObjectId((bacnetProperties.getHMIInstanceId()));
			network.add(hmiDeviceNw.getHmiDeviceName(), hmiDeviceNw);
		}
		if (getEnabled() && null!=hMIDevPresentInBacnet) {
			BValue deValue = network.get(hMIDevPresentInBacnet);
			BHMIDevice device = (BHMIDevice) deValue;
			device.getConfig().getDeviceObject().setObjectId((getBacnetProperties().getHMIInstanceId()));
		}

	}

	private String getHMIDeviceNamePresentInBacnetNetwork() {
		String deviceNameString = null;
		BComponent[] devices = getBacnetNetwork().getChildren(BHMIDevice.class);
		if (devices.length > 0) {
			return devices[0].getName();
		}

		return deviceNameString;
	}

	public void removeHMIDeviceEntryFromBacnet() {
		String deviceNameString = getHMIDeviceNamePresentInBacnetNetwork();
		if (null != deviceNameString) {
			getBacnetNetwork().remove(deviceNameString);
		}
	}

  public boolean getHMIFirmwareFromFileSystem() {
    BValue value = get(HMI_PROP_NAME_FW_SOURCE);
    return (value instanceof BBoolean) && ((BBoolean)value).getBoolean();
  }

  public void handleServiceEnabled(String lexiconModuleName, String lexiconKey) {
	  if(!getEnabled()) {
		  throw new LocalizableRuntimeException(lexiconModuleName, lexiconKey);
	  }
  }
  public void doCommissionHMI() {
	  handleServiceEnabled(TYPE.getModule().getModuleName(), LEX_MSG_HMI);
    synchronized (syncObject) {
      BBaseHMIJob job = new BCommissionHMIJob(this);
      job.submit(Context.NULL);
    }
  }

  public void doNotifyChangesToHMI() {
	  handleServiceEnabled(TYPE.getModule().getModuleName(), LEX_MSG_HMI);
    synchronized (syncObject) {
      new Thread() {
        public void run() {
          logger.info(HonPlantControllerHMILexicon.getLexicon().getText("hmi.commission.syncstatus.notify.data.changed"));
          try {
            getHMICommhandler().sendDynamicDataSyncRequest(0);
          } catch (BacnetException e) {
        	  HonPlantControllerHMILogger.finest(e.getMessage(), e);
          }
        }
      }.start();
    }
  }

	public void establishDeviceCert() {
		// HMI -> CONTROLLER:
		// BACNET_PRIVATE_TRANSFER: request CPU Certificate
		BDynamicEnum certificateValue = getDeviceCertificate();
		BFacets enumFacet = certificateValue.getEnumFacets();
		BEnumRange enumRange = (BEnumRange) enumFacet.get(BFacets.RANGE);
		final String certAlias = SlotPath.unescape(enumRange.getTag(BEATS_CERT_ORIDINAL));
		Optional<BCertManagerService> service = getCertManagerService();
		if (!service.isPresent()) {
			HonPlantControllerHMILogger.warning(MessageFormat.format(
					"BHonPlantControllerHMIService.establishDeviceCert(): No manager found for Cert: {0}", certAlias));
			return;
		}
		try {
			ICryptoManager certMgr = CertManagerFactory.getInstance();
			ITrustStore sysStore = certMgr.getSystemTrustStore();
			deviceCert = getCertFromSystemTrustStore(certAlias);
			ECPrivateKey privateKey = null;
			if (sysStore instanceof NTrustStore) {

				java.security.Key key = AccessController.doPrivileged((PrivilegedAction<java.security.Key>) () -> {
					try {
						KeyStore kStore = ((NTrustStore) sysStore).getKeyStore();
						return kStore.getKey(certAlias, KEY_STORE_KEY.toCharArray());
					} catch (Exception e) {
						HonPlantControllerHMILogger.finest("could not get the key from the keystore", e);
					}
					return null;
				});
				if (key != null && key instanceof ECPrivateKey) {
					privateKey = (ECPrivateKey) key;
				}
			}
			if (deviceCert != null && privateKey != null) {
				HonPlantControllerHMILogger.finest("BHonPlantControllerHMIService.establish: Exchange");
				getHmiInitializationListener().setPrivKeyController(privateKey);
				getHmiInitializationListener().setPubKeyController((ECPublicKey) deviceCert.getPublicKey());
			} else {
				HonPlantControllerHMILogger.finest(
						"The {0} Certificate not found in the system trust store or private key is null", certAlias);
			}
		} catch (Exception e) {
			HonPlantControllerHMILogger.finest("Could not get entry for alias", e);
		}
	}

  public void doStartKeyExchange() throws Exception {
	HonPlantControllerHMILogger.finest("BHonPlantControllerHMIService.Exchange: Starting Exchange");
    establishDeviceCert();

    //Request Cert from HMI
    getHmiInitializationListener().keyExchange();
   }

  public X509Certificate getCertFromSystemTrustStore(String certAlias) throws Exception {
    X509Certificate cert = null;

    Optional<BCertManagerService> service = getCertManagerService();
	if (service.isPresent()) {
		ICryptoManager certMgr = CertManagerFactory.getInstance();
		ITrustStore sysStore = certMgr.getSystemTrustStore();
		if (sysStore.containsAlias(certAlias)) {
			cert = sysStore.getCertificate(certAlias);
		}else {
			HonPlantControllerHMILogger.finest("could not find certificate : {0}",certAlias);
		}
	}
    return cert;
  }

	private BDynamicEnum getAvailableCertificates() throws Exception {
		List<String> tags = new ArrayList<>();
		tags.add(SlotPath.escape(HONEYWELL_BEATS_CERT));
		tags.add(SlotPath.escape(HONEYWELL_CA));
		BEnumRange range = BEnumRange.make(tags.toArray(new String[0]));
		return BDynamicEnum.make(range);
	}

  public void doFirmwareUpdateHMI(BString filename) {
	  handleServiceEnabled(TYPE.getModule().getModuleName(), LEX_MSG_HMI);
	    synchronized (syncObject) {
      new BHMIFileDownloadJob(this, filename).submit(null);
    }
  }

  public void doGetLanguageFileHMI() {
	  handleServiceEnabled(TYPE.getModule().getModuleName(), LEX_MSG_HMI);
    // do nothing. All is handled in work bench
  }

  public void doLanguageFileHMI(BString filename) {
	  handleServiceEnabled(TYPE.getModule().getModuleName(), LEX_MSG_HMI);
    synchronized (syncObject) {
      new BHMIFileDownloadJob(this, filename).submit(null);
    }
  }

  // please note: language can be null
  public boolean sendLanguageFile(String language) {
    char langId;
    if ("chn".equals(language)) {
      langId = 'C';
    } else if ("kor".equals(language)) {
      langId = 'K';
    } else {
      // not supported language
      return false;
    }

    Path firmwarePath = Sys.getStationHome().toPath().resolve(FIRMWARE_FOLDER_NAME);
    File firmwareFolder = firmwarePath.toFile();
    if (!firmwareFolder.isDirectory()) {
      // missing or invalid folder
      return false;
    }
    File[] files = firmwareFolder.listFiles();
    if (files == null) {
      // empty folder
      return false;
    }

    // get the files which supports the requested language
    List<LangFileInfo> langFiles = new ArrayList<>();
    for (File file : files) {
      String filename = file.getName();
      int lastDot = filename.lastIndexOf('.');
      if ((lastDot > 0) && filename.substring(lastDot).equals(HMI_LANGUAGE_FILE_EXTENSION)) {
        final String LANG_FILE_IDENT = "-App_v";
        int indexIdentApp = filename.indexOf(LANG_FILE_IDENT);
        if (indexIdentApp >= 0) {
          String languageVersion = filename.substring(indexIdentApp + LANG_FILE_IDENT.length(), lastDot);
          char lastChar = languageVersion.charAt(languageVersion.length() - 1);
          if (lastChar == langId) {
            String versStr = languageVersion.substring(0, languageVersion.length() - 1);
            Version version = new Version(versStr);
            langFiles.add(new LangFileInfo(filename, version));
          }
        }
      }
    }
    if (langFiles.isEmpty()) {
      // no file found which supports the requested language
      return false;
    }

    // get the language file with the highest version
    LangFileInfo fileInfo = langFiles.get(0);
    for (int index = 1; index < langFiles.size(); index++) {
      if (fileInfo.version.compareTo(langFiles.get(index).version) < 0) {
        fileInfo = langFiles.get(index);
      }
    }

    // send the file to the HMI
    // Use async action to send the response of RequestLanguageFile before the start of the download job
    languageFileHMI(BString.make(fileInfo.filename));

    return true;
  }
  
	public boolean isHMIDevice(BBacnetObjectIdentifier hmiObjectId,BBacnetAddress bacnetAddress) {
		final String LOG_MSG_PREFIX = TYPE.getTypeName() + ".isHMIDevice";
		boolean isHMIDevice = false;
		HMICommCheck hmiCommCheck;
		try {
			hmiCommCheck = new HMICommCheck(LOG_MSG_PREFIX);
		} catch (LocalizableException exc) {
			// the message is shown in HMICommCheck
			HonPlantControllerHMILogger.log(Level.SEVERE,
					MessageFormat.format("{0}: Exception occurred while checking for HMI device. {1}:{2}", LOG_MSG_PREFIX, exc.getMessage(), exc));
			return false;
		}

		BBacnetClientLayer clientLayer = hmiCommCheck.clientLayer;
		BBacnetAddress hmiAddress = bacnetAddress;
	
		try {
			String modelName = ((BString) BHMICommHandler.readProperty(clientLayer, hmiObjectId, hmiAddress,
					BBacnetPropertyIdentifier.MODEL_NAME, BacnetConst.ASN_CHARACTER_STRING, BString.DEFAULT))
							.getString();
			if (modelName.equalsIgnoreCase("HMI-DN") || modelName.equalsIgnoreCase("HMI-WL") || modelName.equalsIgnoreCase("CPO-MMI")) {
				isHMIDevice = true;
			}
		} catch (BacnetException exc) {

			HonPlantControllerHMILogger.log(Level.WARNING,
					MessageFormat.format("{0} problem at ''{1}''", LOG_MSG_PREFIX, "read model name"), exc);
		}
		
		return isHMIDevice;
	}

  public void verifyFirmware() {
	 
    final String LOG_MSG_PREFIX = TYPE.getTypeName() + ".checkForFwUpdate";
     if (getHMIFirmwareFromFileSystem()) {
      HonPlantControllerHMILogger.finest("{0} get file from station", LOG_MSG_PREFIX);
      return;
    }
	if (isFirmwareDownloadCancelled()) {
		HonPlantControllerHMILogger.info(
				"{0} : Firmware job has been already cancelled. Firmware job will not be triggered until done manually or after station restart",
				LOG_MSG_PREFIX);
		return;
	}

    HMICommCheck hmiCommCheck;
    try {
      hmiCommCheck = new HMICommCheck(LOG_MSG_PREFIX);
    } catch (LocalizableException exc) {
      // the message is shown in HMICommCheck
      return;
    }

    BBacnetClientLayer clientLayer = hmiCommCheck.clientLayer;
    BBacnetObjectIdentifier hmiObjectId = hmiCommCheck.hmiObjectId;
    BBacnetAddress hmiAddress = hmiCommCheck.hmiAddress;

    while (true) {
      try {
        Thread.sleep(1000);
      } catch (InterruptedException exc) {
        Thread.currentThread().interrupt();
        return;
      }
     String action = "read system status";
      try {
        BBacnetDeviceStatus systemStatus;
        systemStatus = (BBacnetDeviceStatus) BHMICommHandler.readProperty(clientLayer, hmiObjectId, hmiAddress,
            BBacnetPropertyIdentifier.SYSTEM_STATUS,
            BacnetConst.ASN_ENUMERATED,
            BBacnetDeviceStatus.DEFAULT);
        HonPlantControllerHMILogger.finest(MessageFormat.format("{0} systemStatus[{1}]", LOG_MSG_PREFIX, systemStatus));
        if (systemStatus.getOrdinal() != BBacnetDeviceStatus.OPERATIONAL) {
          continue;
        }

        action = "read model name";
        String modelName = ((BString) BHMICommHandler.readProperty(clientLayer, hmiObjectId, hmiAddress,
            BBacnetPropertyIdentifier.MODEL_NAME,
            BacnetConst.ASN_CHARACTER_STRING,
            BString.DEFAULT)).getString();
        HonPlantControllerHMILogger.finest(MessageFormat.format("{0} modelName[{1}]", LOG_MSG_PREFIX, modelName));
        if (modelName.isEmpty()) {
          HonPlantControllerHMILogger.warning(MessageFormat.format("{0} HMI not supported, model name: ''{1}''",
              LOG_MSG_PREFIX, modelName));
          return;
        }

        action = "read firmware revision";
        String hmiRevision = ((BString) BHMICommHandler.readProperty(clientLayer, hmiObjectId, hmiAddress,
            BBacnetPropertyIdentifier.FIRMWARE_REVISION,
            BacnetConst.ASN_CHARACTER_STRING,
            BString.DEFAULT)).getString();
        HonPlantControllerHMILogger.info(MessageFormat.format("{0} hmiRevision[{1}]", LOG_MSG_PREFIX, hmiRevision));

        HonPlantControllerHMILogger.info(MessageFormat.format("{0} HMI is ready for update of the firmware",
            LOG_MSG_PREFIX));

        action = "read firmware file from module";
        String imageFileName = getFwFileNameFromModule();
        if (imageFileName == null) {
          HonPlantControllerHMILogger.warning(MessageFormat.format("{0} HMI firmware file not available",
              LOG_MSG_PREFIX));
          return;
        }
        byte[] fwImage = readFileFromModule(imageFileName);
        byte[] fwVersionData = new byte[HMI_HEADER_IDENT_MAX_LENGTH];
        System.arraycopy(fwImage, HMI_HEADER_IDENT_OFFSET, fwVersionData, 0, HMI_HEADER_IDENT_MAX_LENGTH);

        action = "check firmware version";
        String fwIdent = getFileIdent(fwVersionData);
        if (fwIdent.equals(hmiRevision)) {
          HonPlantControllerHMILogger.info(MessageFormat.format("{0} HMI contains already firmware version ''{1}''",
              LOG_MSG_PREFIX, fwIdent));
          return;
        } else if (isRevisionOlder(fwIdent, hmiRevision)) {
          HonPlantControllerHMILogger.info(MessageFormat.format("{0} HMI firmware file (version: ''{1}'') is older " +
                  "then the one in the HMI (version: ''{2}'')",
              LOG_MSG_PREFIX, fwIdent, hmiRevision));
          return;
        }

        action = "invoke firmwareUpdateHMI";
        firmwareUpdateHMI(BString.DEFAULT);
        return;
      } catch (Exception exc) {
        if (Thread.currentThread().isInterrupted()) {
          HonPlantControllerHMILogger.log(Level.SEVERE,
              MessageFormat.format("{0} interrupted at ''{1}''", LOG_MSG_PREFIX, action),
              exc);
          return;
        }
        HonPlantControllerHMILogger.log(Level.WARNING,
            MessageFormat.format("{0} problem at ''{1}''", LOG_MSG_PREFIX, action), exc);
      }
    }
   
  }

  public boolean runFileDownloadHMI(BString filename, BSimpleJob job)
      throws Exception {
    final String LOG_MSG_PREFIX = getType().getTypeName() + ".runFileDownloadHMI";

    boolean isFwUpdate = isFwUpdate(filename);

    String identFileType;
    String identWork;
    String lexKeyAlreadyRunning;
    String lexKeyWrongModel;
    String lexKeyAlreadyLoaded;
    String lexKeyStartFileDownload;
    String lexKeyEndFileDownload;
    String lexKeyFileDownloadCancelled;
    if (isFwUpdate) {
      identFileType = "firmware";
      identWork = "firmware update";
      lexKeyAlreadyRunning = "hmi.downloadFirmware.already.running";
      lexKeyWrongModel = "hmi.downloadFirmware.file.wrong.model";
      lexKeyAlreadyLoaded = "hmi.downloadFirmware.already.loaded";
      lexKeyStartFileDownload = "hmi.downloadFirmware.start.download";
      lexKeyEndFileDownload = "hmi.downloadFirmware.end.download";
      lexKeyFileDownloadCancelled = "hmi.downloadFirmware.cancel.download";
    } else {
      identFileType = "language";
      identWork = "language download";
      lexKeyAlreadyRunning = "hmi.downloadLanguage.already.running";
      lexKeyWrongModel = "hmi.downloadLanguage.file.wrong.model";
      lexKeyAlreadyLoaded = "hmi.downloadLanguage.already.loaded";
      lexKeyStartFileDownload = "hmi.downloadLanguage.start.download";
      lexKeyEndFileDownload = "hmi.downloadLanguage.end.download";
      lexKeyFileDownloadCancelled = "hmi.downloadLanguage.cancel.download";
    }

    if (hmiFileUpdateIsRunning) {
      HonPlantControllerHMILogger.warning("HMI " + identWork + " is already running");
      throw new LocalizableException(LEX, lexKeyAlreadyRunning);
    }

    Path filePath = null;
    String imageFileName = null;
    if (isFwUpdate && BString.DEFAULT.equals(filename)) {
      // get firmware file from Niagara module
      imageFileName = getFwFileNameFromModule();
      if (imageFileName == null) {
        HonPlantControllerHMILogger.warning("HMI " + identWork + " no file available");
        throw new LocalizableException(LEX, "hmi.downloadFirmware.file.not.available");
    }
    } else {
      // get file from file system of the station
      Path firmwarePath = Sys.getStationHome().toPath().resolve(FIRMWARE_FOLDER_NAME);
      filePath = firmwarePath.resolve(filename.getString());
      imageFileName = filePath.toString();
    }

    HonPlantControllerHMILogger.info(MessageFormat.format("{0} filename[{1}] imageFileName[{2}]", LOG_MSG_PREFIX,
                                                          filename, imageFileName));

    HMICommCheck hmiCommCheck = new HMICommCheck(LOG_MSG_PREFIX);
    HMIIAmListener hmiIAmListener = hmiCommCheck.hmiIAmListener;

    if (job != null) {
      job.progress(job.getProgress() + 1);
    }

    try {
      hmiFileUpdateIsRunning = true;
      setIsHmiJobActive(true);
      byte[] fwImage;
      if (filePath == null) {
        // get file from Niagara module
        fwImage = readFileFromModule(imageFileName);
      } else {
        // get file from file system of the station
        fwImage = Files.readAllBytes(filePath);
      }
      byte[] fwVersionData = new byte[HMI_HEADER_IDENT_MAX_LENGTH];
      System.arraycopy(fwImage, HMI_HEADER_IDENT_OFFSET, fwVersionData, 0, HMI_HEADER_IDENT_MAX_LENGTH);
      String fwIdent = getFileIdent(fwVersionData);
      HonPlantControllerHMILogger.info(MessageFormat.format("{0} {1} image length[{2,number,0}] revision[{3}]",
          LOG_MSG_PREFIX, identWork, fwImage.length, fwIdent));

      // ReadPropery(System_status (112), Firmware_version (44), ...)

      BBacnetStack bacnetStack = hmiCommCheck.bacnetStack;
      BBacnetClientLayer clientLayer = hmiCommCheck.clientLayer;
      BBacnetObjectIdentifier hmiObjectId = hmiCommCheck.hmiObjectId;
      BBacnetAddress hmiAddress = hmiCommCheck.hmiAddress;

      BBacnetDeviceStatus systemStatus;
      systemStatus = (BBacnetDeviceStatus) BHMICommHandler.readProperty(clientLayer, hmiObjectId, hmiAddress,
          BBacnetPropertyIdentifier.SYSTEM_STATUS,
          BacnetConst.ASN_ENUMERATED,
          BBacnetDeviceStatus.DEFAULT);
      HonPlantControllerHMILogger.finest(MessageFormat.format("{0} systemStatus[{1}]", LOG_MSG_PREFIX, systemStatus));

      String modelName = ((BString) BHMICommHandler.readProperty(clientLayer, hmiObjectId, hmiAddress,
          BBacnetPropertyIdentifier.MODEL_NAME,
          BacnetConst.ASN_CHARACTER_STRING,
          BString.DEFAULT)).getString();

      HonPlantControllerHMILogger.info(MessageFormat.format("{0} modelName[{1}]", LOG_MSG_PREFIX, modelName));

      if (modelName.isEmpty()) {
        HonPlantControllerHMILogger.warning(MessageFormat.format("{0} HMI {1} file ({2}) is not compatible " +
                "with HMI, model name: ''{3}''", LOG_MSG_PREFIX,
            identFileType, fwIdent, modelName));
        throw new LocalizableException(LEX.module.getModuleName(), lexKeyWrongModel,
            new Object[]{fwIdent, modelName});
      }

      String hmiRevision = null;
      if (isFwUpdate) {
        hmiRevision = ((BString) BHMICommHandler.readProperty(clientLayer, hmiObjectId, hmiAddress,
            BBacnetPropertyIdentifier.FIRMWARE_REVISION,
            BacnetConst.ASN_CHARACTER_STRING,
            BString.DEFAULT)).getString();

        HonPlantControllerHMILogger.info(MessageFormat.format("{0} hmiRevision[{1}]", LOG_MSG_PREFIX, hmiRevision));
        if (fwIdent.equals(hmiRevision))
        {
          HonPlantControllerHMILogger.info(MessageFormat.format("[0} HMI {1} file is already loaded in the HMI, " +
                                                                "version: ''{2}''", LOG_MSG_PREFIX, identFileType,
                                                                hmiRevision));
          if (job != null)
          {
            job.log().message(LEX.module.getModuleName(), lexKeyAlreadyLoaded, hmiRevision);
          }
          return false;
        } else if (isRevisionOlder(fwIdent, hmiRevision)) {
        	HonPlantControllerHMILogger.info(MessageFormat.format("{0} HMI {1} file (version: ''{2}'') is older then " +
                "the one in the HMI (version: ''{3}'')", LOG_MSG_PREFIX,
            identFileType, fwIdent, hmiRevision));
        	throw new LocalizableException(LEX.module.getModuleName(), "hmi.downloadFirmware.file.is.older",
            new Object[]{fwIdent, hmiRevision});
        }
      }

      int maxAPDULengthAccepted =
          ((BBacnetUnsigned) BHMICommHandler.readProperty(clientLayer, hmiObjectId, hmiAddress,
              BBacnetPropertyIdentifier.MAX_APDU_LENGTH_ACCEPTED,
              BacnetConst.ASN_UNSIGNED,
              BInteger.DEFAULT)).getInt();

      HonPlantControllerHMILogger.finest(MessageFormat.format("{0} maxAPDULengthAccepted[{1,number,0}]",
          LOG_MSG_PREFIX, maxAPDULengthAccepted));

      BBacnetBitString protocolServicesSupported =
          (BBacnetBitString) BHMICommHandler.readProperty(clientLayer, hmiObjectId, hmiAddress,
              BBacnetPropertyIdentifier.PROTOCOL_SERVICES_SUPPORTED,
              BacnetConst.ASN_BIT_STRING, BBacnetBitString.DEFAULT);
      int atomicWriteFileIndex = BacnetBitStringUtil.getBitIndex(BacnetBitStringUtil.BACNET_SERVICES_SUPPORTED,
          "atomicWriteFile");
      boolean isAtomicWriteFileSupported = protocolServicesSupported.getBit(atomicWriteFileIndex);

      HonPlantControllerHMILogger.finest(MessageFormat.format("{0} protocolServicesSupported[{1}] " +
              "isAtomicWriteFileSupported[{2}]", LOG_MSG_PREFIX,
          protocolServicesSupported, isAtomicWriteFileSupported));

      if (job != null) {
        job.progress(job.getProgress() + 1);
        job.log().message(LEX.module.getModuleName(), lexKeyStartFileDownload);
        if (isFwUpdate) {
          job.log().message(LEX.module.getModuleName(), "hmi.downloadFirmware.information",
              new String[]{hmiRevision, fwIdent, Integer.toString(fwImage.length)});
        } else {
          job.log().message(LEX.module.getModuleName(), "hmi.downloadLanguage.information",
              new String[]{fwIdent, Integer.toString(fwImage.length)});
        }
      }

      if (isFwUpdate) {
        // reinitializeDevice(StartFirmwareDownload(126))
        reinitializeDevice(bacnetStack, hmiAddress, HMIReinitializeDeviceDownloadRequest.START_FIRMWARE_DOWNLOAD);
      } else {
        // reinitializeDevice(StartLanguageDownload(128))
        reinitializeDevice(bacnetStack, hmiAddress, HMIReinitializeDeviceDownloadRequest.START_LANGUAGE_DOWNLOAD);
      }
      hmiIAmListener.resetHmiIsPresent();

      if (job != null) {
        job.progress(job.getProgress() + 1);
      }

      if (!isFwUpdate) {
        // readProperty(configuraion_files)
        BBacnetObjectIdentifier configurationFiles =
            (BBacnetObjectIdentifier) BHMICommHandler.readProperty(clientLayer, hmiObjectId, hmiAddress,
                BBacnetPropertyIdentifier.CONFIGURATION_FILES,
                BacnetConst.ASN_OBJECT_IDENTIFIER,
                BBacnetObjectIdentifier.DEFAULT);
        HonPlantControllerHMILogger.finest(MessageFormat.format("{0} configurationFiles[{1}]",
            LOG_MSG_PREFIX, configurationFiles));
      }

      // readProperty(File x, file type), x = 0 for firmware, x = 1 for language
      BBacnetObjectIdentifier hmiFileObjId = BBacnetObjectIdentifier.make(BBacnetObjectType.FILE,
          (isFwUpdate) ? 0 : 1);
      BString fileType = (BString) BHMICommHandler.readProperty(clientLayer, hmiFileObjId, hmiAddress,
          BBacnetPropertyIdentifier.FILE_TYPE,
          BacnetConst.ASN_CHARACTER_STRING, BString.DEFAULT);
      HonPlantControllerHMILogger.finest(MessageFormat.format("{0} hmiFileObjId[{1}] fileType[{2}]", LOG_MSG_PREFIX,
              hmiFileObjId, fileType));

      // TODO is check fileType for 'frm' or 'lang' necessary?

      if (job != null) {
        job.progress(job.getProgress() + 1);
      }

      // writeProperty(File x, file size = 0)
      BHMICommHandler.writeProperty(clientLayer, hmiFileObjId, hmiAddress, BBacnetPropertyIdentifier.FILE_SIZE,
          BacnetConst.ASN_UNSIGNED, BBacnetUnsigned.make(0));

      if (job != null) {
        job.progress(job.getProgress() + 1);
      }

      // atomicWriteFile(File x)
      boolean success = writeFileData(clientLayer, hmiFileObjId, hmiAddress, fwImage, maxAPDULengthAccepted,
          isAtomicWriteFileSupported, bacnetStack, job);
      
		if (!success) {
			if (job != null) {
				job.log().message(LEX.module.getModuleName(), lexKeyFileDownloadCancelled);
				job.setJobState(BJobState.canceled);
			}
			if(isFwUpdate) {
				isFirmwareDownloadCancelled = true;
			}
			return false;
		}

      if (job != null) {
        job.progress(99);
        job.log().message(LEX.module.getModuleName(), lexKeyEndFileDownload);
      }

      if (isFwUpdate) {
        // reinitializeDevice(EndFirmwareDownload(127))
        reinitializeDevice(bacnetStack, hmiAddress, HMIReinitializeDeviceDownloadRequest.END_FIRMWARE_DOWNLOAD);
      } else {
        // reinitializeDevice(EndLanguageDownload(129))
        reinitializeDevice(bacnetStack, hmiAddress, HMIReinitializeDeviceDownloadRequest.END_LANGUAGE_DOWNLOAD);
      }
      return true;
    } catch (Exception exc) {
      // unexpected exception
      HonPlantControllerHMILogger.log(Level.SEVERE,
                                      MessageFormat.format("{0} something went wrong", LOG_MSG_PREFIX), exc);
      throw exc;
    } finally {
      hmiFileUpdateIsRunning = false;
      setIsHmiJobActive(false);
      if (job != null) {
        job.progress(100);
      }
    }
  }

  protected String getFwFileNameFromModule() {
    final String LOG_MSG_PREFIX = TYPE.getTypeName() + ".getFwFileNameFromModule";

    // get firmware file from Niagara module
    String fwFolderName = "module://honFirmwarePackage/res/HMIFirmware";
    BOrd fwFolderOrd = BOrd.make(fwFolderName);
    BDirectory fwFolder = (BDirectory)fwFolderOrd.get();
    for (BIFile file : fwFolder.listFiles()) {
      HonPlantControllerHMILogger.finest(MessageFormat.format("{0} HMI firmware file.filePath[{1}] " +
                                                            "file.fileName[{2}]", LOG_MSG_PREFIX, file.getFilePath(),
                                                            file.getFileName()));
      if ((file instanceof BDataFile) && ("." + file.getExtension()).equals(HMI_FIRMWARE_FILE_EXTENSION)) {
        HonPlantControllerHMILogger.finest(MessageFormat.format("{0} HMI firmware   using", LOG_MSG_PREFIX));
        return file.getFilePath().toString();
      }
    }
    return null;
  }

  /**
   * check whether file contains an older firmware revision then the one in the HMI or not
   *
   * @param fwIdent     revision of the firmware contained in the file
   * @param hmiRevision revision of the firmware currently in the HMI
   * @return true:  revision of the file is older then or equal to the HMI,
   *         false: revision is newer or downgrade is allowed
   */
  protected boolean isRevisionOlder(String fwIdent, String hmiRevision) {
    if (getSupportHmiFirmwareDowngrade()) {
      return false;
    }

    // example for revision: HMI_FW_v1.5.0.12
    final String SEPARATOR = "_FW_v";
    String[] fwIdentParts = fwIdent.split(SEPARATOR);
    String[] hmiRevisionParts = hmiRevision.split(SEPARATOR);
    if ((fwIdentParts.length != 2) || (hmiRevisionParts.length != 2)) {
      return true;
    }

    Version fwVersion = new Version(fwIdentParts[1]);
    Version hmiVersion = new Version(hmiRevisionParts[1]);

    return fwVersion.compareTo(hmiVersion) <= 0;
  }

  protected byte[] readFileFromModule(String imageFileName)
      throws IOException {
    BOrd imageFileOrd = BOrd.make(imageFileName);
    BDataFile imageDataFile = (BDataFile)imageFileOrd.get();
    return imageDataFile.read();
  }

  // reimplementation of BBacnetClientLayer.reinitializeDevice to send states 126 (StartFirmwareDownload),
  // 127 (EndFirmwareDownload), 128 (StartLanguageDownload), and 129 (EndLanguageDownload).
  // Those are not supported by Niagara.
  protected void reinitializeDevice(BBacnetStack bacnetStack, BBacnetAddress hmiAddress, int stateOfUpdate)
      throws BacnetException {
    final String LOG_MSG_PREFIX = getType().getTypeName() + ".reinitializeDevice";

    if (!bacnetStack.isCommInitiationEnabled()) {
      throw new BacnetStackException("Stack Disabled");
    }

    BCharacterSetEncoding encoding = BCharacterSetEncoding.iso10646_UTF8;
    HonPlantControllerHMILogger.fine("BacnetClientLayer.reinitializeDevice(" + stateOfUpdate + "; " +
            HMI_PASSWORD_REINITIALIZE_DEVICE + "; " + encoding + ")");

    HMIReinitializeDeviceDownloadRequest request;
    request = new HMIReinitializeDeviceDownloadRequest(stateOfUpdate, HMI_PASSWORD_REINITIALIZE_DEVICE, encoding);
    HonPlantControllerHMILogger.fine(MessageFormat.format("{0} request hmiAddress[{1}] request[{2}]", LOG_MSG_PREFIX,
                                                          hmiAddress, request));
    BacnetServicePrimitive response = bacnetStack.getTransport().sendConfirmedRequestSimple(request, hmiAddress,
                                                                                            BNetworkPriority.normal);
    HonPlantControllerHMILogger.fine(MessageFormat.format("{0} response hmiAddress[{1}] response[{2}]",
                                                          LOG_MSG_PREFIX, hmiAddress, response));
  }

  // reimplementation of BBacnetFile.writeFileDataStream to control the progress bar of BSimpleJob
  protected boolean writeFileData(BBacnetClientLayer clientLayer, BBacnetObjectIdentifier bacFileObj,
                                   BBacnetAddress hmiAddress, byte[] fwImage, int maxAPDULengthAccepted,
                                   boolean isAtomicWriteFileSupported, BBacnetStack bacnetStack, BSimpleJob job)
      throws BacnetException {
    final String LOG_MSG_PREFIX = getType().getTypeName() + ".writeFileData";
    final String LOG_MSG_FORMAT_PROGRESS = "{0} {1,number,0} bytes written";
    final int LOG_LIMIT_PERIOD = 100000;

    if (!isAtomicWriteFileSupported) {
      throw new UnsupportedOperationException(BACNET_LEXICON.getText("serviceNotSupported.atomicWriteFile"));
    }

    int writeLength = fwImage.length;
    int startProgress = (job == null) ? 0 : job.getProgress();
    int diffProgress = 99 - startProgress;

    // Determine device's data return capability.  First take minimum of
    // device's and our own max APDU.  Then subtract a safety factor because
    // some devices abort if you get close to (but not over) the limit.
    int maxApdu = maxAPDULengthAccepted;
    int myMax = BBacnetNetwork.localDevice().getMaxAPDULengthAccepted();
    if (myMax < maxApdu) {
      maxApdu = myMax;
    }
    maxApdu -= ACK_HEADER_SIZE;

    // If we either know we need to send it in multiple requests, or the single
    // request failed, try the multiple request approach.
    int logLimit = LOG_LIMIT_PERIOD;
    int start = 0;
    HonPlantControllerHMILogger.info(MessageFormat.format(LOG_MSG_FORMAT_PROGRESS, LOG_MSG_PREFIX, start));
    if (job != null) {
      job.log().message(LEX.module.getModuleName(), LEXICON_KEY_PROGRESS_DOWNLOAD, Integer.toString(start));
    }
    int len = maxApdu;
    do {
      int copylen = len;
      if (start + copylen > writeLength) {
        copylen = writeLength - start;
      }
      byte[] copyData = new byte[copylen];
      System.arraycopy(fwImage, start, copyData, 0, copylen);
      clientLayer.atomicWriteFileStream(hmiAddress, bacFileObj, start, copyData);
      start += copylen;
      
      if (job != null) {
    	if(job.getJobState().equals(BJobState.canceling)) {
    		reinitializeDevice(bacnetStack, hmiAddress, HMIReinitializeDeviceDownloadRequest.ABORT_DOWNLOAD);
    		return false;
    	}
        int newProgress = startProgress + (start * diffProgress) / writeLength;
        int currProgress = job.getProgress();
        if (currProgress != newProgress) {
          job.progress(newProgress);
        }
      }

      if (start > logLimit) {
        logLimit += LOG_LIMIT_PERIOD;
        HonPlantControllerHMILogger.info(MessageFormat.format(LOG_MSG_FORMAT_PROGRESS, LOG_MSG_PREFIX, start));
        if (job != null) {
          job.log().message(LEX.module.getModuleName(), LEXICON_KEY_PROGRESS_DOWNLOAD, Integer.toString(start));
        }
      }
    }
    while (start < writeLength);

    HonPlantControllerHMILogger.info(MessageFormat.format(LOG_MSG_FORMAT_PROGRESS, LOG_MSG_PREFIX, start));
    if (job != null) {
      job.log().message(LEX.module.getModuleName(), LEXICON_KEY_PROGRESS_DOWNLOAD, Integer.toString(start));
    }
    return true;
  }

  public BBoolean doIsHMIPresent() {
	  handleServiceEnabled(TYPE.getModule().getModuleName(), LEX_MSG_HMI);
    if (hmiFileUpdateIsRunning || (getBacnetNetwork() == null) || !bacnetNetwork.getEnabled() ||
        (hmiCommHandler == null)) {
      // firmware update is already running, missing commissioning, network was deleted or disabled, ...
      return BBoolean.FALSE;
    }

    HMIIAmListener hmiIAmListener = hmiCommHandler.getHmiIAmListener();
    return BBoolean.make((hmiIAmListener != null) && hmiIAmListener.hmiIsPresent());
  }

  public BString doInitiateFileCopy(BString filename) {
	  handleServiceEnabled(TYPE.getModule().getModuleName(), LEX_MSG_HMI);
    if (Sys.getStation() == null) {
      return BString.make(LEX.getText("station.not.running"));
    }

    String fileName = filename.toString();
    HonPlantControllerHMILogger.finest(MessageFormat.format("InitiateFileCopy started for ''{0}''", fileName));
    try {
      Path firmwarePath = Sys.getStationHome().toPath().resolve(FIRMWARE_FOLDER_NAME);
      if (!firmwarePath.toFile().exists()) {
        firmwarePath = Files.createDirectories(firmwarePath);
      }

      Path firmwareFilePath = firmwarePath.resolve(fileName);
      if (Files.deleteIfExists(firmwareFilePath)) {
        HonPlantControllerHMILogger.info(MessageFormat.format("file ''{0}'' already exists inside station''s " +
                                                              "firmware folder: deleted", fileName));
      }
      Files.createFile(firmwareFilePath);
      HonPlantControllerHMILogger.info(MessageFormat.format("New file is created inside station''s firmware folder " +
                                                            "with name ''{0}''", fileName));
    } catch (IOException e) {
      HonPlantControllerHMILogger.finest(e.getMessage(), e);
      return BString.make(LEX.getText("hmi.downloadFile.Exception", fileName, e));
    }

    return BString.make(SUCCESS);
  }

  public BString doCopyBinaryString(BHMIFirmwareDownloadConfig config) {
	  handleServiceEnabled(TYPE.getModule().getModuleName(), LEX_MSG_HMI);
    if (Sys.getStation() == null) {
      return BString.make(LEX.getText("station.not.running"));
    }

    String fileName = config.getFileName();
    try {
      byte[] binaries = Base64.getDecoder().decode(config.getBinaryString());
      Path firmwarePath = Sys.getStationHome().toPath().resolve(FIRMWARE_FOLDER_NAME);
      Path filePath = firmwarePath.resolve(fileName);
      Files.write(filePath, binaries, StandardOpenOption.APPEND);
    } catch (IOException e) {
      HonPlantControllerHMILogger.finest(e.getMessage(), e);
      return BString.make(LEX.getText("hmi.downloadFile.Exception", fileName, e));
    }

    return BString.make(SUCCESS);
  }

  public static String getFileIdent(byte[] content) {
    return new String(content).trim();
  }

  public static boolean isFwUpdate(BString filename) {
    if (BString.DEFAULT.equals(filename)) {
      return true;
    }
    String nameOfFile = filename.getString();
    int lastDot = nameOfFile.lastIndexOf('.');
    if (lastDot < 0) {
      return true;
    }
    return nameOfFile.substring(lastDot).equals(HMI_FIRMWARE_FILE_EXTENSION);
  }

  public BBacnetNetwork getBacnetNetwork() {
	  try {
	    if (bacnetNetwork == null) {
	      bacnetNetwork = getBacnetProperties().getBacnetNetwork();
	    }
	  }catch (Exception e) {
		  return null;
	}
    return bacnetNetwork;
  }

  private Optional<BCertManagerService> getCertManagerService() {
    Type certManType = Sys.getType(BCertManagerService.class);
    Optional<BIService> result = Sys.findService(certManType);

    if (result.isPresent()) {
      return Optional.of((BCertManagerService)result.get());
    }
    return Optional.empty();
  }

  public void setBacnetNetwork(BBacnetNetwork bacnetNetwork) {
    this.bacnetNetwork = bacnetNetwork;
  }


  public Optional<Integer> shouldUpdateHMI() {
    return checkHmiSyncStatus();
  }

  public Optional<Integer> checkHmiSyncStatus() {
    int result = 0;
    BBacnetNetwork network = getBacnetNetwork();
    BBacnetExportTable exportTable = (BBacnetExportTable) network.getLocalDevice().getExportTable();
    BComponent[] export = exportTable.getChildComponents();
    HashMap<String, BComponent> currentExportDPTable = new HashMap<>();
    HashMap<String, BComponent> currentExportSchedulesTable = new HashMap<>();
    HashMap<String, BComponent> currentExportCalendarsTable = new HashMap<>();

    for (BComponent comp : export) {
      if (comp instanceof BBacnetPointDescriptor) {
        currentExportDPTable.put(comp.getName(), comp);
      } else if (comp instanceof BBacnetScheduleDescriptor){
        currentExportSchedulesTable.put(comp.getName(), comp);
      } else if (comp instanceof BBacnetCalendarDescriptor){
    	  currentExportCalendarsTable.put(comp.getName(), comp);
      }
    }

    // looks like this is the first check
    if (null == exportDPTableSync || !exportDPTableSync.equals(currentExportDPTable)) {
      result += BHMISyncCategory.datapoints.getOrdinal();
    } 
    exportDPTableSync = currentExportDPTable;

    // looks like this is the first check
    if (null == exportSchedulesTableSync || !exportSchedulesTableSync.equals(currentExportSchedulesTable)) {
      result += BHMISyncCategory.schedules.getOrdinal();
    } 
    exportSchedulesTableSync = currentExportSchedulesTable;

    if (null == exportCalendarTableSync || !exportCalendarTableSync.equals(currentExportCalendarsTable)) {
        result += BHMISyncCategory.schedules.getOrdinal();
    } 
    exportCalendarTableSync = currentExportCalendarsTable;

    List<BUser> users = HonPlantControllerUserUtils.getUsers();

    HashMap<String, BUser> currentUserTable = new HashMap<>();
    for (BUser user : users) {
      currentUserTable.put(user.getName(), user);
    }

    if (null == userTableSync || !userTableSync.equals(currentUserTable)) {
    	result += BHMISyncCategory.user.getOrdinal();
    } 
    userTableSync = currentUserTable;

    // if odd datapoints was not added yet.
    if (BHMISyncStatus.outOfSync.equals(getHome().getFALSyncStatus()) && (result % 2 != 1)) {
      result += BHMISyncCategory.datapoints.getOrdinal();
    }

    if (getAlarmPriorityConfiguration().getSyncStatus().equals(BHMISyncStatus.outOfSync)) {
      result += BHMISyncCategory.alarmPriority.getOrdinal();
    }

    return result != 0 ? Optional.of(result) : Optional.empty();
  }

  public void resetFALSyncStatus() {
    getHome().resetFALSyncStatus();
    setFalSyncStatus(BHMISyncStatus.inSync);
  }

  public void restAlarmPrioritySyncStatus() {
    getAlarmPriorityConfiguration().setSyncStatus(BHMISyncStatus.inSync);
  }

  private static class LangFileInfo {
    private final String filename;
    private final Version version;

    public LangFileInfo(String filename, Version version) {
      this.filename = filename;
      this.version = version;
    }

    @Override
    public String toString() {
      return "name[" + filename + "] vers[" + version + "]";
    }
  }

  public byte[] getCertBytes() throws CertificateEncodingException {
    return deviceCert.getEncoded();
  }

  public Subscriber getFALSubscriber() {
    if (null == subscriber)
      return null;

    return subscriber.getFALSubscriber();
  }

  private class HMICommCheck {
    private final HMIIAmListener hmiIAmListener;
    private final BBacnetStack bacnetStack;
    private final BBacnetClientLayer clientLayer;
    private final BBacnetObjectIdentifier hmiObjectId;
    private final BBacnetAddress hmiAddress;

    public HMICommCheck(String logMsgPrefix)
        throws LocalizableException {
      BNetworkPort hmiPort = HonPlantControllerBacnetUtils.getHMIPort(getBacnetNetwork(), getBacnetProperties());
      if (hmiPort == null) {
        HonPlantControllerHMILogger.finest("{0} missing HMI port", logMsgPrefix);
        throw new LocalizableException(LEX, "hmi.downloadFile.port.missing");
      }
      if (hmiCommHandler == null) {
        HonPlantControllerHMILogger.finest("{0} missing HMI communication handler", logMsgPrefix);
        throw new LocalizableException(LEX, "hmi.downloadFile.comm.handler.missing");
      }
      hmiIAmListener = hmiCommHandler.getHmiIAmListener();
      if (hmiIAmListener == null) {
        HonPlantControllerHMILogger.finest("{0} missing HMI I-AM listener", logMsgPrefix);
        throw new LocalizableException(LEX, "hmi.downloadFile.iam.listener.missing");
      }

      bacnetStack = (BBacnetStack) bacnetNetwork.getBacnetComm();
      clientLayer = bacnetStack.getClient();
      int hmiNetwork = hmiPort.getNetworkNumber();
      hmiObjectId = getBacnetProperties().getHMIInstanceId();  //getHMiInstanceId
      hmiAddress = new BBacnetAddress(hmiNetwork, hmiIAmListener.getHmiMacAddress());
    }
  }

// Attributes
  private BBacnetNetwork bacnetNetwork = null;

  private BHMICommHandler hmiCommHandler = null;

  public BHMICommHandler getHMICommhandler() {
    return hmiCommHandler;
  }

  public void setHMICommhandler(BHMICommHandler hmiCommHandler) {
    this.hmiCommHandler = hmiCommHandler;
  }

  public void registerHMIDevice(){
      try {
  		if(Sys.isStation() && Sys.atSteadyState()) {
  			setBacnetNetwork(null);
			if(null != getHMICommhandler()) {
              getHMICommhandler().unregisterListeners();
            }
  			setHMICommhandler(null);
  			BBacnetNetwork bacNetwork = this.getBacnetProperties().getBacnetNetwork();
  			setBacnetNetwork(bacNetwork);
  			setHMICommhandler(null);
  			registerListeners();
  		}
      } catch (Exception e) {
    	  HonPlantControllerHMILogger.finest(e.getMessage(), e);
      }
  }
  
	public void setFirmwareDownloadCancelled(boolean isFirmwareDownloadCancelled) {
		this.isFirmwareDownloadCancelled = isFirmwareDownloadCancelled;
	}

	public boolean isFirmwareDownloadCancelled() {
		return isFirmwareDownloadCancelled;
	}
	
	public String getSerialNumber() {
		return serialNumber;
	}

	public void setSerialNumber(String serialNumber) {
		this.serialNumber = serialNumber;
	}

	public BBacnetObjectIdentifier getInstanceIdFromHMI() {
		return instanceIdFromHMI;
	}

	public void setInstanceIdFromHMi(BBacnetObjectIdentifier instanceIdFromHMI) {
		this.instanceIdFromHMI = instanceIdFromHMI;
	}

	public boolean isHMIInstSet() {
		return isHMIInstSet;
	}

	public void setHMIInstSet(boolean isHMIInstSet) {
		this.isHMIInstSet = isHMIInstSet;
	}
	
	
	
	
  private BHonPlantControllerHMISubscriber subscriber;
  private static final Object syncObject = new Object();

  private static final Lexicon LEX = Lexicon.make(BHonPlantControllerHMI.class);
  private static final Lexicon BACNET_LEXICON = Lexicon.make("bacnet");
  public static final String SUCCESS = "SUCCESS";
  public static final String FIRMWARE_FOLDER_NAME = "firmware";
  public static final int HMI_OBJECT_ID = 99;
  private static final int ACK_HEADER_SIZE = 30;
  public static final int HMI_HEADER_IDENT_OFFSET = 0x100;
  public static final int HMI_HEADER_IDENT_MAX_LENGTH = 40;
  public static final String LEXICON_KEY_PROGRESS_DOWNLOAD = "hmi.downloadFile.progress.download";
  protected boolean hmiFileUpdateIsRunning = false;
  public static final String HMI_FIRMWARE_FILE_EXTENSION = ".frm";
  public static final String HMI_LANGUAGE_FILE_EXTENSION = ".lang";
  public static final String HMI_PROP_NAME_FW_SOURCE = "fwSource";
  private static final String HMI_PASSWORD_REINITIALIZE_DEVICE = "Zone";
  private boolean isFirmwareDownloadCancelled = false;

  private HashMap<String, BComponent> exportDPTableSync = null;
  private HashMap<String, BComponent> exportSchedulesTableSync = null;
  private HashMap<String, BComponent> exportCalendarTableSync = null;
  private HashMap<String, BUser> userTableSync = null;

  public static final Logger logger = Logger.getLogger("HonPlantControllerHMI");

  private static final BDynamicEnum NO_CERTS_FOUND = BDynamicEnum.make(BEnumRange.make(new String[] {"NoCertsFound"}));
  private static final String KEY_STORE_KEY = "changeit";
  private static final String HONEYWELL_CA = "acs products ca";
  private static final String HONEYWELL_BEATS_CERT = "hon_advanced_niagara_controller";
  private static final String LEX_MSG_HMI="hmi.service.errormsgHmi";
  private String serialNumber = null;
  X509Certificate deviceCert;
  private static final int BEATS_CERT_ORIDINAL = 0;
  private BBacnetObjectIdentifier instanceIdFromHMI =null;
  private boolean isHMIInstSet =false;
  private static BHMIDevice hmiDevice= new BHMIDevice();
  
  private  static final int SIXTY = 60;
  private  static final int HUNDREAD = 100;
  

}
