/*
Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
This file contains trade secrets of Honeywell International, Inc. No part
may be reproduced or transmitted in any form by any means or for any
purpose without the express written permission of Honeywell.
*/
package com.honeywell.honplantcontroller.hmi.util;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.baja.bacnet.datatypes.BBacnetArray;
import javax.baja.bacnet.datatypes.BBacnetObjectIdentifier;
import javax.baja.bacnet.enums.BBacnetObjectType;
import javax.baja.bacnet.export.BIBacnetExportObject;
import javax.baja.bacnet.export.BLocalBacnetDevice;
import javax.baja.control.BControlPoint;
import javax.baja.naming.BOrd;
import javax.baja.sys.BComponent;
import javax.baja.sys.BObject;
import javax.baja.sys.BVector;
import javax.baja.sys.Sys;

import com.honeywell.honbacnethelper.export.BHonDeviceExtDescriptor;
import com.honeywell.honbacnethelper.export.BHonFastAccessList;
import com.honeywell.honbacnethelper.export.BIHonBacnetCustomDescriptor;
import com.honeywell.honplantcontroller.hmi.service.BHonPlantControllerHMI;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerBacnetPointExportUtils;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerBacnetScheduleExportUtils;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerHMILexicon;
import com.honeywell.honplantcontroller.hmi.utils.HonPlantControllerHMILogger;
import com.honeywell.honplantcontroller.hmi.view.BFalAssignment;
import com.honeywell.honplantcontroller.hmi.view.BFastAccessList;
import com.tridium.bacnet.datatypes.BSvoSubordinate;
import com.tridium.bacnet.stack.server.BBacnetExportTable;

public class HonPlantControllerFastAccessListExporter {
    public HonPlantControllerFastAccessListExporter() {
        //
    }

    public ArrayList exportFalTolocalDevice(BFastAccessList fal) {
        Map<String, List<String>> dpFalNotExportable = new HashMap<>();
        Map<String, List<String>> dpFalObjectsNotExportable = new HashMap<>();

        ArrayList result = new ArrayList<>();
        result.add(dpFalNotExportable);
        result.add(dpFalObjectsNotExportable);


        this.exportTable = getExportTable(fal);

        if (this.exportTable == null) {
            return result;
        }

        BHonDeviceExtDescriptor<?, ?>[] deviceExtDescriptors = exportTable.getChildren(BHonDeviceExtDescriptor.class);
        if (deviceExtDescriptors.length == 0)
        {
            return result;
        }
        deviceExtDescriptor = deviceExtDescriptors[0];
        if (deviceExtDescriptor == null) {
            return result;
        }

        if (!isFALObjectsExportable(fal, dpFalObjectsNotExportable)) {
            //Log the error and don't export the FAL
            HonPlantControllerHMILogger.warning("[exportFAL] Not exported. Objects are not exportable!");
            dpFalNotExportable.put("[" + fal.getDisplayName(null) + "]", Collections.singletonList(fal.getDisplayName(null)));
            return result;
        }

        BHonFastAccessList exportedFAL = exportFAL(fal);
        if (null != exportedFAL) {
            updateAlreadyExportedObjectsSet(this.exportTable, fal, exportedFAL);
            exportFALObjects(fal, exportedFAL);
        }

        return result;
    }

    private BBacnetExportTable getExportTable(BFastAccessList fal) {
        BHonPlantControllerHMI service = (BHonPlantControllerHMI) fal.getParent().getParent().getParent();
        return (BBacnetExportTable) service.getBacnetNetwork().getLocalDevice().getExportTable();

    }

    private void updateAlreadyExportedObjectsSet(BBacnetExportTable bacnetExportTable, BFastAccessList fal, BHonFastAccessList exportedFAL) {

        if (null == bacnetExportTable || null == fal || null == exportedFAL) {
            return;
        }

        BSvoSubordinate[] fastAccesListEntries = exportedFAL.getSubordinates().getChildren(BSvoSubordinate.class);

        fal.loadSlots();
        BVector objects = fal.getObjectsList();
        objects.lease(10);
        BComponent[] children = objects.getChildComponents();

        List<BFalAssignment> falAssigments = Stream.of(children)
            .filter(BFalAssignment.class::isInstance)
            .map(BFalAssignment.class::cast)
            .collect(Collectors.toList());

        Stream.of(fastAccesListEntries).forEach(exportedFALentry -> {

            BBacnetObjectIdentifier objectId = exportedFALentry.getReference().getObjectId();

            BIBacnetExportObject exportedObject = bacnetExportTable.byObjectId(objectId);

            boolean found = falAssigments.stream().anyMatch(falAssigment -> falAssigment.getObjectOrd().getOrd().equals(exportedObject.getObjectOrd()));

            if (!found) {
                exportedFAL.removeEntry(objectId);
            }
        });
    }

    private boolean isAlreadyExported(BOrd handleOrd, BHonFastAccessList fal, BBacnetExportTable exportTable) {
        BBacnetObjectIdentifier objectId = exportTable.byObjectOrd(handleOrd);

        if (null == objectId) {
            return false;
        }

        return fal.isObjectIdExported(objectId);
    }

    private Optional<BHonFastAccessList> isFalAlreadyExported(String falName) {
        BHonFastAccessList[] exportedFALs = deviceExtDescriptor.getFastAccessLists().getChildren(BHonFastAccessList.class);

        return Stream.of(exportedFALs).filter(fal -> fal.getObjectName().equalsIgnoreCase(falName)).findFirst();
    }

    private BHonFastAccessList exportFAL(BFastAccessList fastAccessList) {

        fastAccessList.lease();
        fastAccessList.getAbsoluteOrd();
        BVector objects = fastAccessList.getObjectsList();
        objects.lease(10);
        BComponent[] children = objects.getChildComponents();

        if (children.length <= 0) {
            //do not export empty FAL
            return null;
        }

        try {
            this.exportTable.lease();
            int instanceNumber = this.exportTable.getNextInstance(BBacnetObjectType.STRUCTURED_VIEW);
            Optional<BHonFastAccessList> fal = isFalAlreadyExported(fastAccessList.getName());

            if (!fal.isPresent()) {
                BHonFastAccessList exportedFAL = BHonFastAccessList.make(fastAccessList.getDisplayName(null), instanceNumber);
                exportedFAL.setServiceFALOrd(fastAccessList.getHandleOrd());
                exportFalInDeviceExt(exportedFAL);
                return exportedFAL;
            }
            return fal.get();
        } catch (Exception e) {
            HonPlantControllerHMILogger.warning("exportFAL", e);
        }

        return null;

    }

    private void exportFALObjects(BFastAccessList fastAccessList, BHonFastAccessList exportedFAL) {

        fastAccessList.lease();
        BVector objects = fastAccessList.getObjectsList();
        BComponent[] children = objects.getChildComponents();
        exportTable.lease();
        for (int i = 0; i < children.length; i++) {
            BComponent child = children[i];
            child.lease();
            if (child instanceof BFalAssignment) {
                try {
                    BOrd falChildOrd = ((BFalAssignment) child).getObjectOrd().getOrd();
                    if (!isAlreadyExported(falChildOrd, exportedFAL, exportTable)) {
                        addToExportTable(exportTable, falChildOrd, exportedFAL);
                    }
                } catch (Exception e) {
                    HonPlantControllerHMILogger.warning("can not export fal object" + e.getMessage());
                }
            }
        }
    }

    private boolean isFALObjectsExportable(BFastAccessList fastAccessList, Map<String, List<String>> dpFalObjectsNotExportable) {
        fastAccessList.lease();
        BVector objects = fastAccessList.getObjectsList();
        BComponent[] children = objects.getChildComponents();

        //Boolean used to check if we have at least 1 good element in the FAL
        boolean ok = false;
        for (int i = 0; i < children.length; i++) {
            BComponent child = children[i];
            child.lease();
            if (child instanceof BFalAssignment) {
                BOrd falChildOrd = ((BFalAssignment) child).getObjectOrd().getOrd();
                try {
                    BControlPoint controlPoint = (BControlPoint) falChildOrd.resolve(Sys.getStation()).get();
                    if ((!HonPlantControllerBacnetPointExportUtils.checkFittingAlarmExt(controlPoint)) ||
                        (!HonPlantControllerBacnetPointExportUtils.checkEnumRange(controlPoint))) {
                        dpFalObjectsNotExportable.put("[" + controlPoint.getSlotPath() + "]", Collections.singletonList(controlPoint.getName()));
                    } else {
                        //At least 1 good element exists
                        ok = true;
                    }
                } catch (Exception e) {
                    HonPlantControllerHMILogger.warning(MessageFormat.format(getText("hmi.commission.finished.fal.exportable.exception"), e.getMessage()));
                }
            }
        }
        return ok;
    }

    private void addToExportTable(BBacnetExportTable exportTable, BOrd falObject, BHonFastAccessList exportedFAL) {

        BBacnetObjectIdentifier objectId = HonPlantControllerBacnetScheduleExportUtils.findObjectIdForOrd(exportTable, falObject);
        if (null == exportTable || null == objectId) {
            return;
        }
        BBacnetObjectIdentifier deviceId = ((BLocalBacnetDevice) exportTable.getParent()).getObjectId();

        BIBacnetExportObject exportObject = exportTable.byObjectId(objectId);
        BIHonBacnetCustomDescriptor exportedObject = (BIHonBacnetCustomDescriptor) exportObject;
        exportedObject.setPointInFAL(true);

        exportedFAL.addEntry(deviceId, objectId);

    }

    private void exportFalInDeviceExt(BHonFastAccessList exportedFAL) {
        BBacnetArray fals = deviceExtDescriptor.getFastAccessLists();
        fals.addElement(exportedFAL);
    }

    private String getText(String lexiconMessage, Object... obj) {
        return HonPlantControllerHMILexicon.getLexicon().getText(lexiconMessage, obj);
    }

    private BBacnetExportTable exportTable;
    private BHonDeviceExtDescriptor deviceExtDescriptor;
}
