/*
Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
This file contains trade secrets of Honeywell International, Inc. No part
may be reproduced or transmitted in any form by any means or for any
purpose without the express written permission of Honeywell.
*/
package com.honeywell.honplantcontroller.hmi.utils;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.stream.Stream;

import javax.baja.alarm.BAlarmClass;
import javax.baja.alarm.ext.BAlarmSourceExt;
import javax.baja.alarm.ext.BOffnormalAlgorithm;
import javax.baja.alarm.ext.offnormal.BBooleanChangeOfStateAlgorithm;
import javax.baja.alarm.ext.offnormal.BBooleanCommandFailureAlgorithm;
import javax.baja.alarm.ext.offnormal.BEnumChangeOfStateAlgorithm;
import javax.baja.alarm.ext.offnormal.BEnumCommandFailureAlgorithm;
import javax.baja.bacnet.BBacnetNetwork;
import javax.baja.bacnet.enums.BBacnetObjectType;
import javax.baja.bacnet.export.BBacnetLoopDescriptor;
import javax.baja.bacnet.export.BBacnetNotificationClassDescriptor;
import javax.baja.bacnet.export.BBacnetPointDescriptor;
import javax.baja.bacnet.export.BIBacnetExportObject;
import javax.baja.bacnet.point.BBacnetProxyExt;
import javax.baja.collection.BITable;
import javax.baja.collection.TableCursor;
import javax.baja.control.BBooleanPoint;
import javax.baja.control.BBooleanWritable;
import javax.baja.control.BControlPoint;
import javax.baja.control.BEnumPoint;
import javax.baja.control.BEnumWritable;
import javax.baja.control.BNumericPoint;
import javax.baja.control.BNumericWritable;
import javax.baja.control.ext.BAbstractProxyExt;
import javax.baja.data.BIDataValue;
import javax.baja.naming.BOrd;
import javax.baja.sys.BComponent;
import javax.baja.sys.BEnumRange;
import javax.baja.sys.BFacets;
import javax.baja.sys.BIBoolean;
import javax.baja.sys.BIEnum;
import javax.baja.sys.BINumeric;
import javax.baja.sys.BObject;
import javax.baja.sys.BString;
import javax.baja.sys.LocalizableRuntimeException;
import javax.baja.sys.Type;
import javax.baja.tag.Id;
import javax.baja.tag.Tags;
import javax.baja.util.BTypeSpec;

import com.honeywell.honbacnethelper.export.BHonBacnetAnalogInputDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetAnalogOutputDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetAnalogValueDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetAnalogValuePrioritizedDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetBinaryInputDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetBinaryOutputDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetBinaryValueDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetBinaryValuePrioritizedDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetMultiStateInputDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetMultiStateOutputDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetMultiStateValueDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetMultiStateValuePrioritizedDescriptor;
import com.honeywell.honbacnethelper.export.BIHonBacnetCustomDescriptor;
import com.honeywell.honplantcontroller.hmi.enums.BHMIPointExportOptions;
import com.honeywell.honplantcontroller.hmi.job.BCommissionHMIJob;
import com.honeywell.honplantcontroller.hmi.limitations.HonHMILimitations;
import com.honeywell.honplantcontroller.hmi.view.BFastAccessList;
import com.tridium.bacnet.stack.server.BBacnetExportTable;
import com.tridium.kitControl.BLoopPoint;

public class HonPlantControllerBacnetPointExportUtils {

    private HonPlantControllerBacnetPointExportUtils() {
        //This is a utility-class, so there is no need to instantiate it.
    }


    /**
     * Helper-Method to export all control points in given BacnetNetwork
     *
     * @param bacnetNetwork
     */
    public static Optional<Map<String, Map<String, List<String>>>> exportPoints(BBacnetNetwork bacnetNetwork, BHMIPointExportOptions pointExportOption) {

        boolean limitReached = false;

        Optional<Map<String, Map<String, List<String>>>> result = Optional.empty();

        if (bacnetNetwork == null) {
            HonPlantControllerHMILogger.checkAndLog(Level.FINE, "Could not export points, bacnetNetwork was null");
            return result;
        }

        BBacnetExportTable exportTable = (BBacnetExportTable) bacnetNetwork.getLocalDevice().getExportTable();

        HonPlantControllerBacnetScheduleExportUtils.removeInvalidDescriptors(exportTable);

        final BOrd ord = BOrd.make("slot:/|bql:select * from " + "control:ControlPoint");
        final BITable<BControlPoint> table = (BITable<BControlPoint>) ord.resolve(bacnetNetwork).get();
        final TableCursor<BControlPoint> c = table.cursor();

        Stream<BControlPoint> points = c.stream();
        Stream<BControlPoint> end = points;

        // BStringPoint and BStringWritable are not supported by HMI

        //intialize Optional error/warning lists
        errorTypesAndErrors = new HashMap<>();

        dpTypesAndAlarmErrors = new HashMap<>();
        dpTypesAndPrecisionErrors = new HashMap<>();
        dpEnumRangeErrors = new HashMap<>();
        dpObjectNameWarnings = new HashMap<>();

        AtomicInteger counter = new AtomicInteger(0);

        switch (pointExportOption.getOrdinal()) {
            case BHMIPointExportOptions.ALL_CONTROL_POINTS:
                end = points.filter(point -> true);

                break;
            case BHMIPointExportOptions.ALL_DATA_POINTS:
                end = points.filter(point -> HonPlantControllerBacnetPointExportUtils.isPointInFAL(point) || HonPlantControllerBacnetPointExportUtils.isPlainControllPoint(point) || HonPlantControllerBacnetPointExportUtils.isPanelbusPoint(point));
                break;

            case BHMIPointExportOptions.TAGGED_POINTS:
                end = points.filter(point -> HonPlantControllerBacnetPointExportUtils.isPointInFAL(point) || HonPlantControllerBacnetPointExportUtils.isPointSelected(point));

                break;
            case BHMIPointExportOptions.ONLY_FAL:
                end = points.filter(HonPlantControllerBacnetPointExportUtils::isPointInFAL);
                break;

            default:
                break;
        }

        if (pointExportOption.equals(BHMIPointExportOptions.allControlPoints)) {
            limitReached = end.map(point -> {
                exportControlPoint(exportTable, counter, point);
                return counter.get() >= HonHMILimitations.getMaxNumberOfExportedPoints();
            }).anyMatch(it -> it);

        } else {

            limitReached = end.map(point -> {
                exportDataPoint(exportTable, counter, point);
                return counter.get() >= HonHMILimitations.getMaxNumberOfExportedPoints();
            }).anyMatch(it -> it);
        }

        if (limitReached) {
            limitExceeded();
        }
        if (!dpTypesAndPrecisionErrors.isEmpty()) {
            errorTypesAndErrors.put(PRECISION_ERROR, dpTypesAndPrecisionErrors);
        }
        if (!dpTypesAndAlarmErrors.isEmpty()) {
            errorTypesAndErrors.put(ALARM_ERROR, dpTypesAndAlarmErrors);
        }
        if(!dpEnumRangeErrors.isEmpty()) {
            errorTypesAndErrors.put(ENUM_RANGE_ERROR, dpEnumRangeErrors);
        }
        if (!dpObjectNameWarnings.isEmpty()) {
            errorTypesAndErrors.put(OBJECT_NAME_WARNING, dpObjectNameWarnings);
        }

        if (!errorTypesAndErrors.isEmpty()) {
            result = Optional.of(errorTypesAndErrors);
        }

        return result;
    }

    /**
     * export a loop point as bacnet loop descriptor
     *
     * @param exportTable
     * @param controlPoint
     * @param exportedCount
     * @return
     */
    public static boolean exportLoopPoint(BBacnetExportTable exportTable, BControlPoint controlPoint, AtomicInteger exportedCount) {
        if (controlPoint instanceof BLoopPoint) {
            // has to be examined before BNumericPoint because it is derived from
            // BNumericPoint
            BIBacnetExportObject bacnetDescriptor = new BBacnetLoopDescriptor();
            HonPlantControllerBacnetScheduleExportUtils.exportBacnetObject(controlPoint, bacnetDescriptor, exportTable, BBacnetObjectType.LOOP, "loop", dpObjectNameWarnings);
            exportedCount.incrementAndGet();
            return true;
        }

        return false;
    }

    /**
     * check if control point is a plain point or writable (non-function block)
     *
     * @param controlPoint
     * @return
     */
    public static boolean isPlainControllPoint(BControlPoint controlPoint) {
        Type controlPointType = controlPoint.getType();

        if (controlPointType.equals(BNumericPoint.TYPE)
            || controlPointType.equals(BNumericWritable.TYPE)
            || controlPointType.equals(BBooleanPoint.TYPE)
            || controlPointType.equals(BBooleanWritable.TYPE)
            || controlPointType.equals(BEnumWritable.TYPE)
            || controlPointType.equals(BEnumPoint.TYPE)) {

            return true;
        }
        return false;
    }

    /**
     * export control points (function blocks ie. loop, ramp, sine wave and data points)
     *
     * @param exportTable
     * @param exportedCount
     * @param controlPoint
     */
    public static void exportControlPoint(BBacnetExportTable exportTable, AtomicInteger exportedCount, BControlPoint controlPoint) {
        if (!exportLoopPoint(exportTable, controlPoint, exportedCount)) {
            exportDataPoint(exportTable, exportedCount, controlPoint);
        }
    }


    /**
     * Export data points (numeric, boolean, enum) from station to HMI
     *
     * @param exportTable
     * @param exportedCount
     * @param controlPoint
     */
    public static void exportDataPoint(BBacnetExportTable exportTable, AtomicInteger exportedCount, BControlPoint controlPoint) {
        if (checkFittingAlarmExt(controlPoint)) {
            if (controlPoint instanceof BINumeric) {
                exportNumeric(controlPoint, exportTable);
                exportedCount.incrementAndGet();
            }
            else if(controlPoint instanceof BIBoolean) {
                exportBoolean(controlPoint, exportTable);
                exportedCount.incrementAndGet();
            }
            else if(controlPoint instanceof BIEnum) {
                exportEnum(controlPoint, exportTable);
                exportedCount.incrementAndGet();
            }
        }
    }

    /**
     * check if point is in a FAL
     *
     * @param controlPoint
     * @return
     */
    public static boolean isPointInFAL(BControlPoint controlPoint) {

        Tags tags = controlPoint.tags();

        Optional<BIDataValue> falTag = tags.get(BFastAccessList._tagIdFALname);

        return falTag.isPresent();

    }

    /**
     * Check if point was marked for export to HMI
     *
     * @param controlPoint
     * @return
     */
    public static boolean isPointSelected(BControlPoint controlPoint) {

        Tags tags = controlPoint.tags();

        Optional<BIDataValue> falTag = tags.get(EXPORT_TO_HMI);

        return falTag.isPresent();

    }

    private static void exportPrecisionValue(BControlPoint sourceExportComponent, BIBacnetExportObject bacnetExportObject) {
        if ((bacnetExportObject instanceof BIHonBacnetCustomDescriptor) && (sourceExportComponent instanceof BNumericPoint)) {
            BFacets facets = sourceExportComponent.getFacets();
            int precision = facets.geti(BFacets.PRECISION, -1);
            if (precision < 0 || precision > 5) {
                //log this as warning
                buildJoblogForPrecissionError(sourceExportComponent);
            }
            //set precision to limit of 0 or 5
            precision = (precision < 0) ? 0: (precision > 5) ? 5: precision;

            ((BIHonBacnetCustomDescriptor)bacnetExportObject).setPointPrecision(precision);
        }
        
    }

    private static void populateErrorMap(Map<String, List<String>> dpTypesAndDps, BControlPoint controlPoint, String pointTypeName) {
        if ((null == pointTypeName) || (null == dpTypesAndDps) || (null == controlPoint)) {
            return;
        }

        dpTypesAndDps.put("[" + controlPoint.getSlotPath() + "]", Collections.singletonList(pointTypeName));
    }

    public static boolean enumRangeStartingWithZero(BIEnum enumPoint) {
        try {
            BFacets facets = enumPoint.getEnumFacets();
            BObject range = facets.get("range");
            if (range instanceof BEnumRange) {
                int[] ordinals = ((BEnumRange) range).getOrdinals();
                if (null != ordinals && ordinals[0] == 0) {
                    return true;
                }
            }
        } catch (Exception e) {
            //range is not defined or starting from 0
            HonPlantControllerHMILogger.checkAndLog(Level.SEVERE, "enumRangeStartingWithZero.exception: range is not defined or starting from 0");
        }

        return false;
    }


    /**
     * Helper function to export numeric points
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportNumeric(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        if (isPanelbusPoint(controlPoint)) {
            exportPanelbusNumeric(controlPoint, exportTable);
        } else if (isBacnetPoint(controlPoint)) {
            exportBacnetNumeric(controlPoint, exportTable);
        } else {
            exportSoftwareNumeric(controlPoint, exportTable);
        }
    }

    /**
     * Helper function to export software numeric points
     *  @param controlPoint
     * @param exportTable
     */
    public static void exportSoftwareNumeric(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        if (controlPoint instanceof BNumericWritable) {
            exportAsAnalogPrioritizedValue(controlPoint, exportTable);
            return;
        }

        exportAsAnalogValue(controlPoint, exportTable);
    }

    /**
     * Helper function to export boolean points
     *  @param controlPoint
     * @param exportTable
     */
    public static void exportBoolean(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        if (isPanelbusPoint(controlPoint)) {
            exportPanelbusBoolean(controlPoint, exportTable);
        } else if (isBacnetPoint(controlPoint)) {
            exportBacnetBoolean(controlPoint, exportTable);
        } else {
            exportSoftwareBoolean(controlPoint, exportTable);
        }
    }

    /**
     * Helper function to export software boolean points
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportSoftwareBoolean(BControlPoint controlPoint, BBacnetExportTable exportTable) {

        if (controlPoint instanceof BBooleanWritable) {
            exportAsBooleanPrioritizedValue(controlPoint, exportTable);
            return;
        }

        exportAsBooleanValue(controlPoint, exportTable);
    }

    /**
     * Helper function to export enum points
     *  @param controlPoint
     * @param exportTable
     */
    public static void exportEnum(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        if (!validateEnumRange((BIEnum)controlPoint)) {
            //The point should be skipped!
            buildJobLogForEnumRangeError(controlPoint);
            return;
        }

        if (isBacnetPoint(controlPoint)) {
            exportBacnetEnum(controlPoint, exportTable);
        } else {
            exportSoftwareEnum(controlPoint, exportTable);
        }
    }

    /**
     * Helper function to export software enum points
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportSoftwareEnum(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        if (controlPoint instanceof BEnumWritable) {
            exportAsEnumPrioritizedValue(controlPoint, exportTable);
            return;
        }
        exportAsEnumValue(controlPoint, exportTable);
    }


    public static boolean isPanelbusPoint(BControlPoint controlPoint) {
        return controlPoint.getProxyExt().getType().is(BTypeSpec.make("clPanelBus:PanelbusProxyExt").getTypeInfo());
    }

    public static boolean isBacnetPoint(BControlPoint controlPoint) {
        return controlPoint.getProxyExt() instanceof BBacnetProxyExt;
    }

    /**
     * Helper function to export Panelbus Analog points
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportPanelbusNumeric(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BAbstractProxyExt proxyExt = controlPoint.getProxyExt();

        if (proxyExt.getType().is(BTypeSpec.make("clPanelBus:PanelbusInputProxyExt").getTypeInfo())) {
            exportAsAnalogInput(controlPoint, exportTable);
        } else if (proxyExt.getType().is(BTypeSpec.make("clPanelBus:PanelbusOutputProxyExt").getTypeInfo())) {
            exportAsAnalogOutput(controlPoint, exportTable);
        }
    }

    /**
     * Helper function to export Panelbus Boolean points
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportPanelbusBoolean(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BAbstractProxyExt proxyExt = controlPoint.getProxyExt();

        if (proxyExt.getType().is(BTypeSpec.make("clPanelBus:PanelbusInputProxyExt").getTypeInfo())) {
            exportAsBooleanInput(controlPoint, exportTable);
        } else if (proxyExt.getType().is(BTypeSpec.make("clPanelBus:PanelbusOutputProxyExt").getTypeInfo())) {
            exportAsBooleanOutput(controlPoint, exportTable);
        }
    }

    /**
     * Helper function to export Bacnet Analog data points
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportBacnetNumeric(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BAbstractProxyExt proxyExt = controlPoint.getProxyExt();

        BBacnetProxyExt bacnetProxyExt = (BBacnetProxyExt) proxyExt;

        int objectType = bacnetProxyExt.getObjectId().getObjectType();

        switch (objectType) {
            case BBacnetObjectType.ANALOG_INPUT:
                exportAsAnalogInput(controlPoint, exportTable);
                break;

            case BBacnetObjectType.ANALOG_OUTPUT:
                exportAsAnalogOutput(controlPoint, exportTable);
                break;

            case BBacnetObjectType.ANALOG_VALUE:
                exportAsAnalogValue(controlPoint, exportTable);
                break;

            default:
                break;
        }
    }

    /**
     * Helper function to export Bacnet Binary data points
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportBacnetBoolean(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BAbstractProxyExt proxyExt = controlPoint.getProxyExt();

        BBacnetProxyExt bacnetProxyExt = (BBacnetProxyExt) proxyExt;

        int objectType = bacnetProxyExt.getObjectId().getObjectType();

        switch (objectType) {
            case BBacnetObjectType.BINARY_INPUT:
                exportAsBooleanInput(controlPoint, exportTable);
                break;

            case BBacnetObjectType.BINARY_OUTPUT:
                exportAsBooleanOutput(controlPoint, exportTable);
                break;

            case BBacnetObjectType.BINARY_VALUE:
                exportAsBooleanValue(controlPoint, exportTable);
                break;

            default:
                break;
        }
    }

    /**
     * Helper function to export Bacnet Multi state data points
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportBacnetEnum(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BAbstractProxyExt proxyExt = controlPoint.getProxyExt();

        BBacnetProxyExt bacnetProxyExt = (BBacnetProxyExt) proxyExt;

        int objectType = bacnetProxyExt.getObjectId().getObjectType();

        switch (objectType) {
            case BBacnetObjectType.MULTI_STATE_INPUT:
                exportAsEnumInput(controlPoint, exportTable);
                break;

            case BBacnetObjectType.MULTI_STATE_OUTPUT:
                exportAsEnumOutput(controlPoint, exportTable);
                break;

            case BBacnetObjectType.MULTI_STATE_VALUE:
                exportAsEnumValue(controlPoint, exportTable);
                break;

            default:
                break;
        }
    }


    /**
     * Helper function to export a control point using Analog Input Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsAnalogInput(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetAnalogInputDescriptor();
        String nameInExportTable = "analogInput";
        int bacnetObjectType = BBacnetObjectType.ANALOG_INPUT;

        exportPrecisionValue(controlPoint, bacnetDescriptor);
        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    /**
     * Helper function to export a control point using Analog Output Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsAnalogOutput(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetAnalogOutputDescriptor();
        String nameInExportTable = "analogOutput";
        int bacnetObjectType = BBacnetObjectType.ANALOG_OUTPUT;

        exportPrecisionValue(controlPoint, bacnetDescriptor);
        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    /**
     * Helper function to export a control point using Analog Value Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsAnalogValue(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetAnalogValueDescriptor();
        String nameInExportTable = "analogValue";
        int bacnetObjectType = BBacnetObjectType.ANALOG_VALUE;

        exportPrecisionValue(controlPoint, bacnetDescriptor);
        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    /**
     * Helper function to export a control point using Analog Prioritized Value Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsAnalogPrioritizedValue(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetAnalogValuePrioritizedDescriptor();
        String nameInExportTable = "analogValue";
        int bacnetObjectType = BBacnetObjectType.ANALOG_VALUE;

        exportPrecisionValue(controlPoint, bacnetDescriptor);
        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    /**
     * Helper function to export a control point using Binary Input Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsBooleanInput(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetBinaryInputDescriptor();
        String nameInExportTable = "binaryInput";
        int bacnetObjectType = BBacnetObjectType.BINARY_INPUT;

        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    /**
     * Helper function to export a control point using Binary Output Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsBooleanOutput(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetBinaryOutputDescriptor();
        String nameInExportTable = "binaryOutput";
        int bacnetObjectType = BBacnetObjectType.BINARY_OUTPUT;

        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    /**
     * Helper function to export a control point using Binary Value Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsBooleanValue(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetBinaryValueDescriptor();
        String nameInExportTable = "binaryValue";
        int bacnetObjectType = BBacnetObjectType.BINARY_VALUE;

        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    /**
     * Helper function to export a control point using Binary Prioritized Value Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsBooleanPrioritizedValue(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetBinaryValuePrioritizedDescriptor();
        String nameInExportTable = "binaryValue";
        int bacnetObjectType = BBacnetObjectType.BINARY_VALUE;

        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    /**
     * Helper function to export a control point using Multi State Input Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsEnumInput(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetMultiStateInputDescriptor();
        String nameInExportTable = "multiStateInput";
        int bacnetObjectType = BBacnetObjectType.MULTI_STATE_INPUT;

        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    /**
     * Helper function to export a control point using Multi State Output Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsEnumOutput(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetMultiStateOutputDescriptor();
        String nameInExportTable = "multiStateOutput";
        int bacnetObjectType = BBacnetObjectType.MULTI_STATE_OUTPUT;

        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    /**
     * Helper function to export a control point using Multi State Value Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsEnumValue(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetMultiStateValueDescriptor();
        String nameInExportTable = "multiStateValue";
        int bacnetObjectType = BBacnetObjectType.MULTI_STATE_VALUE;

        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    /**
     * Helper function to export a control point using Multi State Prioritized Value Descriptor
     *
     * @param controlPoint
     * @param exportTable
     */
    public static void exportAsEnumPrioritizedValue(BControlPoint controlPoint, BBacnetExportTable exportTable) {
        BIBacnetExportObject bacnetDescriptor = new BHonBacnetMultiStateValuePrioritizedDescriptor();
        String nameInExportTable = "multiStateValue";
        int bacnetObjectType = BBacnetObjectType.MULTI_STATE_VALUE;

        exportBacnetPointObject(controlPoint, bacnetDescriptor, exportTable, bacnetObjectType, nameInExportTable);
    }

    public static void buildJobLogForEnumRangeError(BControlPoint controlPoint) {
        String pointTypeName = (controlPoint instanceof BEnumWritable) ? "EnumWritable" : "EnumPoint";
        populateErrorMap(dpEnumRangeErrors, controlPoint, pointTypeName);
    }

    public static boolean validateEnumRange(BIEnum enumPoint) {
        if (enumRangeStartingWithZero(enumPoint)) {
            return false;
        }

        BFacets facets = enumPoint.getEnumFacets();

        BObject rangeFacet = facets.get("range");

        if (rangeFacet instanceof BEnumRange) {
            BEnumRange enumRange = (BEnumRange) rangeFacet;
            return !enumRange.isNull();
        }

        return false;
    }

    public static void buildJoblogForPrecissionError(BControlPoint controlPoint) {
        if (null == controlPoint || null == dpTypesAndPrecisionErrors) {
            return;
        }

        String pointTypeName = (controlPoint instanceof BNumericWritable) ? "NumericWritable" : "NumericPoint";
        populateErrorMap(dpTypesAndPrecisionErrors, controlPoint, pointTypeName);
    }

    public static boolean checkFittingAlarmExt(BControlPoint controlPoint) {
        if (null == controlPoint || null == dpTypesAndAlarmErrors) {
            return false;
        }

        String pointTypeName = null;

        if (controlPoint instanceof BBooleanWritable) {
            if(!checkFittingAlarmExt((BBooleanWritable)controlPoint)) {
                pointTypeName = "BooleanWritable";
            }
        } else if (controlPoint instanceof BBooleanPoint) {
            if(!checkFittingAlarmExt((BBooleanPoint)controlPoint) )  {
                pointTypeName = "BooleanPoint";
            }
        } else if (controlPoint instanceof BEnumWritable) {
            if(!checkFittingAlarmExt((BEnumWritable)controlPoint)) {
                pointTypeName = "EnumWritable";
            }

        } else if (controlPoint instanceof BEnumPoint) {
            if(!checkFittingAlarmExt((BEnumPoint)controlPoint)) {
                pointTypeName = "EnumPoint";
            }
        }

        if (pointTypeName != null)
            populateErrorMap(dpTypesAndAlarmErrors, controlPoint, pointTypeName);

        return (pointTypeName == null);
    }

    public static boolean checkFittingAlarmExt(BBooleanWritable booleanWritable) {

        BAlarmSourceExt[] alarmSourceExts = booleanWritable.getChildren(BAlarmSourceExt.class);

        //make sure to check all alarmSources under a point, and break as soon as an invalid source is found
        for (BAlarmSourceExt bAlarmSourceExt : alarmSourceExts) {
            BOffnormalAlgorithm offnormalAlgorithm = bAlarmSourceExt.getOffnormalAlgorithm();

            boolean isHardwarePoint = isHardwarePoint(booleanWritable);
            if (!isHardwarePoint && !(offnormalAlgorithm instanceof  BBooleanChangeOfStateAlgorithm)) {
                return false;
            }
            if (isHardwarePoint && !(offnormalAlgorithm instanceof BBooleanCommandFailureAlgorithm)) {
                return false;
            }
        }

        return true;
    }

    public static boolean checkFittingAlarmExt(BBooleanPoint booleanPoint) {

        BAlarmSourceExt[] alarmSourceExts = booleanPoint.getChildren(BAlarmSourceExt.class);

        for (BAlarmSourceExt bAlarmSourceExt : alarmSourceExts) {
            BOffnormalAlgorithm offnormalAlgorithm = bAlarmSourceExt.getOffnormalAlgorithm();
            if (!(offnormalAlgorithm instanceof BBooleanChangeOfStateAlgorithm)) {
                return false;
            }
        }

        return true;
    }

    public static boolean checkFittingAlarmExt(BEnumWritable enumWritable) {

        BAlarmSourceExt[] alarmSourceExts = enumWritable.getChildren(BAlarmSourceExt.class);

        for (BAlarmSourceExt bAlarmSourceExt : alarmSourceExts) {
            BOffnormalAlgorithm offnormalAlgorithm = bAlarmSourceExt.getOffnormalAlgorithm();
            boolean isHardwarePoint = isHardwarePoint(enumWritable);

            if (!isHardwarePoint && !(offnormalAlgorithm instanceof BEnumChangeOfStateAlgorithm)) {
                return false;
            }
            if (isHardwarePoint && !(offnormalAlgorithm instanceof  BEnumCommandFailureAlgorithm)) {
                return false;
            }
        }
        return true;
    }

    public static boolean checkFittingAlarmExt(BEnumPoint enumPoint) {

        BAlarmSourceExt[] alarmSourceExts = enumPoint.getChildren(BAlarmSourceExt.class);

        for (BAlarmSourceExt bAlarmSourceExt : alarmSourceExts) {
            BOffnormalAlgorithm offnormalAlgorithm = bAlarmSourceExt.getOffnormalAlgorithm();
            if (!(offnormalAlgorithm instanceof BEnumChangeOfStateAlgorithm)) {
                return false;
            }
        }
        return true;
    }

    public static boolean checkEnumRange(BControlPoint controlPoint) {
        if (controlPoint instanceof BEnumPoint){
            return validateEnumRange((BIEnum) controlPoint);
        }

        return true;
    }

    public static boolean exportAlarmClasses(BBacnetNetwork bacnetNetwork) {

        if (bacnetNetwork == null) {
            HonPlantControllerHMILogger.checkAndLog(Level.FINE, "Could not export alarm classes, bacnetNetwork was null");
            return false;
        }

        BBacnetExportTable exportTable = (BBacnetExportTable) bacnetNetwork.getLocalDevice().getExportTable();

        final BOrd ord = BOrd.make("slot:/|bql:select * from " + "alarm:AlarmClass");
        final BITable<BAlarmClass> table = (BITable<BAlarmClass>) ord.resolve(bacnetNetwork).get();
        final TableCursor<BAlarmClass> c = table.cursor();

        // BStringPoint and BStringWritable are not supported by HMI

        int i = 0;
        for (BAlarmClass alarmClass : c) {
            BIBacnetExportObject bacnetDescriptor = new BBacnetNotificationClassDescriptor();
            HonPlantControllerBacnetScheduleExportUtils.exportAlarmClass(alarmClass, bacnetDescriptor, exportTable, BBacnetObjectType.NOTIFICATION_CLASS, alarmClass.getName(), i);
            ++i;
        }
        return true;
    }

    public static boolean exportBacnetPointObject(BComponent sourceExportComponent, BIBacnetExportObject bacnetExportObject, BBacnetExportTable exportTable, int objectType, String nameInExportTable) {
        boolean success = HonPlantControllerBacnetScheduleExportUtils.exportBacnetObject(sourceExportComponent, bacnetExportObject, exportTable, objectType, nameInExportTable, dpObjectNameWarnings);
        if (success) {
            ((BBacnetPointDescriptor) bacnetExportObject).makeWritable(BString.make("in1,in8"));
        }
        return success;
    }

    protected static void limitExceeded()
        throws LocalizableRuntimeException {

        throw new LocalizableRuntimeException(BCommissionHMIJob.TYPE.getModule().getModuleName(),
            "hmi.limitations.exceeded.point.max.exportedPoints",
            new Object[]{HonHMILimitations.getMaxNumberOfExportedPoints()});
        
    }
    public static boolean isHardwarePoint(BControlPoint controlPoint) {
        return isPanelbusPoint(controlPoint) || isBacnetPoint(controlPoint);
    }

    //Attributes
    public static final String ALARM_ERROR = "Alarm Error";
    public static final String PRECISION_ERROR = "Precision Error";
    public static final String ENUM_RANGE_ERROR = "Enum Range Error";
    public static final String OBJECT_NAME_WARNING = "Object Name Mismatch";


    private static Map<String, Map<String, List<String>>> errorTypesAndErrors = new HashMap<>();

    private static Map<String, List<String>> dpTypesAndAlarmErrors = new HashMap<>();
    private static Map<String, List<String>> dpTypesAndPrecisionErrors = new HashMap<>();
    private static Map<String, List<String>> dpEnumRangeErrors = new HashMap<>();
    private static Map<String, List<String>> dpObjectNameWarnings = new HashMap<>();
    public static final String HMI_POINT_EXPORT_NAMESPACE = "honHMI";
    public static final String HMI_POINT_EXPORT_TAG = "HMIPoint";
    public static final Id EXPORT_TO_HMI = Id.newId(HMI_POINT_EXPORT_NAMESPACE + ":" + HMI_POINT_EXPORT_TAG);
}
