/*
Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
This file contains trade secrets of Honeywell International, Inc. No part
may be reproduced or transmitted in any form by any means or for any
purpose without the express written permission of Honeywell.
*/


package com.honeywell.honplantcontroller.hmi.utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.baja.bacnet.BBacnetNetwork;
import javax.baja.bacnet.datatypes.BBacnetArray;
import javax.baja.bacnet.datatypes.BBacnetBitString;
import javax.baja.bacnet.datatypes.BBacnetDeviceObjectPropertyReference;
import javax.baja.bacnet.datatypes.BBacnetListOf;
import javax.baja.bacnet.datatypes.BBacnetObjectIdentifier;
import javax.baja.bacnet.enums.BBacnetObjectType;
import javax.baja.bacnet.export.BBacnetBooleanScheduleDescriptor;
import javax.baja.bacnet.export.BBacnetCalendarDescriptor;
import javax.baja.bacnet.export.BBacnetPointDescriptor;
import javax.baja.bacnet.export.BBacnetScheduleDescriptor;
import javax.baja.bacnet.export.BIBacnetExportObject;
import javax.baja.bacnet.export.BLocalBacnetDevice;
import javax.baja.collection.BITable;
import javax.baja.collection.TableCursor;
import javax.baja.control.BControlPoint;
import javax.baja.naming.BOrd;
import javax.baja.naming.SlotPath;
import javax.baja.schedule.BAbstractSchedule;
import javax.baja.schedule.BBooleanSchedule;
import javax.baja.schedule.BCalendarSchedule;
import javax.baja.schedule.BEnumSchedule;
import javax.baja.schedule.BNumericSchedule;
import javax.baja.schedule.BStringSchedule;
import javax.baja.schedule.BWeeklySchedule;
import javax.baja.sys.BComplex;
import javax.baja.sys.BComponent;
import javax.baja.sys.BDynamicEnum;
import javax.baja.sys.BEnumRange;
import javax.baja.sys.BFacets;
import javax.baja.sys.BObject;
import javax.baja.sys.BString;
import javax.baja.sys.Context;
import javax.baja.sys.Knob;
import javax.baja.sys.Property;
import javax.baja.sys.Type;
import javax.baja.util.BFormat;

import com.honeywell.honbacnethelper.export.BHonBacnetAnalogInputDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetAnalogOutputDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetAnalogValueDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetAnalogValuePrioritizedDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetBinaryOutputDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetBooleanScheduleDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetEnumScheduleDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetMultiStateOutputDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetNumericScheduleDescriptor;
import com.honeywell.honbacnethelper.export.BHonBacnetStringScheduleDescriptor;
import com.honeywell.honbacnethelper.export.BIHonBacnetCustomDescriptor;
import com.honeywell.honbacnethelper.util.HonBacnetScheduleUtil;
import com.honeywell.honplantcontroller.hmi.comm.BIHonPlantControllerHMIInterface;
import com.honeywell.honplantcontroller.hmi.limitations.HonHMILimitations;
import com.honeywell.honplantcontroller.hmi.limitations.HonLimitationResult;
import com.honeywell.honplantcontroller.hmi.service.BHonPlantControllerHMI;
import com.tridium.bacnet.stack.server.BBacnetExportTable;

public class HonPlantControllerBacnetScheduleExportUtils {

    private HonPlantControllerBacnetScheduleExportUtils() {
        //This is a utility-class, so there is no need to instanciate it.
    }

    /**
     * Helper-Method for exporting a Weekly-Schedule using Bacnet
     *
     * @param schedule
     * @param bacnetScheduleDescriptor
     * @param exportTable
     */
    public static void exportSchedule(BAbstractSchedule schedule, BIBacnetExportObject bacnetScheduleDescriptor, BBacnetExportTable exportTable) {
        Map<String, List<String>> dpObjectNameWarnings = new HashMap<>();
        exportSchedule(schedule, bacnetScheduleDescriptor, exportTable, dpObjectNameWarnings);
    }

    public static void exportSchedule(BAbstractSchedule schedule, BIBacnetExportObject bacnetScheduleDescriptor, BBacnetExportTable exportTable, Map<String, List<String>> dpObjectNameWarnings) {
        exportBacnetObject(schedule, bacnetScheduleDescriptor, exportTable, BBacnetObjectType.SCHEDULE, "schedule", dpObjectNameWarnings);
    }

    /**
     * Helper-Method for exporting a Calender-Schedule using Bacnet
     *
     * @param calendar
     * @param bacnetCalendarDescriptor
     * @param exportTable
     */
    public static void exportCalendar(BAbstractSchedule calendar, BIBacnetExportObject bacnetCalendarDescriptor, BBacnetExportTable exportTable) {
        Map<String, List<String>> dpObjectNameWarnings = new HashMap<>();
        exportCalendar(calendar, bacnetCalendarDescriptor, exportTable, dpObjectNameWarnings);
    }

    public static void exportCalendar(BAbstractSchedule calendar, BIBacnetExportObject bacnetCalendarDescriptor, BBacnetExportTable exportTable, Map<String, List<String>> dpObjectNameWarnings) {
        exportBacnetObject(calendar, bacnetCalendarDescriptor, exportTable, BBacnetObjectType.CALENDAR, "calendar", dpObjectNameWarnings);
    }

    public static String convertSlotPathToObjectName(SlotPath slotPath) {
        return slotPath.toString().replace(SLOT, "").replace("/", ".");
    }

    public static boolean exportBacnetObject(BComponent sourceExportComponent, BIBacnetExportObject bacnetExportObject, BBacnetExportTable exportTable, int objectType, String nameInExportTable, Map<String, List<String>> dpObjectNameWarnings) {
        BHonPlantControllerHMI hmiDevice = getHMIDevice();
        if (hmiDevice == null) {
            HonPlantControllerHMILogger.checkAndLog(Level.SEVERE, "[exportBacnetObject] HMI Device is not present!");
            return false;
        }

        BOrd sourceExportComponentOrd = sourceExportComponent.getHandleOrd();
        // BBacnetObjectIdentifier foundExportedIdentifier = exportTable.byObjectOrd(sourceExportComponentOrd);
        BBacnetObjectIdentifier foundExportedIdentifier = findObjectIdForOrd(exportTable, sourceExportComponentOrd);
        if (foundExportedIdentifier != null) {
            HonPlantControllerHMILogger.checkAndLog(Level.FINE, "Object is already exported bacnetIdentifier" + foundExportedIdentifier + "current object name " + sourceExportComponent.getName());
            //update objectName only as a result to a rename action
            BIBacnetExportObject byObjectId = exportTable.byObjectId(foundExportedIdentifier);
            BIBacnetExportObject byObjectName = exportTable.byObjectName(sourceExportComponent.getName());
			
            // Check and update the description for already exported object
			BComponent proxyExt = (BComponent) sourceExportComponent.get(PROXY_EXT);
			if (null != proxyExt && null != proxyExt.get(DESCRIPTION)) {
				((BComplex) byObjectId).set(DESCRIPTION, proxyExt.get(DESCRIPTION));
			}
			
            //Update StateText of Schedule/Calendar Objects
            if (byObjectId instanceof BHonBacnetEnumScheduleDescriptor
                && bacnetExportObject instanceof BHonBacnetEnumScheduleDescriptor) {
                BHonBacnetEnumScheduleDescriptor descriptor = (BHonBacnetEnumScheduleDescriptor) exportTable.byObjectId(foundExportedIdentifier);
                descriptor.setScheduleStateText((BBacnetArray) ((BHonBacnetEnumScheduleDescriptor) bacnetExportObject).getScheduleStateText().newCopy());
            } else if (byObjectId instanceof BHonBacnetBooleanScheduleDescriptor
                && bacnetExportObject instanceof BHonBacnetBooleanScheduleDescriptor) {
                BHonBacnetBooleanScheduleDescriptor descriptor = (BHonBacnetBooleanScheduleDescriptor) exportTable.byObjectId(foundExportedIdentifier);
                descriptor.setScheduleStateText((BBacnetArray) ((BHonBacnetBooleanScheduleDescriptor) bacnetExportObject).getScheduleStateText().newCopy());
                ;
            }

            if (byObjectId instanceof BHonBacnetAnalogInputDescriptor && bacnetExportObject instanceof BHonBacnetAnalogInputDescriptor) {
                BHonBacnetAnalogInputDescriptor descriptor = (BHonBacnetAnalogInputDescriptor)exportTable.byObjectId(foundExportedIdentifier);
                descriptor.setPointPrecision(((BHonBacnetAnalogInputDescriptor)bacnetExportObject).getPointPrecision());
            } else if (byObjectId instanceof BHonBacnetAnalogOutputDescriptor && bacnetExportObject instanceof BHonBacnetAnalogOutputDescriptor) {
                BHonBacnetAnalogOutputDescriptor descriptor = (BHonBacnetAnalogOutputDescriptor)exportTable.byObjectId(foundExportedIdentifier);
                descriptor.setPointPrecision(((BHonBacnetAnalogOutputDescriptor)bacnetExportObject).getPointPrecision());
            } else if (byObjectId instanceof BHonBacnetAnalogValueDescriptor && bacnetExportObject instanceof BHonBacnetAnalogValueDescriptor) {
                BHonBacnetAnalogValueDescriptor descriptor = (BHonBacnetAnalogValueDescriptor)exportTable.byObjectId(foundExportedIdentifier);
                descriptor.setPointPrecision(((BHonBacnetAnalogValueDescriptor)bacnetExportObject).getPointPrecision());
            } else if (byObjectId instanceof BHonBacnetAnalogValuePrioritizedDescriptor && bacnetExportObject instanceof BHonBacnetAnalogValuePrioritizedDescriptor) {
                BHonBacnetAnalogValuePrioritizedDescriptor descriptor = (BHonBacnetAnalogValuePrioritizedDescriptor)exportTable.byObjectId(foundExportedIdentifier);
                descriptor.setPointPrecision(((BHonBacnetAnalogValuePrioritizedDescriptor)bacnetExportObject).getPointPrecision());
            }

            //Check if objectName should be updated, then raise a Notification to user
            if ((sourceExportComponent instanceof BControlPoint) || (sourceExportComponent instanceof BWeeklySchedule) || (sourceExportComponent instanceof BCalendarSchedule)) {
                BFormat format = (sourceExportComponent instanceof BControlPoint) ? hmiDevice.getPointName() :
                    (sourceExportComponent instanceof BWeeklySchedule) ? hmiDevice.getScheduleName() : hmiDevice.getCalendarName();

                //1. Try to use BFormat to rename the Point
                String objectName = format.format(sourceExportComponent);
                if (!byObjectId.getObjectName().equals(objectName)) {
                    //2. Point Name not identical to BFormat, check source component name
                    objectName = SlotPath.unescape(sourceExportComponent.getName());
                    if (!byObjectId.getObjectName().equals(SlotPath.unescape(objectName))) {
                        //3. Point Name not equal to source component name, use the SlotPath for uniqueness of object Name
                        objectName = SlotPath.unescape(sourceExportComponent.getSlotPath().toString());
						if (objectName.startsWith(SLOT)) {
							objectName = convertSlotPathToObjectName(sourceExportComponent.getSlotPath());
						}
                        if (!byObjectId.getObjectName().equals(SlotPath.unescape(objectName))) {
                            //The object objectName needs to be updated! Try again to use steps 1, 2 and 3
                            objectName = format.format(sourceExportComponent);
                            byObjectName = exportTable.byObjectName(objectName);
                            if (byObjectName != null) {
                                //2. Point Name Duplicates, try using source component name
                                objectName = SlotPath.unescape(sourceExportComponent.getName());
                                byObjectName = exportTable.byObjectName(objectName);
                                if (byObjectName != null) {
                                    //3. Point Name duplicate, use the SlotPath for uniqueness of object Name
                                    objectName = SlotPath.unescape(sourceExportComponent.getSlotPath().toString());
									if (objectName.startsWith(SLOT)) {
										objectName = convertSlotPathToObjectName(sourceExportComponent.getSlotPath());
									}
                                }
                            }
                        }
                    }

                    if (!byObjectId.getObjectName().equals(objectName)) {
                        dpObjectNameWarnings.put("[" + sourceExportComponent.getSlotPath() + "]", Collections.singletonList(byObjectId.getObjectName()));
                    }
                }
            } else {
                if (!(byObjectId.getObjectName().equals(SlotPath.unescape(sourceExportComponent.getName()))) &&
                    !(byObjectId.getObjectName().equals(SlotPath.unescape(sourceExportComponent.getSlotPath().toString())))) {
                    String objectName = byObjectId.getObjectName();
                    dpObjectNameWarnings.put("[" + sourceExportComponent.getSlotPath() + "]", Collections.singletonList(objectName));
                }
            }

            return true;
        }

        int instanceNumer = exportTable.getNextInstance(objectType);
        bacnetExportObject.setObjectId(BBacnetObjectIdentifier.make(objectType, instanceNumer));
        bacnetExportObject.setObjectOrd(sourceExportComponent.getHandleOrd(), Context.NULL);
        //HMI display name should be configurable via BFormat
        //BFormat.make("%parent.name%")
        if ((sourceExportComponent instanceof BControlPoint) || (sourceExportComponent instanceof BWeeklySchedule) || (sourceExportComponent instanceof BCalendarSchedule)) {
            //1. Try to use BFormat to rename the Point
            BFormat format = (sourceExportComponent instanceof BControlPoint) ? hmiDevice.getPointName() :
                (sourceExportComponent instanceof BWeeklySchedule) ? hmiDevice.getScheduleName() : hmiDevice.getCalendarName();
            String objectName = format.format(sourceExportComponent);
            BIBacnetExportObject byObjectName = exportTable.byObjectName(objectName);
            if (byObjectName != null) {
                //2. Point Name Duplicates, try using source component name
                objectName = SlotPath.unescape(sourceExportComponent.getName());
                byObjectName = exportTable.byObjectName(objectName);
                if (byObjectName != null) {
                    //3. Point Name duplicate, use the SlotPath for uniqueness of object Name
                    objectName = SlotPath.unescape(sourceExportComponent.getSlotPath().toString());
					if (objectName.startsWith(SLOT)) {
						objectName = convertSlotPathToObjectName(sourceExportComponent.getSlotPath());
					}
                }
            }
            bacnetExportObject.setObjectName(objectName);
        } else {
            //This is an other component which is not a ControlPoint, not a Schedule and not a calendar
            if (exportTable.byObjectName(sourceExportComponent.getName()) == null) {
                bacnetExportObject.setObjectName(SlotPath.unescape(sourceExportComponent.getName()));
            } else {
                bacnetExportObject.setObjectName(SlotPath.unescape(convertSlotPathToObjectName(sourceExportComponent.getSlotPath())));
            }
        }
        if (sourceExportComponent instanceof BControlPoint && bacnetExportObject instanceof BBacnetPointDescriptor && ((BControlPoint) sourceExportComponent).getStatus().isOverridden()) {
            ((BBacnetPointDescriptor) bacnetExportObject).setBacnetStatusFlags(BBacnetBitString.make(new boolean[]{false, false, true, false}));
        }
        exportTable.export(bacnetExportObject);
        exportTable.add(nameInExportTable + "_" + instanceNumer, (BComponent) bacnetExportObject);
        
        return true;
    }

    public static boolean exportAlarmClass(BComponent sourceExportComponent, BIBacnetExportObject
        bacnetExportObject, BBacnetExportTable exportTable, int objectType, String nameInExportTable, int identifier) {

        BOrd sourceExportComponentOrd = sourceExportComponent.getHandleOrd();
        // BBacnetObjectIdentifier foundExportedIdentifier = exportTable.byObjectOrd(sourceExportComponentOrd);
        BBacnetObjectIdentifier foundExportedIdentifier = findObjectIdForOrd(exportTable, sourceExportComponentOrd);
        if (foundExportedIdentifier != null) {
            HonPlantControllerHMILogger.checkAndLog(Level.FINE, "Object is already exported bacnetIdentifier" + foundExportedIdentifier + "object name " + sourceExportComponent.getName());
            return false;
        }

        int instanceNumer = exportTable.getNextInstance(objectType);
        bacnetExportObject.setObjectId(BBacnetObjectIdentifier.make(objectType, instanceNumer));
        bacnetExportObject.setObjectOrd(sourceExportComponent.getHandleOrd(), Context.NULL);
        if (exportTable.byObjectName(sourceExportComponent.getName()) == null) {
            bacnetExportObject.setObjectName(SlotPath.unescape(sourceExportComponent.getName()));
        } else {
            bacnetExportObject.setObjectName(SlotPath.unescape(convertSlotPathToObjectName(sourceExportComponent.getSlotPath())));
        }
        exportTable.export(bacnetExportObject);
        exportTable.add(nameInExportTable + "_" + instanceNumer, (BComponent) bacnetExportObject);
        return true;
    }

    /**
     * helper function for finding objectId by sourceOrd
     *
     * @param exportTable
     * @param sourceExportComponentOrd
     * @return
     */
    public static BBacnetObjectIdentifier findObjectIdForOrd(BBacnetExportTable exportTable, BOrd
        sourceExportComponentOrd) {
        return Stream.of(exportTable.getChildren(BIBacnetExportObject.class)).filter(export -> export.getObjectOrd().equals(sourceExportComponentOrd)).map(BIBacnetExportObject::getObjectId).findFirst().orElse(null);
    }

    public static void removeInvalidDescriptors(BBacnetExportTable exportTable) {
        BBacnetExportTable[] exportTables = new BBacnetExportTable[]{exportTable};
        Set<BIHonBacnetCustomDescriptor> descriptors = Arrays.stream(exportTables[0].getChildComponents())
            .filter(BIHonBacnetCustomDescriptor.class::isInstance)
            .map(BIHonBacnetCustomDescriptor.class::cast)
            .collect(Collectors.toSet());

        Iterator<BIHonBacnetCustomDescriptor> iterator = descriptors.iterator();
        while (iterator.hasNext()) {
            BIHonBacnetCustomDescriptor descriptor = iterator.next();
            if (descriptor instanceof BBacnetPointDescriptor) {
                BControlPoint controlPoint = ((BBacnetPointDescriptor) descriptor).getPoint();
                if (controlPoint == null || controlPoint.getName() == null) {
                    exportTable.remove(((BBacnetPointDescriptor) descriptor).getName());
                }
            } else if (descriptor instanceof BBacnetScheduleDescriptor) {
                BWeeklySchedule weeklySchedule = (BWeeklySchedule) ((BBacnetScheduleDescriptor) descriptor).getObject();
                if (weeklySchedule == null || weeklySchedule.getName() == null) {
                    exportTable.remove(((BBacnetScheduleDescriptor) descriptor).getName());
                }
            } else if (descriptor instanceof BBacnetCalendarDescriptor) {
                BCalendarSchedule calendarSchedule = (BCalendarSchedule) ((BBacnetCalendarDescriptor) descriptor).getObject();
                if (calendarSchedule == null || calendarSchedule.getName() == null) {
                    exportTable.remove(((BBacnetCalendarDescriptor) descriptor).getName());
                }
            }
        }
    }

    /**
     * Helper-Method to export all Weekly-Schedules and Calendar-Schedules in given BacnetNetwork
     *
     * @param bacnetNetwork
     */
    public static HonLimitationResult exportSchedulesAndCalenders(BBacnetNetwork bacnetNetwork) {

        if (bacnetNetwork == null) {
            HonPlantControllerHMILogger.checkAndLog(Level.FINE, "Could not export schedulesAndCalendars, bacnetNetwork was null");
            return HonLimitationResult.NOK;
        }

        BLocalBacnetDevice localDevice = bacnetNetwork.getLocalDevice();
        BBacnetObjectIdentifier deviceId = localDevice.getObjectId();
        BBacnetExportTable exportTable = (BBacnetExportTable) localDevice.getExportTable();

        final BOrd ord = BOrd.make("slot:/|bql:select * from schedule:AbstractSchedule");
        final BITable<BAbstractSchedule> table = (BITable<BAbstractSchedule>) ord.resolve(bacnetNetwork).get();
        final TableCursor<BAbstractSchedule> c = table.cursor();
        int exportedSchedulesCount = 0;
        int exportedCalendarCount = 0;
        List<BAbstractSchedule> irmScheduleList = new ArrayList<BAbstractSchedule>();
        
        for (BAbstractSchedule abstractSchedule : c) {
        	BComplex parentObject = abstractSchedule.getParent();
        	while(null!=parentObject) {
        		if(parentObject.getType().toString().equals(IRM_PROGRAM)) {
        			irmScheduleList.add(abstractSchedule);
        			break;
        		}
        		parentObject = parentObject.getParent();
        	}
        	
        	
            if (abstractSchedule instanceof BBooleanSchedule && !irmScheduleList.contains(abstractSchedule)) {
                if (exportedSchedulesCount >= HonHMILimitations.getMaxNumberOfSchedules()) {
                    return HonLimitationResult.EXCEEDED;
                }
                BHonBacnetBooleanScheduleDescriptor scheduleDescriptor = new BHonBacnetBooleanScheduleDescriptor();

                // HMI supports only enumeration for boolean schedules (work bench: 'BinaryPV' as Schedule Data Type)
                BDynamicEnum binaryPV = BDynamicEnum.make(BBacnetBooleanScheduleDescriptor.ENUMERATED_IDX,
                    BBacnetBooleanScheduleDescriptor.BOOL_DATA_TYPE_RANGE);
                scheduleDescriptor.setScheduleDataType(binaryPV);
                BObject trueTextObj = ((BBooleanSchedule) abstractSchedule).getFacets().get("trueText");
                BObject falseTextObj = ((BBooleanSchedule) abstractSchedule).getFacets().get("falseText");
                if (trueTextObj instanceof BString && falseTextObj instanceof BString) {
                    BBacnetArray makeScheduleStateText = HonBacnetScheduleUtil.makeScheduleStateText(new int[]{0, 1}, new String[]{falseTextObj.toString(), trueTextObj.toString()});
                    if (null != makeScheduleStateText) {
                        scheduleDescriptor.setScheduleStateText(makeScheduleStateText);
                    }
                }
                addObjPropRefs(scheduleDescriptor, abstractSchedule, BBooleanSchedule.out, deviceId,
                    BHonBacnetBinaryOutputDescriptor.TYPE, exportTable);
                exportSchedule(abstractSchedule, scheduleDescriptor, exportTable);
                ++exportedSchedulesCount;
            } else if (abstractSchedule instanceof BNumericSchedule && !irmScheduleList.contains(abstractSchedule)) {
                if (exportedSchedulesCount >= HonHMILimitations.getMaxNumberOfSchedules()) {
                    return HonLimitationResult.EXCEEDED;
                }
                BHonBacnetNumericScheduleDescriptor scheduleDescriptor = new BHonBacnetNumericScheduleDescriptor();

                addObjPropRefs(scheduleDescriptor, abstractSchedule, BNumericSchedule.out, deviceId,
                    BHonBacnetAnalogOutputDescriptor.TYPE, exportTable);
                exportSchedule(abstractSchedule, scheduleDescriptor, exportTable);
                ++exportedSchedulesCount;
            } else if (abstractSchedule instanceof BEnumSchedule && !irmScheduleList.contains(abstractSchedule)) {
                if (exportedSchedulesCount >= HonHMILimitations.getMaxNumberOfSchedules()) {
                    return HonLimitationResult.EXCEEDED;
                }
                BHonBacnetEnumScheduleDescriptor scheduleDescriptor = new BHonBacnetEnumScheduleDescriptor();
                BFacets facets = ((BEnumSchedule) abstractSchedule).getFacets();
                BObject range = facets.get("range");
                if (range instanceof BEnumRange) {
                    int[] ordinals = ((BEnumRange) range).getOrdinals();
                    if (null != ordinals && ordinals.length > 0) {
                        String[] tags = new String[ordinals.length];
                        for (int i = 0; i < tags.length; i++) {
                            tags[i] = ((BEnumRange) range).getTag(ordinals[i]);
                        }
                        BBacnetArray makeScheduleStateText = HonBacnetScheduleUtil.makeScheduleStateText(ordinals, tags);
                        if (null != makeScheduleStateText) {
                            scheduleDescriptor.setScheduleStateText(makeScheduleStateText);
                        }
                    }
                }

                addObjPropRefs(scheduleDescriptor, abstractSchedule, BEnumSchedule.out, deviceId,
                    BHonBacnetMultiStateOutputDescriptor.TYPE, exportTable);
                exportSchedule(abstractSchedule, scheduleDescriptor, exportTable);
                ++exportedSchedulesCount;
            } else if (abstractSchedule instanceof BStringSchedule && !irmScheduleList.contains(abstractSchedule)) {
                if (exportedSchedulesCount >= HonHMILimitations.getMaxNumberOfSchedules()) {
                    return HonLimitationResult.EXCEEDED;
                }
                BHonBacnetStringScheduleDescriptor scheduleDescriptor = new BHonBacnetStringScheduleDescriptor();
                addObjPropRefs(scheduleDescriptor, abstractSchedule, BStringSchedule.out, deviceId,
                    BHonBacnetStringScheduleDescriptor.TYPE, exportTable);
                exportSchedule(abstractSchedule, scheduleDescriptor, exportTable);
                ++exportedSchedulesCount;
            } else if (abstractSchedule instanceof BCalendarSchedule && !irmScheduleList.contains(abstractSchedule)) {
                if (exportedCalendarCount >= HonHMILimitations.getMaxNumberOfCalendars()) {
                    return HonLimitationResult.EXCEEDED_CALENDARS;
                }
                BBacnetCalendarDescriptor scheduleDescriptor = new BBacnetCalendarDescriptor();
                exportCalendar(abstractSchedule, scheduleDescriptor, exportTable);
                ++exportedCalendarCount;
            }
        }

        return HonLimitationResult.OK;
    }

    private static void addObjPropRefs(BBacnetScheduleDescriptor scheduleDescriptor, BAbstractSchedule
        abstractSchedule,
                                       Property outProp, BBacnetObjectIdentifier deviceId, Type typeOutput,
                                       BBacnetExportTable exportTable) {
        BBacnetListOf objPropRefs = scheduleDescriptor.getListOfObjectPropertyReferences();
        Knob[] outKnobs = abstractSchedule.getKnobs(outProp);
        for (Knob outKnob : outKnobs) {
            BComponent tgtComp = outKnob.getTargetComponent();
            if (tgtComp instanceof BControlPoint) {
                BIBacnetExportObject[] outputs = exportTable.getExportedObjects(typeOutput);
                for (BIBacnetExportObject output : outputs) {
                    if (tgtComp.getHandleOrd().equals(output.getObjectOrd())) {
                        BBacnetDeviceObjectPropertyReference devObjPropRef;
                        devObjPropRef = new BBacnetDeviceObjectPropertyReference(output.getObjectId());
                        devObjPropRef.setDeviceId(deviceId);
                        objPropRefs.addListElement(devObjPropRef, null);
                    }
                }
            }
        }
    }

    private static BHonPlantControllerHMI getHMIDevice() {
        BHonPlantControllerHMI hmiDevice = null;

        BIHonPlantControllerHMIInterface honPlantControllerMMIInterface = BIHonPlantControllerHMIInterface.getHonPlantControllerHMIInterface();
        if (null != honPlantControllerMMIInterface) {
            hmiDevice = honPlantControllerMMIInterface.getHMIDevice();
        }

        return hmiDevice;
    }

    private static String getText(String lexiconMessage, Object... obj) {
        return HonPlantControllerHMILexicon.getLexicon().getText(lexiconMessage, obj);
    }
    
    public static final String IRM_PROGRAM = "honIrmConfig:IrmProgram";
    private static final String SLOT = "slot:/";
    private static final String DESCRIPTION = "description";
    private static final String PROXY_EXT = "proxyExt";
}

