/*
Copyright (c) 2022 Honeywell International, Inc. All rights reserved.
This file contains trade secrets of Honeywell International, Inc. No part
may be reproduced or transmitted in any form by any means or for any
purpose without the express written permission of Honeywell.
*/


package com.honeywell.honplantcontroller.hmi.utils;



import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import javax.baja.alarm.BAlarmService;
import javax.baja.bacnet.BBacnetNetwork;
import javax.baja.bacnet.datatypes.BBacnetObjectIdentifier;
import javax.baja.bacnet.enums.BBacnetObjectType;
import javax.baja.bacnet.export.BIBacnetExportObject;
import javax.baja.bacnet.export.BLocalBacnetDevice;
import javax.baja.driver.BDriverContainer;
import javax.baja.sys.BComponent;
import javax.baja.sys.BStation;
import javax.baja.sys.Context;
import javax.baja.sys.Flags;
import javax.baja.sys.Property;
import javax.baja.util.BFormat;

import com.honeywell.honbacnethelper.export.BHonBacnetEventLogDescriptor;
import com.honeywell.honbacnethelper.export.BHonDeviceExtDescriptor;
import com.honeywell.honplantcontroller.hmi.alarm.BHMIAlarmConsoleRecipient;
import com.honeywell.honplantcontroller.hmi.comm.BHonPlantControllerBacnetProperties;
import com.honeywell.honplantcontroller.hmi.comm.BHMIBacnetMstpLinkLayer;
import com.tridium.bacnet.stack.BBacnetStack;
import com.tridium.bacnet.stack.link.BBacnetLinkLayer;
import com.tridium.bacnet.stack.link.mstp.BBacnetMstpLinkLayer;
import com.tridium.bacnet.stack.network.BBacnetNetworkLayer;
import com.tridium.bacnet.stack.network.BNetworkPort;
import com.tridium.bacnet.stack.server.BBacnetExportTable;


public class HonPlantControllerBacnetUtils {

	private HonPlantControllerBacnetUtils() {
    //This is a utility-class, so there is no need to instanciate it.
  }

  public static BBacnetNetwork findBacnetNetwork(BStation station) {

    if(station == null) {
      return null;
    }

    BDriverContainer[] driverContainer = station.getChildren(BDriverContainer.class);
    if(null == driverContainer || driverContainer.length <= 0)
    	return null;
    return Arrays.stream(driverContainer[0].getChildComponents())
    .filter(BBacnetNetwork.class::isInstance)
    .map(BBacnetNetwork.class::cast)
    .findFirst().orElse(null);
  }
    public static Set<BBacnetNetwork> findBacnetNetworks(BStation station) {

        if(station == null) {
            return null;
        }

        BDriverContainer[] driverContainer = station.getChildren(BDriverContainer.class);
        if(null == driverContainer || driverContainer.length <= 0)
            return null;
        return Arrays.stream(driverContainer[0].getChildComponents())
                .filter(BBacnetNetwork.class::isInstance)
                .map(BBacnetNetwork.class::cast)
                .collect(Collectors.toSet());
    }

  /**
   * Helpermethod to find find the bacnetnetwork, that is used for communication with HMI
   * @param station
   * @param bacnetProperties
   * @return
   */
//  public static BBacnetNetwork findMMIBacnetNetwork(BStation station, BHonPlantControllerBacnetProperties bacnetProperties) {
//
//    if(station == null || bacnetProperties == null) {
//      return null;
//    }
//    BDriverContainer driverContainer = (BDriverContainer)station.get(DRIVERS);
//    if(driverContainer == null) {
//      return null;
//    }
//
//    return Arrays.stream(driverContainer.getChildComponents())
//    .filter(BBacnetNetwork.class::isInstance)
//    .map(BBacnetNetwork.class::cast)
//    // .filter(network -> network.getLocalDevice().getObjectId().equals(bacnetProperties.getObjectId()))
//    .filter(network -> isMMIBacnetNetwork(network, bacnetProperties))
//    .findFirst().orElse(null);
//  }

  public static boolean compareMstpLinkLayers(BHMIBacnetMstpLinkLayer first, BHMIBacnetMstpLinkLayer second) {

    if(first == null || second == null) {
      return false;
    }
    //As discussed previously we should only compare  the PortName
     return first.getPortName().equals(second.getPortName()) ;
  }

  public static boolean compareMstpPorts(BNetworkPort first, BNetworkPort second) {

    if(first == null || second == null) {
      return false;
    }

    BBacnetLinkLayer linkLayer = first.getLink();
    if(linkLayer instanceof BHMIBacnetMstpLinkLayer) {
      BHMIBacnetMstpLinkLayer networksMstpLink = (BHMIBacnetMstpLinkLayer) linkLayer;
      BHMIBacnetMstpLinkLayer hmiMstpLinkLayer = (BHMIBacnetMstpLinkLayer)second.getLink();
      return HonPlantControllerBacnetUtils.compareMstpLinkLayers(networksMstpLink, hmiMstpLinkLayer);
    }

    return false;
  }

  public static BNetworkPort getMstpPorts(BNetworkPort first, BNetworkPort second) {

	    if(first == null || second == null) {
	      return null;
	    }

	    BBacnetLinkLayer linkLayer = first.getLink();
	    if(linkLayer instanceof BHMIBacnetMstpLinkLayer) {
        BHMIBacnetMstpLinkLayer networksMstpLink = (BHMIBacnetMstpLinkLayer) linkLayer;
        BHMIBacnetMstpLinkLayer hmiMstpLinkLayer = (BHMIBacnetMstpLinkLayer)second.getLink();
	      if(HonPlantControllerBacnetUtils.compareMstpLinkLayers(networksMstpLink, hmiMstpLinkLayer)) {
	    	  networksMstpLink.setMstpAddress(0);
	    	  networksMstpLink.setMaxMaster(1);
	          
	          markPropertyReadonly(networksMstpLink, BHMIBacnetMstpLinkLayer.portName);
	          markPropertyReadonly(networksMstpLink, BHMIBacnetMstpLinkLayer.baudRate);
	          markPropertyReadonly(networksMstpLink, BHMIBacnetMstpLinkLayer.mstpAddress);
	          markPropertyReadonly(networksMstpLink, BHMIBacnetMstpLinkLayer.maxMaster);

	    	  return first;
	      }
	    }

	    return null;
	  }

  /**
   * Helper-Method to check, if a BacnetNework is used for communication with HMI
   * @param bacnetNetwork
   * @param bacnetProperties
   * @return
   */
  public static boolean isHMIBacnetNetwork(BBacnetNetwork bacnetNetwork, BHonPlantControllerBacnetProperties bacnetProperties) {

    if(bacnetNetwork == null || bacnetProperties == null) {
      return false;
    }

    if(!bacnetNetwork.getLocalDevice().getObjectId().equals(bacnetProperties.getObjectId())) {
      return false;
    }

    return containsHMIPort(bacnetNetwork, bacnetProperties);
  }

  /**
   * Helper method to check if given bacnetNetwork contains an HMI-Port
   * @param bacnetNetwork
   * @param bacnetProperties
   * @return
   */
  public static boolean containsHMIPort(BBacnetNetwork bacnetNetwork, BHonPlantControllerBacnetProperties bacnetProperties) {
	if(null == bacnetNetwork)
		return false;
    BBacnetNetworkLayer bacnetNetworkLayer = ((BBacnetStack) bacnetNetwork.getBacnetComm()).getNetwork();
    BNetworkPort[] ports = bacnetNetworkLayer.getChildren(BNetworkPort.class);
    for (BNetworkPort networkPort : ports) {
    	boolean portFound = compareMstpPorts(networkPort, bacnetProperties.createMstpPort());
    	if(portFound) {
    		return portFound;
    	}
	}
    return false;
  }

  public static void removeDuplicatedNetworkPort(BBacnetNetwork bacnetNetwork, BHonPlantControllerBacnetProperties bacnetProperties) {
        if(null == bacnetNetwork)
            return;
        BBacnetNetworkLayer bacnetNetworkLayer = ((BBacnetStack) bacnetNetwork.getBacnetComm()).getNetwork();
        BNetworkPort[] ports = bacnetNetworkLayer.getChildren(BNetworkPort.class);
        ArrayList<BNetworkPort> duplicatedHmiPortList = new ArrayList<BNetworkPort>();
        for (BNetworkPort networkPort : ports) {
            BBacnetLinkLayer linkLayer = networkPort.getLink();
            if(linkLayer instanceof BBacnetMstpLinkLayer && ((BBacnetMstpLinkLayer) linkLayer).getPortName().equals(bacnetProperties.getPortName())) {
                if (linkLayer instanceof BHMIBacnetMstpLinkLayer) {
                    duplicatedHmiPortList.add(networkPort);
                } else {
                    bacnetNetworkLayer.remove(networkPort);
                }
            }
        }

        if(duplicatedHmiPortList.size()>1){
            duplicatedHmiPortList.stream().sorted(Comparator.comparing(BNetworkPort::getEnabled,Comparator.reverseOrder())).collect(Collectors.toList());
            duplicatedHmiPortList.remove(1);
            for (BNetworkPort duplicatedHMIPort : duplicatedHmiPortList) {
                bacnetNetworkLayer.remove(duplicatedHMIPort);
            }
        }
  }

  public static boolean isRunningOnController() {
        // look OS string (force to upper case and remove leading and trailing whitespace
        String os = System.getProperty("os.name").toUpperCase().trim();

        boolean ret = false;

        // compare with known OS name
        if (os.equals("QNX") || os.equals("LINUX"))
            ret = true;

        return ret;
  }

  public static BNetworkPort getHMIPort(BBacnetNetwork bacnetNetwork, BHonPlantControllerBacnetProperties bacnetProperties) {
	    BBacnetNetworkLayer bacnetNetworkLayer = ((BBacnetStack) bacnetNetwork.getBacnetComm()).getNetwork();
	    BNetworkPort[] ports = bacnetNetworkLayer.getChildren(BNetworkPort.class);
	    for (BNetworkPort networkPort : ports) {
	    	BNetworkPort mstpPort = getMstpPorts(networkPort, bacnetProperties.createMstpPort());
	    	if(null != mstpPort) {
	    		return mstpPort;
	    	}
		}
	    return null;
	  }

  /**
   * Helper-Method to initialize a BacnetNetwork, that will be used for communication with HMI
   * @param station
   * @param bacnetProperties
   * @return
   */
  public static BBacnetNetwork createHMIBacnetNetwork(BStation station, BHonPlantControllerBacnetProperties bacnetProperties) {
    if(station == null || bacnetProperties == null) {
      return null;
    }

    BBacnetNetwork bacnetNetwork = new BBacnetNetwork();
    bacnetProperties.createHMIPort(bacnetNetwork);
    bacnetProperties.setBacnetSettings(bacnetNetwork);

    addBacnetNetworkToStation(bacnetNetwork, station);

    try {
      BNetworkPort networkPort = (BNetworkPort) ((BBacnetStack) bacnetNetwork.getBacnetComm()).getNetwork().get("mstPort");

      // TODO: commented out because it fails on windows with nullpointer
      // networkPort.enable();
      // if (networkPort.getFaultCause().equalsIgnoreCase("")) {
      // }

      return bacnetNetwork;
    } catch (Exception e) {
      LOG.warning("Exception enabling BACnet network port: " + e.getMessage());
    }

    return null;
  }

  public static void addBacnetNetworkToStation(BBacnetNetwork bacnetNetwork, BStation station) {
    if(bacnetNetwork == null || station == null) {
      return;
    }
    BDriverContainer[] driverContainer = station.getChildren(BDriverContainer.class);
    if(null == driverContainer || driverContainer.length <= 0)
    	return;

    Property bacnetNetworkProperty = driverContainer[0].add(BACNET_NETWORK+"?", bacnetNetwork);
    driverContainer[0].setDisplayName(bacnetNetworkProperty, BFormat.make("%lexicon(honPlantController:BacnetNetwork)%"),Context.NULL);
  }

  public static boolean addHMIPort(BBacnetNetwork bacnetNetwork, BHonPlantControllerBacnetProperties bacnetProperties) {
    return bacnetProperties.createHMIPort(bacnetNetwork);
  }

  public static BHMIAlarmConsoleRecipient getHMIAlarmConsoleRecipient(BStation station) {

    BAlarmService alarmService = station.getServices().getChildren(BAlarmService.class)[0];

    BHMIAlarmConsoleRecipient alarmConsoleRecipient = alarmService.getChildren(BHMIAlarmConsoleRecipient.class)[0];

    return alarmConsoleRecipient;
  }


	public static void markPropertyHidden(BComponent component, Property property) {
		if(null == component || null == property || null == component.get(property))
			return;
		int flags = component.getFlags(property);
		flags = flags | Flags.HIDDEN;
		component.setFlags(property, flags);
	}

	public static void markPropertyReadonly(BComponent component, Property property) {
		if(null == component || null == property || null == component.get(property))
			return;
		int flags = component.getFlags(property);
		flags = flags | Flags.READONLY;
		component.setFlags(property, flags);
	}

	public static void markPropertyReadonlyAndHidden(BComponent component, Property property) {
		if(null == component || null == property || null == component.get(property))
			return;

		markPropertyHidden(component, property);
		markPropertyReadonly(component, property);
	}

  /**
   * Setting brand information to be provided to hmi over bacnet
   * @param bacnetNetwork
   */
  public static void setBrandInformation(BBacnetNetwork bacnetNetwork) {

    if(null == bacnetNetwork ) {
      return;
    }
    BLocalBacnetDevice localBacnetDevice = bacnetNetwork.getLocalDevice();

    localBacnetDevice.setVendorName(HONEYWELL_BACNET_VENDOR_NAME);
		localBacnetDevice.setVendorId(HONEYWELL_BACNET_VENDOR_ID);

    //TODO: figure out how to get the modelname from platform
		// localBacnetDevice.setModelName(BACNET_SKU_NAME);

  }

  /**
   * Export proprietary object for HMI to consume
   * @param bacnetNetwork
   */
  public static void exportDeviceExtObject(BBacnetNetwork bacnetNetwork) {

    if(null == bacnetNetwork) {
      return;
    }

    BBacnetExportTable exportTable = (BBacnetExportTable) bacnetNetwork.getLocalDevice().getExportTable();

    BIBacnetExportObject foundExportObject = exportTable.byObjectId(BBacnetObjectIdentifier.make(BHonDeviceExtDescriptor.HMI_DEVICE_EXT_OBJECT_TYPE, 0));
    BHonDeviceExtDescriptor deviceExtDescriptor;
    if(null == foundExportObject) {
      deviceExtDescriptor = new BHonDeviceExtDescriptor();
      int instanceNumer = exportTable.getNextInstance(BHonDeviceExtDescriptor.HMI_DEVICE_EXT_OBJECT_TYPE);
      String result = exportTable.export(deviceExtDescriptor);
      //will probably always be null, but just to make sure
      if(result == null) {
        exportTable.add(HON_DEVICE_EXT_NAME+"_"+instanceNumer, deviceExtDescriptor, Flags.HIDDEN);
      }
    }
    else {
      deviceExtDescriptor = (BHonDeviceExtDescriptor)foundExportObject;
    }
    HonPlantControllerBacnetUtils.initializeDeviceExtObject(deviceExtDescriptor);
  }

  /**
   * Export event log object for HMI to consume
   * @param bacnetNetwork
   */
  public static void exportEventLogObject(BBacnetNetwork bacnetNetwork) {

    if(null == bacnetNetwork) {
      return;
    }

    BBacnetExportTable exportTable = (BBacnetExportTable) bacnetNetwork.getLocalDevice().getExportTable();

    BIBacnetExportObject foundExportObject = exportTable.byObjectId(BBacnetObjectIdentifier.make(BBacnetObjectType.EVENT_LOG, 500));
    if(null == foundExportObject) {
      BHonBacnetEventLogDescriptor eventLogDescriptor = new BHonBacnetEventLogDescriptor();
      String result = exportTable.export(eventLogDescriptor);
      //will probably always be null, but just to make sure
      if(result == null) {
        exportTable.add("eventLog_500", eventLogDescriptor, Flags.HIDDEN);
        return;
      }
    }
    LOG.warning("HonBacnetEventLogDescriptor already present in export table");
  }

  /**
   * Initializing {@link BHonDeviceExtDescriptor} with user number and user status
   * @param deviceExtDescriptor
   */
  public static void initializeDeviceExtObject(BHonDeviceExtDescriptor deviceExtDescriptor) {
    long userNumber = HonPlantControllerUserUtils.getNiagaraUserCount();

    int userStatus = HonPlantControllerUserUtils.getUserStatusAsInt(HonPlantControllerUserUtils.getNiagaraUserStatus());
    deviceExtDescriptor.setUserNumber((int) userNumber);
    deviceExtDescriptor.setUserStatus(userStatus);    

  }

  public static void handleRemoveFastAccessListEntry() {

  }

	public static final Logger LOG = Logger.getLogger(HonPlantControllerBacnetUtils.class.getSimpleName());

  private static final String BACNET_NETWORK = "BacnetNetwork";


  public static final int HONEYWELL_BACNET_VENDOR_ID = 17;
  public static final String HONEYWELL_BACNET_VENDOR_NAME = "Honeywell";
  public static final String BACNET_SKU_NAME = "HON-CLBESER";
  public static final String HON_DEVICE_EXT_NAME = "honDeviceExtDescriptor";
  public static final String DRIVERS = "Drivers";
}
