package com.honeywell.honplantcontroller.hmi.utils;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.baja.naming.BOrd;
import javax.baja.naming.SlotPath;
import javax.baja.query.BQueryResult;
import javax.baja.security.BAbstractAuthenticator;
import javax.baja.security.BPassword;
import javax.baja.security.BPasswordAuthenticator;
import javax.baja.sys.BComplex;
import javax.baja.sys.BComponent;
import javax.baja.sys.Sys;
import javax.baja.tag.Entity;
import javax.baja.user.BUser;
import javax.baja.user.BUserService;

import com.honeywell.honplantcontroller.hmi.limitations.HonHMILimitations;
import com.honeywell.honplantcontroller.hmi.user.authenticator.BHonPlantControllerHMIAuthenticator;

/**
 * HonPlantControllerUserUtils
 */
public class HonPlantControllerUserUtils {

  ///CLOVER:OFF
  private HonPlantControllerUserUtils() {}
  ///CLOVER:ON

    /**
     * Provide a Stream of the all Users found by searching for {@link BHonPlantControllerHMIAuthenticator}.
     * The defaultPrototype User which is found in addition is filtered out and the list is also sorted by the user name.
     * @return Stream<Entity> found, filtered and sorted Stream of Users in the {@code baja.BUserService}
     */
  public static Stream<BUser> getAllNiagaraUsers() {
    BOrd query = BOrd.make("neql: n:type like 'honPlantControllerHMI:HonPlantControllerHMIAuthenticator'");
    BQueryResult queryResult = (BQueryResult)query.resolve(Sys.getService(BUserService.TYPE)).get();

    // update using:
    // UserService/User enabled/disabled
    // UserService/User Correct Password
    // UserServiec/User lockout
    // UserService/User Authenticator
    // HMI Authenticator: enabled
    // HMI Authenticator: expiration property

    return queryResult.stream()
        // check if the Authenticator is enabled and not expired
        .filter(entry -> ((BHonPlantControllerHMIAuthenticator)entry).canLogin())
        .map(HonPlantControllerUserUtils::toUser)
        // ignore users which are disabled
        .filter(entry -> entry.getEnabled())
        // Ignore the users which has default password.
        .filter(entry -> !entry.getChildren(BHonPlantControllerHMIAuthenticator.class)[0].getPin().equals(BPassword.DEFAULT))
        // Ignore the defaultPrototype Component which is actually not a user
        .filter(entry -> !entry.getName().equals("defaultPrototype"))
        // Ignore the BACnet user because this user is not used to login to any device.
        .filter(entry -> !(entry).getName().equals("BACnet"))
        // ignore locked out users
        .filter(entry -> !entry.getLockOut())
        // Authenticator should also be enabled!
        .filter(entry -> ((BPasswordAuthenticator)entry.getAuthenticator()).canLogin(entry))
        .distinct();
  }
  
  public static Stream<BUser> getAllNiagaraUsersWithInvalidPin() {
	    BOrd query = BOrd.make("neql: n:type like 'honPlantControllerHMI:HonPlantControllerHMIAuthenticator'");
	    BQueryResult queryResult = (BQueryResult)query.resolve(Sys.getService(BUserService.TYPE)).get();

	    // update using:
	    // UserService/User enabled/disabled
	    // UserService/User Incorrect Password
	    // UserServiec/User lockout
	    // UserService/User Authenticator
	    // HMI Authenticator: enabled
	    // HMI Authenticator: expiration property

	    return queryResult.stream()
	        // check if the Authenticator is enabled and not expired
	        .filter(entry -> ((BHonPlantControllerHMIAuthenticator)entry).canLogin())
	        .map(HonPlantControllerUserUtils::toUser)
	        // ignore users which are disabled
	        .filter(entry -> entry.getEnabled())
	        // Ignore the users which has default password.
	        .filter(entry -> entry.getChildren(BHonPlantControllerHMIAuthenticator.class)[0].getPin().equals(BPassword.DEFAULT))
	        // Ignore the defaultPrototype Component which is actually not a user
	        .filter(entry -> !entry.getName().equals("defaultPrototype"))
	        // Ignore the BACnet user because this user is not used to login to any device.
	        .filter(entry -> !(entry).getName().equals("BACnet"))
	        // ignore locked out users
	        .filter(entry -> !entry.getLockOut())
	        // Authenticator should also be enabled!
	        .filter(entry -> ((BPasswordAuthenticator)entry.getAuthenticator()).canLogin(entry))
	        .distinct();
	  }
  
  /**
   * Gets the list up to the max. of 20 ({@link HonHMILimitations#getMaxNumberOfUsers}) Users.
   * @return Stream of 
   */
  private static Stream<BUser> getLimitedNiagaraUsers() {
    // As per HMI spec only a maximum of 20 Users will be supported.
    return getAllNiagaraUsers().limit(HonHMILimitations.getMaxNumberOfUsers());
  }

  /**
   * Get the number of users.
   * Minimum 1 user and Maximum 20 users.
   * @return number of BHonPlantControllerHMIAuthenticator up to a max. of 20 Entries
   */
  public static long getNiagaraUserCount() {
    return getLimitedNiagaraUsers().count();
  }

  /**
   * Gets a String list of all Niagara users configured with bHonPlantControllerHMIAuthenticator.
   * @return All Niagara Users configured to be supported by the HMI.
   */
  public static List<String> getNiagaraUserList() {
    return getLimitedNiagaraUsers()
                            // Only populate the user name
                            .map(BUser::getName)
                            // filter out all null objects
                            .filter(Objects::nonNull)
                            .collect(Collectors.toList());
  }

  /**
   * Gets a String list of all Niagara users (escaped) configured with bHonPlantControllerHMIAuthenticator.
   * @return All Niagara Users (escaped) configured to be supported by the HMI.
   */
  public static List<String> getUnescapedNiagaraUserList() {
    List<String> users = new ArrayList<>();
    getNiagaraUserList().forEach((e) -> users.add(SlotPath.unescape(e)));

    users.forEach((e) -> HonPlantControllerHMILogger.finest("[getEscapedNiagaraUserList] user: " + e));

    return users;
  }

  /**
   * Tries to get the user which is named after {@code name} from the UserService.
   * If there are more users found than needed only the first in the list will be
   * returned.
   * @param name of the user to find.
   * @return An Optional which contains the user if one is found or non if not.
   */

  public static Optional<BUser> getNiagaraUserbyName(String name) {
    return getAllNiagaraUsers()
                            .filter(Objects::nonNull)
                            .filter(user -> user.getName().equals(name))
                            .findFirst();
  }

  public static List<BUser> getUsers() {
    return getLimitedNiagaraUsers()
                                  .filter(Objects::nonNull)
                                  .collect(Collectors.toList());
  }

  /**
   * Gets a Boolean list of the Status of all Niagara users configured with bHonPlantControllerHMIAuthenticator.
   * NOTE:
   *    - the {@link BPasswordAuthenticator.canLogin} will return true even if expired.
   * @return {@code true} if user is enabled and {@link BAbstractAuthenticator.canLogin} and {@link BHonPlantControllerHMIAuthenticator.canLogin}
   */
  public static List<Boolean> getNiagaraUserStatus() {

    return getLimitedNiagaraUsers().map(HonPlantControllerUserUtils::toUserStatus)
                                    .filter(Objects::nonNull)
                                    .collect(Collectors.toList());
  }

  /**
   * Get a an integer representation of a list of boolean
   * @param userStatus
   * @return
   */
  public static int getUserStatusAsInt(List<Boolean> userStatus) {
    BitSet bitSet = new BitSet(userStatus.size());
    for (int i = 0; i < userStatus.size(); i++) {
      bitSet.set(i, userStatus.get(i));
    }

    int userStatusInteger = 0;

    if(bitSet.length() > 0) {
      userStatusInteger = (int)bitSet.toLongArray()[0];
    }

    return userStatusInteger;
  }
  
  /**
   * Converts the User into either an Boolean<true/false> or null.
   * @param user
   * @return null if entity is not valid or a Boolean if user is valid.
   */
  @SuppressWarnings({
    "squid:S2447" // return null is expected to filter out any users which are not valid for HMI.
  })
  private static final Boolean toUserStatus(BUser user) {
      // No need to check coverage because entry will always be an BHonPlantControllerHMIAuthenticator
      ///CLOVER:OFF
	  try {
        BHonPlantControllerHMIAuthenticator hmiAuthenticator = user.getChildren(BHonPlantControllerHMIAuthenticator.class)[0];
        BAbstractAuthenticator pwAuthenticator = user.getAuthenticator();
        //HMI expect 0=normal user status, 1=user is locked
        return !(user.getEnabled() && hmiAuthenticator.canLogin(user) && pwAuthenticator.canLogin(user));
	  }catch (Exception e) {
		return true;
	}
      ///COLVER:ON
  }

  ///CLOVER:OFF
  private static final BUser toUser(Entity entity) {
    if (entity instanceof BComponent) {
      BComponent comp = (BComponent)entity;

      BComplex parent = comp.getParent();
      return (BUser)parent;
    }
    return null;
  }
  ///CLOVER:ON
}
